#=========================================================================================================
#                                      CreateProjectBoundaries
#=========================================================================================================
#
# Author(s)         :  Clover Point Cartographics (Mike Neal)
#
# Date              :  2009-04-08
#
# Purpose           :  This script is used to create projet boundaries which illustrate the overall
#                      boundaries of each mapping project (BADID). It dissolves boundaries between all
#                      pairs of adjacent polygons that share the same BAPID number, resulting in a polygon
#                      feature class that depicts the overall boundaries of each mapping project. During
#                      the process, all attributes other than BAPID are eliminated. Attributes are then
#                      added to the resulting feature class from the Project Details table.
#
# Arguments         :  argv(1) = INPUT_MLT  (Input feature class for use in creating project boundaries)
#                      argv(2) = INPUT_PD   (Project Details table: either an FGDB table, or the tblProjects
#                                            table from the BAPID Tracker database; see example below)
#
# Example Input     :  X:\fullpath\Tool_CreateProjectBoundaries.py X:\fullpath\TEI_Long_Tbl Full_path_to_project_details
#                      (the 2nd argument must refer either to a TEI_Project_Details table or to the tblProjects table
#                      in the Access database as it would be accessed through ArcCatalog, for example:
#                      X:\fullpath\BAPID_Tracker_BE.mdb\tblProjects)
#
# Outputs           :  TEIS_Master_Project_Boundaries or TEI_Project_Boundaries feature class in the 
#                      Operational_Data.gdb file geodatabase
#
# Dependencies      :  Requires SOILS_Master folder at same folder level as root TEIS_Environment folder
#                      (if you want Soils project boundaries to be included in the output)
#                      If the script Dissolve.py is present in the same folder as this script, it is
#                      invoked to dissolve each slice. In testing, it was found that repeatedly dissolving
#                      within the same script can lead to the script eventually crashing for unknown
#                      reasons (memory leak?). By calling an external script to perform each dissolve, it
#                      seems to eliminate this problem. If the Dissolve.py script is not in place, this
#                      script will run the Dissolve tool each time as normal, and the script may or may not
#                      eventually crash.
#
# Expected Duration :  < 5 minutes for approximately 100,000 records
#                      Approx. 30 hours for 24 million records
#                       On CP Processor server takes ~ 5 hours.
#
# History           :  2012-09-17 (JK): Incorporated "SliceProcessing" logic directly into this script;
#                                       initially the script cursors through the records to obtain 
#                                       OBJECTID values to act as "markers" every 100,000 records; then
#                                       it loops through each set of 100,000 records and dissolves them,
#                                       appending the result to a cumulative dataset; then it does a final
#                                       dissolve, and attaches the Project Details attributes to the final
#                                       result. Runs much faster than dissolving the entire dataset at
#                                       once, and eliminates risk of running out of memory on large datasets.
#
#                   :  2012-10-12 (JK): Update to CPMessage routine: only append messages to PROCESS_LOG
#                                       if the e-mail function is being used (otherwise it can fill memory).
#                                       Added code to use newer, faster cursor object to enumerate slices
#                                       (first step of CreateProjectBoundaries routine) if Arc 10.1 is installed.
#
#                   :  2012-12-27 (JK): Modified script to allow for reading project details from Access 
#                                       database, and to add only the necessary fields from that table to
#                                       the Project Boundaries attribute table.
#
#                   :  2013-01-30 (JK): Added code to check that all BAPIDs in the long table feature class
#                                       are present in the project details table.
#
#                   :  2013-05-23 (JK): Update "Example Input" comments above (ODBC connection to BAPID Tracker
#                                       mdb file not required; you can specify the Windows path to it).
#
#                   :  2013-06-20 (JK): Added error message explaining that any missing BAPIDs in the BAPID 
#                                       Tracker need to be added before running this script again. Added
#                                       examples to message explaining script usage (displays if incorrect 
#                                       number of arguments are provided). Added a check for whether the 
#                                       datasets provided as arguments actually exist. Updated header comments
#                                       (dependencies). Removed code that creates a list of fields to join
#                                       from tblProjects into the dissolved Boundaries feature class, as
#                                       specifying a join field list causes a RunTimeError in Arc 9.3 (it's
#                                       an ESRI bug). Added stra() function which replaces unicode characters 
#                                       with a question mark, and replaced all str() functions throughout 
#                                       script with stra().
#
#                   :  2014-09-18 (EP): Updated script to ask for the location of the soils environment rather 
#                                       than having the TEIS MAster and the Soil master in the same folder.
#
#                   :  2015-11-25 (EP): Updated script to loop through Soils Project Boundaries rather than 
#                                       the soils_Short_table .
#
#=========================================================================================================
#=========================================================================================================
#  Process Status and Return Codes:
#      -1   - Unexpected error
#       0   - Completed Successful Run
#       1   - In progress
#       100 - Invalid Arguments
#       101 - ESRI License Error
#
#=========================================================================================================

#*********************************************************************************************************
#  Program Initialization
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
#  Python and System Modules
#---------------------------------------------------------------------------------------------------------

from __future__ import division

import sys
import string
import os
import time
import arcpy
import pdb


#---------------------------------------------------------------------------------------------------------
#  Arguement Validation
#---------------------------------------------------------------------------------------------------------

# Capture/Validate Parameters
#  Test that the right number of paramters were provided and validated
#  Note that 1 must be added because the Python script is an arguement (0) supplied to the
#  Python Environment.


# If long text parameters are required, use:  any_value = arcpy.GetParameterAsText(0)

if len(sys.argv) <> 4:

    print "\nInvalid number of arguments (" + str(len(sys.argv) - 1) + " supplied) to:"
    print "    " + sys.argv[0] + "\n"
    print " Usage:   " + sys.argv[0] + " [full path to long table feature class] [full path to tblProjects or TEI_Project_Details table]"
    print " Example 1:   " + sys.argv[0] + r" X:\fullpath\Operational_Data.gdb\TEIS_Master_Long_Tbl Y:\fullpath\BAPID_Tracker_BE.mdb\tblProjects W:\fullpath\Soils_master"
    print " Example 2:   " + sys.argv[0] + r" X:\fullpath\Operational_Data.gdb\TEI_Long_Tbl X:\fullpath\Operational_Data.gdb\TEI_Project_Details W:\fullpath\Soils_master"
    sys.exit(100)

PYTHON_SCRIPT      = sys.argv[0]
INPUT_MLT          = sys.argv[1]
INPUT_PD           = sys.argv[2]
INPUT_SOIL         = sys.argv[3]

#---------------------------------------------------------------------------------------------------------
#  Standard/Global Variables/Constants
#---------------------------------------------------------------------------------------------------------

# Global Constants
COMPUTERNAME     = os.environ['COMPUTERNAME']
USERNAME         = os.environ['USERNAME']
START_TIME       = time.ctime(time.time())
START_TIME_SEC   = time.time()
START_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
BEEP             = chr(7)


# Global Variables
#   The Processing_Variables is the collection of messaging and script specific information needed
#   throughout this program.  It is constantly updated, and generally includes all paths and configuration
#   information.  For parameters that are passed in to the script, they should generally be assigned
#   to a Processing_Variable through the Initialize() routine.

Processing_Variables = {}



#---------------------------------------------------------------------------------------------------------
#  CPC Variables/Constants
#---------------------------------------------------------------------------------------------------------




#---------------------------------------------------------------------------------------------------------
#  Set up the ArcGIS Geoprocessing Environment
#---------------------------------------------------------------------------------------------------------

d = arcpy.GetInstallInfo('desktop')
if float(d['Version'][:4]) >= 10.1:
    del arcpy
    import arcpy

#---------------------------------------------------------------------------------------------------------
#  Script Specific Variables/Constants
#---------------------------------------------------------------------------------------------------------

Processing_Variables['BAPID_LIST'] = ['ALL']
OVERWRITE_OPTION = 'REPLACE'



#*********************************************************************************************************
#  Routines
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# TemplateRoutine(Input_FC)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TemplateRoutine(Input_FC):
    CPMessage('    - Template Routine', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TemplateRoutine" 

    result = arcpy.GetCount_management(Input_FC)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Calculating individual values for' + str(Number_of_Features) + ' features.', 4)
    rows = arcpy.SearchCursor(Input_FC)
    row = rows.Next()
    
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    del row
    del rows

    return(Processing_Variables['PROCESS_STATUS'])

#=========================================================================================================
#  Standard Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# stra(value)
#
#       Converts any value to a string. Replaces Python's stra() which incurs an error if the value
#       happens to have a unicode character in it. Unicode characters will be replaced with ?.
#
#---------------------------------------------------------------------------------------------------------
def stra(value):
    return unicode(value).encode('ascii','replace')

#---------------------------------------------------------------------------------------------------------
# Initialize()
#
#       Set up all of the paths and variables that we need to run this script.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def Initialize():
    print '    - Initializing Script'
    
    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Script_Name'] = os.path.split(PYTHON_SCRIPT)[1]
    Processing_Variables['Log_Directory_Path'] = Processing_Variables['Script_Directory_Path'] + r'\log'

    Processing_Variables['Project_Base_Path']      = os.path.split(Processing_Variables['Script_Directory_Path'])[0]
    Processing_Variables['TEI_Master_Path']        = os.path.split(Processing_Variables['Project_Base_Path'])[0]

    
    Processing_Variables['ToolBoxRun'] = 'FALSE'
    # Processing_Variables['SendRunLogToEmail'] = ['mneal@cloverpoint.com', 'chaos_mike@hotmail.com']
    Processing_Variables['Log_Level'] = 4
    Processing_Variables['PROCESS_INFO']   = "RUNNING " + PYTHON_SCRIPT
    Processing_Variables['PROCESS_STATUS'] = 1
    Processing_Variables['PROCESS_LOG'] = []
    Processing_Variables['MessageQueueFile']   = Processing_Variables['Log_Directory_Path'] + '\\' + Processing_Variables['Script_Name'] + '.' + COMPUTERNAME + '.log'

    #### Adding a FGDB Log will significantly slow down performance.
    # Processing_Variables['Log_FGDB_Table'] = Processing_Variables['Project_Base_Path'] + r"\Operational.gdb\Processing_Log"

    if not arcpy.Exists(Processing_Variables['Log_Directory_Path']):
        os.mkdir(Processing_Variables['Log_Directory_Path'])
        CPMessage(r'Created \Scripts\log folder', 1)

    # Load required toolboxes...
    # arcpy.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")
    
    # Set the Geoprocessing environment...
    arcpy.OverwriteOutput = 1
    # arcpy.cellSize = "MINOF"
    # arcpy.workspace = INPUT_LOCATION
    
    
    # CPC Naming Standard: {variable}_Path is the fully qualified path, otherwise it is just the name of the item.
  
    # Processing Variables
   
    # Processing_Variables['OperationalFGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Operational_Data.gdb'
    Processing_Variables['OperationalFGDB_Path']   = os.path.split(INPUT_MLT)[0]
    

    Processing_Variables['Scratch_Path']           = Processing_Variables['Project_Base_Path'] + r'\Scratch'
    Processing_Variables['ScratchFGDB_Path']           = Processing_Variables['Scratch_Path'] + r'\Scratch.gdb'
    arcpy.scratchworkspace = Processing_Variables['ScratchFGDB_Path']
    
    Processing_Variables['Tools_and_Templates_FGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Tools_and_Templates.gdb'
    
    Processing_Variables['Soils_Base_Path']    = INPUT_SOIL
    Processing_Variables['Soils_Data_Path']    = Processing_Variables['Soils_Base_Path'] + r'\Operational_Data.gdb'
    Processing_Variables['TEIS_Master_Long_Tbl']    = INPUT_MLT
    
    Processing_Variables['TEIS_Project_Details']   = INPUT_PD

    Processing_Variables['TEIS_Project_Boundaries']   = INPUT_MLT.replace('Master_Long_Tbl','Master_Project_Boundaries')
    if Processing_Variables['TEIS_Project_Boundaries'] == Processing_Variables['TEIS_Master_Long_Tbl']:
        Processing_Variables['TEIS_Project_Boundaries']    = INPUT_MLT + "_Boundaries"


    Processing_Variables['TEIS_Project_Boundaries_Temp1'] = Processing_Variables['ScratchFGDB_Path'] + r'\CreateProjectBoundaries_Temp1'
    Processing_Variables['TEIS_Project_Boundaries_Temp2'] = Processing_Variables['ScratchFGDB_Path'] + r'\CreateProjectBoundaries_Temp2'
    Processing_Variables['TEIS_Project_Boundaries_Temp3'] = Processing_Variables['ScratchFGDB_Path'] + r'\CreateProjectBoundaries_Temp3'

    
    Processing_Variables['Template_Empty_Project_Boundary']   = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\Template_FGDB_TEIS_Master_Project_Boundaries'
    
    # List of fields taken from the Project Details and Project Tracker tables and added to the boundaries based on BAPID.
    #### May want to get this from the Metadata table eventually?
    Processing_Variables['ProjectBoundaryFields']   = "FCODE;PROJ_TYPE;PROJ_NAME;PROJ_SCALE;MOD_DATE;GEOG_LOC;COMPL_DATE;SURV_DATE;ESIL;TSIL;MAPSH_LST;ORG_NAME;PROJ_SUP;ECO_MAP;TER_MAP;SOIL_MAP;WILD_MAP;DIG_CAP;GIS_SUP;REC_NAME;CLIENT;TRIM_NBR;PHO_TYPE;PHO_SC;PHO_YR;TER_LEG_SC;TER_LEG_TP;PACK_NBR;STBCLS_TP;SLP_UNIT;PROJ_COM;PROJ_ID"

    Processing_Variables['Dissolve_Script'] = Processing_Variables['Script_Directory_Path'] + r'\Dissolve.py'
    
    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# CPMessage(Message)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CPMessage(Message, MessageLevel):
    # Based on the logging level, add messages to the message queues.
    #   Text Run Log   : Implimented
    #   Console        : Implimented
    #   GeoProcessor   : Implimented
    #   Processing Log : Implimented (with e-mail)
    #   FGDB           : Implimented
    
    # Message Levels:
    #   0    : Fatal Errors
    #   1    : Top Level Program
    #   2    : Main routines and events
    #   3    : Warnings
    #   4    : General Information
    #   5    : Debug level detail
    
    # Send to the screen
    if MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToScreen(Message, MessageLevel)    
    
    # Send to the Processing Log
    if 'SendRunLogToEmail' in Processing_Variables:
        Processing_Variables['PROCESS_LOG'].append((Message, MessageLevel))
    
    # Send to the Log File
    if 'MessageQueueFile' in Processing_Variables and MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToFile(Message, Processing_Variables['MessageQueueFile'])

    # Send to the GeoProcessor Message Queue
    #### At ArcGIS 9.3 SP1 there is a bug that also causes any GeoProcessor messages to be echoed to the console.
    #    Added the ToolBoxRun parameter to allow the script to add messages to the GeoProcessor Messaging properly
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and Processing_Variables['ToolBoxRun'] == 'TRUE':
        if MessageLevel == 0:
            arcpy.AddError(Message)
        if MessageLevel == 1 or MessageLevel == 2:
            arcpy.AddMessage(Message)
        if MessageLevel == 3:
            arcpy.AddWarning(Message)

    # Send to the FGDB Processing Log
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and 'Log_FGDB_Table' in Processing_Variables:
        rows = arcpy.InsertCursor(Processing_Variables['Log_FGDB_Table'])

        row = rows.NewRow()
        row.SetValue("Computer_Name", COMPUTERNAME)
        row.SetValue("User_Name", USERNAME)
        row.SetValue("Script_Name", PYTHON_SCRIPT)
        row.SetValue("Event_Time", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
        row.SetValue("Message_Level", MessageLevel)
        row.SetValue("Message", Message)
        row.SetValue("Start_Time", START_TIME_SQL)
        
        rows.InsertRow(row)
        del row
        del rows
    

    return(1)


#---------------------------------------------------------------------------------------------------------
# WriteOutputToScreen(Output_Comment, Header_Style):
#
#   Prints the Output_Comment to the screen with the appropriate formatting
#       0 : "******" Used for Error Sections
#       1 : "======" Used for the Program's Header and Footer
#       2 : "------" Used for New Sections
#       3+:          No delimitation
#   
#---------------------------------------------------------------------------------------------------------
def WriteOutputToScreen(Output_Comment, Header_Style):
    if Header_Style == 0:
        # Errors get beeps, too!
        print BEEP * 3 + "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 2:
        print "\n" + "-" * 79 + "\n"
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"
        
    # print Output_Comment
    arcpy.AddMessage(Output_Comment)
    
    if Header_Style == 0:
        print "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"


#---------------------------------------------------------------------------------------------------------
# WriteOutputToFile(Output_Comment, Run_Log)
#   Record to Log 
#
#
#---------------------------------------------------------------------------------------------------------
def WriteOutputToFile(Output_Comment, Run_Log):
# Run logs are not compatible with concurrency - only one process is expected in debug mode at a time!
# First try to append to an existing file - if that fails create a new file.
    ReturnValue = 1
    try:
        if os.path.isfile(Run_Log) == 1:
            # try to open up the fileName 
            file = open(Run_Log,'a') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Appended comment to the file"
        else:
            # try to open up the fileName 
            file = open(Run_Log,'w') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created new file and put in the comment"
    except:
        # All failures are unexpected in this case...  
        # Maybe an invalid drive name, read only file or something?
        Processing_Variables['PROCESS_STATUS'] = -1
        Processing_Variables['PROCESS_INFO']   = "**ERROR: Can't write to file: " + Run_Log   
        print BEEP + BEEP + BEEP
        ReturnValue = -1
    return(ReturnValue)
        

#---------------------------------------------------------------------------------------------------------
# SetupWorkspace()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SetupWorkspace():
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "SetupWorkspace"  

    CPMessage('    - Setting Up Workspace', 2)
    
    if not arcpy.Exists(INPUT_MLT):
        CPMessage(' **** Specified Master Long Table feature class does not exist: ' + INPUT_MLT, 0)
        sys.exit(100)

    if not arcpy.Exists(INPUT_PD):
        CPMessage(' **** Specified project details table does not exist: ' + INPUT_PD, 0)
        sys.exit(100)

    if not arcpy.Exists(Processing_Variables['ScratchFGDB_Path']):
        CPMessage('        -- Creating Scratch FGDB', 4)
        try:
            arcpy.CreateFileGDB_management(os.path.split(Processing_Variables['ScratchFGDB_Path'])[0], os.path.split(Processing_Variables['ScratchFGDB_Path'])[1])
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created Scratch FGDB"
        except:
            Processing_Variables['PROCESS_STATUS'] = -1
            Processing_Variables['PROCESS_INFO']   = "Error Creating Scratch FGDB"
            CPMessage('** Can not find or create ' + Processing_Variables['ScratchFGDB_Path'], 0)
            sys.exit(100)
    else:
        CPMessage('        -- Scratch FGDB exists, nothing to do.', 4)
       
    if arcpy.Exists(Processing_Variables['OperationalFGDB_Path']):
        CPMessage('        -- Operational FGDB exists.', 4)
    else:
        CPMessage(' **** Operational FGDB does not exist: ' + Processing_Variables['OperationalFGDB_Path'], 0)
        sys.exit(100)

    if arcpy.Exists(Processing_Variables['Tools_and_Templates_FGDB_Path']):
        CPMessage('        -- Tools and Templates FGDB exists.', 4)
    else:
        CPMessage(' **** Tools and Templates FGDB does not exist: ' + Processing_Variables['Tools_and_Templates_FGDB_Path'], 0)
        sys.exit(100)

    if arcpy.Exists(Processing_Variables['TEIS_Master_Long_Tbl']):
        CPMessage('        -- Master Long Table exists.', 4)
    else:
        CPMessage(' **** Master Long Table does not exist: ' + Processing_Variables['TEIS_Master_Long_Tbl'], 0)
        sys.exit(100)

    if arcpy.Exists(Processing_Variables['TEIS_Project_Details']):
        CPMessage('        -- Project Details table exists.', 4)
    else:
        CPMessage(' **** Project Details table does not exist: ' + Processing_Variables['TEIS_Project_Details'], 0)
        sys.exit(100)

    if os.path.exists(Processing_Variables['Soils_Base_Path']):
        CPMessage('        -- Soils_master folder exists.', 4)
    else:
        CPMessage('        -- No Soils_master folder found. No soils data will be incorporated into the output.', 4)

    for i in [1,2,3]:
        if arcpy.Exists(Processing_Variables['TEIS_Project_Boundaries_Temp' + stra(i)]):
            CPMessage('        - Deleting existing temporary boundaries feature class ' + Processing_Variables['TEIS_Project_Boundaries_Temp' + stra(i)], 4)
            while not arcpy.TestSchemaLock(Processing_Variables['TEIS_Project_Boundaries_Temp' + stra(i)]):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['TEIS_Project_Boundaries_Temp' + stra(i)], 3)
                time.sleep(10)            
            arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp' + stra(i)])



    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# SendRunLogToEmail()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SendRunLogToEmail(mailto):
    import smtplib
    CPMessage('  * Sending log to ' + mailto, 2)
    mailfrom = 'GIS_Python@cloverpoint.com'
    # mailto = 'mneal@cloverpoint.com'
    mailsubject = 'Run Log: ' + os.path.split(PYTHON_SCRIPT)[1] + ' (' + COMPUTERNAME + '\\' + USERNAME + ')'
    mailbody = "This is an automatic run log.\n\n"
    
    for LogEntry in Processing_Variables['PROCESS_LOG']:
        mailbody = mailbody + '\n' + LogEntry[0]

    mailmsg = ("From: " + mailfrom + "\r\nTo: " + mailto + "\r\nSubject: " + mailsubject + "\r\n" + mailbody)

    mailserver = smtplib.SMTP('shawmail.gv.shawcable.net')

    # SMTP.login(user, password)

    mailserver.sendmail(mailfrom, mailto, mailmsg)
    mailserver.quit()
    return(Processing_Variables['PROCESS_STATUS'])



#=========================================================================================================
#  Special Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# SanitizeElapsedTime(number_of_seconds)
#
# routine to convert a raw number of seconds into a human-friendly statement of duration
#---------------------------------------------------------------------------------------------------------

def SanitizeElapsedTime(dtInput):

    if dtInput < 120:
        strElapsedTime = stra(int(dtInput)) + ' seconds.'
        
    elif dtInput < 3600:
        strElapsedTime = stra(int(dtInput)) + ' seconds (' + stra(round(dtInput / 60, 1)) + ' minutes).'
        
    else:
        strElapsedTime = stra(int(dtInput)) + ' seconds (' + stra(round(dtInput / 3600, 2)) + ' hours).'
        
    return strElapsedTime


#---------------------------------------------------------------------------------------------------------
# CreateProjectBoundaries()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CreateProjectBoundaries():
    # Creates a dissolved version of all boundaries that match the list of BAPIDs.  The only item will be BAPID.
    
    CPMessage('    - Creating Project Boundaries', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "CreateProjectBoundaries"     

    ProjDetails_Dictionary = {}
    Number_Processed = 0
    
    CPMessage('        - Reading Project Details table', 4)
    
    rows = arcpy.SearchCursor(Processing_Variables['TEIS_Project_Details'])
    row = rows.next()
    while row:
        BAPID = row.getValue('BAPID')
        PROJ_TYPE = stra(row.getValue('PROJ_TYPE')).rstrip(' ').replace('None','')
        ProjDetails_Dictionary[BAPID] = {'PROJ_TYPE': PROJ_TYPE}

        Number_Processed = Number_Processed + 1
        row = rows.next()
    CPMessage('            - ' + stra(len(ProjDetails_Dictionary.keys())) + ' BAPIDs found in Project Details table', 4)

    CPMessage('        - Determining number of slices to process and checking BAPID details', 4)

    global d
    if len(d['Version'].split('.')) > 2:
        dv = d['Version'].rsplit('.',len(d['Version'].split('.'))-2)[0]


        Object_ID_Field = arcpy.ListFields(Processing_Variables['TEIS_Master_Long_Tbl'], "*", "OID")[0].name
        result = arcpy.GetCount_management(Processing_Variables['TEIS_Master_Long_Tbl'])
        Number_of_Features = int(result.getOutput(0))
        bapid_dict = {}
        bapid_not_found_dict = {}
        rows = arcpy.SearchCursor(Processing_Variables['TEIS_Master_Long_Tbl'])
        row = rows.next()

        Number_Processed = 0
        Slice_Marker_List = []
        while row:
            Number_Processed = Number_Processed + 1
            if Number_Processed % 500000 == 0:
                current_oid = row.getValue(Object_ID_Field)
                Slice_Marker_List.append(current_oid)

            BAPID = row.getValue('BAPID')
            try:
                ProjDetails_Dictionary[BAPID]
            except:
                try:
                    bapid_not_found_dict[BAPID]
                except:
                    bapid_not_found_dict[BAPID] = True
            try:
                bapid_dict[BAPID]
            except:
                bapid_dict[BAPID] = True

            if Number_Processed % 1000000 == 0:
                CPMessage('            - Read ' + stra(Number_Processed) + ' of ' + stra(Number_of_Features) + ' records (' + stra(len(bapid_dict.keys())) + ' unique BAPID(s) found)', 4)
            row = rows.next()

        if len(bapid_not_found_dict.keys()) > 0:
            bapids_not_found_string = ''
            bapid_not_found_list = bapid_not_found_dict.keys()
            bapid_not_found_list.sort()
            for bapid in bapid_not_found_list:
                bapids_not_found_string += stra(bapid) + ', '
            bapids_not_found_string = bapids_not_found_string[:-2]
            CPMessage('    ***** BAPIDs not found in BAPID Tracker/Project Details table: ' + bapids_not_found_string, 0)
            CPMessage('    ***** Please add these BAPIDs to the BAPID Tracker (or TEI_Project_Details table) before running this script again.', 0)
            sys.exit()

        CPMessage('            - Read all ' + stra(Number_Processed) + ' records. All ' + stra(len(bapid_dict.keys())) + ' unique BAPIDs found in BAPID Tracker/Project Details table.', 4)

        if Number_Processed % 500000 == 0:
            Slice_Marker_List.remove(current_oid)
        Slice_Marker_List.append(999999999)
        del row
        del rows

    elif float(d['Version']) >= 10.1:

        Object_ID_Field = arcpy.ListFields(Processing_Variables['TEIS_Master_Long_Tbl'], "*", "OID")[0].name
        Number_of_Features = int(arcpy.GetCount_management(Processing_Variables['TEIS_Master_Long_Tbl']).getOutput(0))
        Number_Processed = 0
        Slice_Marker_List = []
        bapid_dict = {}
        bapid_not_found_dict = {}
        exec_string = '''with arcpy.da.SearchCursor(Processing_Variables['TEIS_Master_Long_Tbl'], [Object_ID_Field,'BAPID']) as cursor:
    for row in cursor:
        Number_Processed = Number_Processed + 1
        if Number_Processed % 500000 == 0:
            Slice_Marker_List.append(row[0])

        try:
            ProjDetails_Dictionary[row[1]]
        except:
            try:
                bapid_not_found_dict[row[1]]
            except:
                bapid_not_found_dict[row[1]] = True
        try:
            bapid_dict[row[1]]
        except:
            bapid_dict[row[1]] = True

        if Number_Processed % 1000000 == 0:
            CPMessage('            - Read ' + stra(Number_Processed) + ' of ' + stra(Number_of_Features) + ' records (' + stra(len(bapid_dict.keys())) + ' unique BAPID(s) found)', 4)

    if len(bapid_not_found_dict.keys()) > 0:
        bapids_not_found_string = ''
        bapid_not_found_list = bapid_not_found_dict.keys()
        bapid_not_found_list.sort()
        for bapid in bapid_not_found_list:
            bapids_not_found_string += stra(bapid) + ', '
        bapids_not_found_string = bapids_not_found_string[:-2]
        CPMessage('    ***** BAPIDs not found in BAPID Tracker/Project Details table: ' + bapids_not_found_string, 0)
        CPMessage('    ***** Please add these BAPIDs to the BAPID Tracker (or TEI_Project_Details table) before running this script again.', 0)
        sys.exit()

    CPMessage('            - Read all ' + stra(Number_Processed) + ' records. All ' + stra(len(bapid_dict.keys())) + ' unique BAPIDs found in BAPID Tracker/Project Details table.', 4)
    if Number_Processed % 500000 == 0:
        Slice_Marker_List.remove(row[0])
    Slice_Marker_List.append(999999999)
    del row
    del cursor'''
        exec(exec_string)
        
    else:
        CPMessage('** Unsupported ArcGIS version (' + stra(d['Version']) + ') running on this machine.',0)
        sys.exit()
    
    # First do the TEIS_Master_Long_Tbl features.
    CPMessage('        - Creating TEIS Boundaries', 4)
    
    if len(Slice_Marker_List) == 0:
        # Process the entire feature class

        CPMessage('            - Dissolving', 4)
        arcpy.Dissolve_management(Processing_Variables['TEIS_Master_Long_Tbl'], Processing_Variables['TEIS_Project_Boundaries_Temp1'], 'BAPID', '#', 'SINGLE_PART')

    else:
        Previous_Marker = 0
        Slice_Counter = 0
        dtCalcStart = time.time()
        for Current_Marker in Slice_Marker_List:
            Slice_Counter += 1
            where_clause = '"OBJECTID" > ' + stra(Previous_Marker) + ' AND "OBJECTID" <= ' + stra(Current_Marker)
            CPMessage('            - Exporting temporary slice ' + stra(Slice_Counter) + ' of ' + stra(len(Slice_Marker_List)), 4)
            arcpy.FeatureClassToFeatureClass_conversion(Processing_Variables['TEIS_Master_Long_Tbl'],os.path.split(Processing_Variables['TEIS_Project_Boundaries_Temp3'])[0],os.path.split(Processing_Variables['TEIS_Project_Boundaries_Temp3'])[1],where_clause)
            CPMessage('            - Dissolving (slice ' + stra(Slice_Counter) + ' of ' + stra(len(Slice_Marker_List)) + ')', 4)
            if Slice_Counter == 1:
                if os.path.exists(Processing_Variables['Dissolve_Script']):
                    os.system(Processing_Variables['Dissolve_Script'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp3'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp1'] + " BAPID SINGLE_PART")
                else:
                    arcpy.Dissolve_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'], Processing_Variables['TEIS_Project_Boundaries_Temp1'], 'BAPID', '#', 'SINGLE_PART')
                arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'])

            else:
                if os.path.exists(Processing_Variables['Dissolve_Script']):
                    os.system(Processing_Variables['Dissolve_Script'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp3'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp2'] + " BAPID SINGLE_PART")
                else:
                    arcpy.Dissolve_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'], Processing_Variables['TEIS_Project_Boundaries_Temp2'], 'BAPID', '#', 'SINGLE_PART')
                arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'])
                CPMessage('            - Appending', 4)
                arcpy.Append_management(Processing_Variables['TEIS_Project_Boundaries_Temp2'], Processing_Variables['TEIS_Project_Boundaries_Temp1'], 'TEST')
                arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp2'])

            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Slice_Counter) * (len(Slice_Marker_List) - Slice_Counter)
            CPMessage('            - Processed ' + stra(Slice_Counter) + ' of ' + stra(len(Slice_Marker_List)) + ' slices (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining to process all slices)', 4)
            Previous_Marker = Current_Marker
            
    # Then create and add boundaries for the the Soils data, if it exists
    if os.path.exists(Processing_Variables['Soils_Data_Path']):
        # for soil_layer in ['Soil_Short_Tbl','BCSLCv2r2','BCSLCv3r2','CANSIS100K','CAPAMP','CAPAMP2','CAPAMP2AG','Seamless','TEIS_Soils_of_BC','ReportsOnly']:
        for soil_layer in ['SOILS_Project_Boundaries']:
            current_soil_fc = Processing_Variables['Soils_Data_Path'] + '\\' + soil_layer
            if arcpy.Exists(current_soil_fc):
                # Dissolve the entire feature class
                CPMessage('        - Creating Soils Boundaries for ' + soil_layer, 4)

                CPMessage('            - Dissolving ' + soil_layer, 4)
                if os.path.exists(Processing_Variables['Dissolve_Script']):
                    os.system(Processing_Variables['Dissolve_Script'] + " " + current_soil_fc + " " + Processing_Variables['TEIS_Project_Boundaries_Temp2'] + " BAPID SINGLE_PART")
                else:
                    arcpy.Dissolve_management(current_soil_fc,Processing_Variables['TEIS_Project_Boundaries_Temp2'],'BAPID','#','SINGLE_PART')
                
                CPMessage('            - Appending to main Project Boundaries', 4)
                arcpy.Append_management(Processing_Variables['TEIS_Project_Boundaries_Temp2'], Processing_Variables['TEIS_Project_Boundaries_Temp1'], 'TEST')
                arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp2'])
            else:
                CPMessage('** Expected Soils dataset does not exist: ' + current_soil_fc, 3)

    CPMessage('        - Second dissolve on BAPID', 3)
    if os.path.exists(Processing_Variables['Dissolve_Script']):
        os.system(Processing_Variables['Dissolve_Script'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp1'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp2'] + " BAPID MULTI_PART")
    else:
        arcpy.Dissolve_management(Processing_Variables['TEIS_Project_Boundaries_Temp1'], Processing_Variables['TEIS_Project_Boundaries_Temp2'], 'BAPID', '#', 'MULTI_PART')
    arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp1'])

    CPMessage('        - Third dissolve on BAPID', 3)
    if os.path.exists(Processing_Variables['Dissolve_Script']):
        os.system(Processing_Variables['Dissolve_Script'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp2'] + " " + Processing_Variables['TEIS_Project_Boundaries_Temp3'] + " BAPID MULTI_PART")
    else:
        arcpy.Dissolve_management(Processing_Variables['TEIS_Project_Boundaries_Temp2'], Processing_Variables['TEIS_Project_Boundaries_Temp3'], 'BAPID', '#', 'MULTI_PART')
    arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp2'])

    CPMessage('        - Adding attributes from Project Details table', 3)

    arcpy.JoinField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'], 'BAPID', Processing_Variables['TEIS_Project_Details'], 'BAPID')
    field_list = [f for f in arcpy.ListFields(Processing_Variables['TEIS_Project_Boundaries_Temp3'])]
    for field in field_list:
        if field.name == "AA_COM" and field.length > 254:
            arcpy.CalculateField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'],"AA_COM","unicode(!AA_COM!).encode('ascii','replace')[:254].replace('None','')","PYTHON")
        elif field.name == "BDRY_ONLY" and field.type == "SmallInteger":
            arcpy.AddField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'],"BDRY_ONLY1","TEXT","#","#","1")
            code_blk = '''def bdry_yn(bdry_only_num):
    if bdry_only_num == 0:
        return("N")
    else:
        return("Y")'''
            arcpy.CalculateField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'],'BDRY_ONLY1','bdry_yn(!BDRY_ONLY!)','PYTHON',code_blk)
            arcpy.DeleteField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'],'BDRY_ONLY')
            arcpy.AddField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'],"BDRY_ONLY","TEXT","#","#","1")
            arcpy.CalculateField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'],'BDRY_ONLY','!BDRY_ONLY1!','PYTHON')
            arcpy.DeleteField_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'],'BDRY_ONLY1')

    if arcpy.Exists(Processing_Variables['TEIS_Project_Boundaries']):
        CPMessage('        - Deleting existing output feature class: ' + Processing_Variables['TEIS_Project_Boundaries'], 4)
        while not arcpy.TestSchemaLock(Processing_Variables['TEIS_Project_Boundaries']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['TEIS_Project_Boundaries'], 3)
            time.sleep(10)            
        arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries'])    

    CPMessage('        - Copying Project Boundaries template', 4)
    arcpy.Copy_management(Processing_Variables['Template_Empty_Project_Boundary'], Processing_Variables['TEIS_Project_Boundaries'])
    
    CPMessage('        - Appending records to final output: ' + Processing_Variables['TEIS_Project_Boundaries'], 4)
    arcpy.Append_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'], Processing_Variables['TEIS_Project_Boundaries'], 'NO_TEST')
    arcpy.Delete_management(Processing_Variables['TEIS_Project_Boundaries_Temp3'])

    CPMessage('        - Deleting temporary datasets', 4)

    delete_list = [Processing_Variables['TEIS_Project_Boundaries_Temp1'],Processing_Variables['TEIS_Project_Boundaries_Temp2'],Processing_Variables['TEIS_Project_Boundaries_Temp3']]
    for fc_to_delete in delete_list:
        if arcpy.Exists(fc_to_delete):
            while not arcpy.TestSchemaLock(fc_to_delete):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + fc_to_delete, 3)
                time.sleep(10)            
            arcpy.Delete_management(fc_to_delete)

    return(Processing_Variables['PROCESS_STATUS'])
    

#*********************************************************************************************************
#  Main Program
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# Run Main
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain():

    Initialize()
    SetupWorkspace()

    CreateProjectBoundaries()

    return Processing_Variables['PROCESS_STATUS']

#*********************************************************************************************************
#  Execution
#*********************************************************************************************************


runMain()



#---------------------------------------------------------------------------------------------------------
#  Clean-up
#---------------------------------------------------------------------------------------------------------


END_TIME       = time.ctime(time.time())
END_TIME_SEC   = time.time()
END_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
EXECUTION_TIME = END_TIME_SEC - START_TIME_SEC
if EXECUTION_TIME < 120:
    EXECUTION_TIME_STRING = stra(int(EXECUTION_TIME)) + ' seconds.'
elif EXECUTION_TIME < 3600:
    EXECUTION_TIME_STRING = stra(round(EXECUTION_TIME / 60, 1)) + ' minutes.'
else:
    EXECUTION_TIME_STRING = stra(round(EXECUTION_TIME / 3600, 2)) + ' hours.'


CPMessage('    End Time: ' + END_TIME, 2)
CPMessage('    Execution Time: ' + EXECUTION_TIME_STRING, 4)
CPMessage('Complete: ' + PYTHON_SCRIPT, 1)
CPMessage('Next Script: CreateShortTable',1) 
CPMessage('Preconditions for next step: ', 1)
# arcpy.CheckInExtension("spatial")
#del gp

# Report Program Execution Status back to the calling program


if 'SendRunLogToEmail' in Processing_Variables:
    for Email in Processing_Variables['SendRunLogToEmail']:
        SendRunLogToEmail(Email)

# sys.exit(Processing_Variables['PROCESS_STATUS'])

#*********************************************************************************************************
#  End of Program
#*********************************************************************************************************
