#=========================================================================================================
#                                        CPC GeoData Checker
#=========================================================================================================
#
#
# Author(s)         :  Clover Point Cartographics (Mike Neal)
#
# Date              :  2010-06-11
#
# Purpose           :  This script is a stand alone data checker. It provides a summary of the locations
#                      it checks as well as listing the number of files, vector datasets, raster datasets,
#                      and databases. It checks datasets for geometry and projection. It also lists the
#                      attributes of datasets, including attribute parameters such as Alias, Type, Length,
#                      Scale, and Precision.
#
# Arguments         :  argv(1) = INPUT_LOCATION
#
# Example Input     :  X:\fullpath\GeoData_Checker.py X:\fullpath\Folder
#
# Outputs           :  ExtractionReport.html in a local temporary folder 
#
# Dependencies      :  None
#
# Expected Duration :  < 2 hours
#
# History           :  None
#
#=========================================================================================================
#=========================================================================================================
#  Process Status and Return Codes:
#      -1   - Unexpected error
#       0   - Completed Successful Run
#       1   - In progress
#       100 - Invalid Arguments
#       101 - ESRI License Error
#       999 - License Expired
#
#=========================================================================================================

# import py_compile
# py_compile.compile(r'P:\Projects\08039_GIS_Data_Checker\Working\GeoData_Checker.py')

#---------------------------------------------------------------------------------------------------------
#  Python and System Modules
#---------------------------------------------------------------------------------------------------------

from __future__ import division
import sys
import string
import os
import time
import shutil
# import re
# import dircache
# import win32com.client

# import cpc_gis_common

#*********************************************************************************************************
#  Program Initialization
#*********************************************************************************************************

PROGRAM_VERSION = '0.9.2b'
LICENSE = 'Public Beta'




#---------------------------------------------------------------------------------------------------------
#  Arguement Validation
#---------------------------------------------------------------------------------------------------------

# Capture/Validate Parameters
#  Test that the right number of paramters were provided and validated
#  Note that 1 must be added because the Python script is an arguement (0) supplied to the
#  Python Environment.


# If long text parameters are required, use:  any_value = gp.GetParameterAsText(0)

if len(sys.argv) < 2 or len(sys.argv) > 2 :

    print "\nInvalid number of arguments (" + str(len(sys.argv) - 1) + " supplied) to:"
    print "    " + sys.argv[0] + "\n"
    print " Usage:   " + sys.argv[0] + " [ARG1] "
    print "    [ARG1]   : Path to input directory"
    # print "    [ARG2]   : Path to output directory"
    # print "    {ARG3}   : No options"
    sys.exit(100)

PYTHON_SCRIPT   = sys.argv[0]
INPUT_LOCATION  = sys.argv[1]

#---------------------------------------------------------------------------------------------------------
#  Standard/Global Variables/Constants
#---------------------------------------------------------------------------------------------------------

# Global Constants
COMPUTERNAME     = os.environ['COMPUTERNAME']
USERNAME         = os.environ['USERNAME']
CPC_LOCAL_TEMP_SPACE = os.environ['TEMP'] + '\\CP_GeoData_Checker'
START_TIME       = time.ctime(time.time())
START_TIME_SEC   = time.time()
START_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

# Global Variables
#   Note that to assign new values to these inside a function it is necessary to use the global statement
#   within the function itself.
PROCESS_INFO     = "RUNNING " + PYTHON_SCRIPT
PROCESS_STATUS   = 1

#---------------------------------------------------------------------------------------------------------
#  License Validation
#---------------------------------------------------------------------------------------------------------
EXPIRY_DATE = time.mktime([2010,04,01,0,0,0,5,0,-1])

print '\n'
print ' Clover Point GeoData Checker\n'
print ' Version:',PROGRAM_VERSION
print '\n  ** FREE PUBLIC BETA **\n'
print ' For more information please contact:'
print '    Web    : www.cloverpoint.com'
print '    E-mail : mike.n@cloverpoint.com'
print '    Phone  : 1-888-457-2724'


# if START_TIME_SEC > EXPIRY_DATE:
#     print '\n This version of the GeoData Checker is expired.'
#    print ' Please contact Clover Point for a new version. \n'
#    sys.exit(999)
# else:
#    print ' This version is licensed until ' + time.ctime(EXPIRY_DATE)
#    print '\n'


#---------------------------------------------------------------------------------------------------------
#  CPC Variables/Constants
#---------------------------------------------------------------------------------------------------------

# CPC_SCRIPT_WRITE_OUTPUT  = "P:\\GISResource\\Enterprise\\Common\\CPC_Common_WriteOutput.py"
# CPC_QA_LOG_FILE          = "P:\\GISResource\\Enterprise\\Log\\QA_" + COMPUTERNAME + ".log"

# WRITE_OUTPUT = os.system(CPC_SCRIPT_WRITE_OUTPUT + ' "Executing ' + PYTHON_SCRIPT + '" 1 1 1 ' + CPC_QA_LOG_FILE)
# PKZIP = 'P:\\GISResource\\Enterprise\\Bin\\pkzipc.exe'

#---------------------------------------------------------------------------------------------------------
#  Set up the ArcGIS Geoprocessing Environment
#---------------------------------------------------------------------------------------------------------

# WRITE_OUTPUT = os.system(CPC_SCRIPT_WRITE_OUTPUT + ' "  Testing Python/ArcGIS version" 0 0 1 ' + CPC_QA_LOG_FILE)
# WRITE_OUTPUT = os.system(CPC_SCRIPT_WRITE_OUTPUT + ' "    ' + sys.version + '" 0 0 1 ' + CPC_QA_LOG_FILE)


# For ArcGIS 9.2:
try:
    import arcgisscripting
    gp = arcgisscripting.create()

# Get any required extensions
# if gp.CheckExtension("spatial") == "available":
#     gp.CheckOutExtension("spatial")
# else:
#     sys.exit(101)

except:
    # WRITE_OUTPUT = os.system(CPC_SCRIPT_WRITE_OUTPUT + ' "    ArcGIS Failed" 3 1 1 ' + CPC_QA_LOG_FILE)
    print " **** ArcGIS Initialization Failed"
    sys.exit(101)

# Load required toolboxes...
# gp.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")

# Set the Geoprocessing environment...
gp.OverwriteOutput = 1
gp.cellSize = "MINOF"
# gp.workspace = INPUT_LOCATION
gp.scratchworkspace = CPC_LOCAL_TEMP_SPACE

#---------------------------------------------------------------------------------------------------------
#  Script Specific Variables/Constants
#---------------------------------------------------------------------------------------------------------


#*********************************************************************************************************
#  Routines
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# TemplateRoutine()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TemplateRoutine():

    return(0)

#=========================================================================================================
#  Standard Routines
#=========================================================================================================

#=========================================================================================================
#  Special Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# GetProcessingList(INPUT_LOCATION)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return list of files to be processed (no path names).
#
#---------------------------------------------------------------------------------------------------------

# def GetProcessingList(INPUT_LOCATION):
#     # Get a list of all zip files in the INPUT_LOCATION
#     All_Files = os.listdir(INPUT_LOCATION)
#     Processing_List = []
#     for Current_File in All_Files:
#         if Current_File.endswith('.zip'):
#             Processing_List.append(Current_File)
#
#     #### Debug:
#     Processing_List = ['CPC_20080611.zip','ICIS_Test.zip']
#     Processing_List = ['Basic_Test.zip']
#
#
#     return(Processing_List)

def CreateStagingArea(DataPackage, PackageMetadata):
    StagingAreaPath = CPC_LOCAL_TEMP_SPACE
    if os.path.isdir(StagingAreaPath) == 1:
        print '  ** WARNING: Staging Path already existed (removing):', StagingAreaPath
        shutil.rmtree(StagingAreaPath)
        PackageMetadata['ProcessingWarnings'].append('Staging area existed and was removed before processing.')

    # Create the staging area, and copy the zip file there.
    os.makedirs(StagingAreaPath)
    os.makedirs(StagingAreaPath + '\\Extracted')
    os.makedirs(StagingAreaPath + '\\Temp')
    # shutil.copyfile(INPUT_LOCATION + '\\' + DataPackage, StagingAreaPath + '\\' + DataPackage)
    PackageMetadata['CheckGeometryTable'] = StagingAreaPath + '\\GeomChk.dbf'
    return(StagingAreaPath)

def ExtractDataPackage(DataPackage, StagingArea):
    # Make a copy for performance and for safety
    print '    - Copying files to temporary space for analysis'
    os.chdir(StagingArea + '\\Extracted')
    os.system('xcopy ' + DataPackage + ' ' + StagingArea + '\\Extracted' + ' /e /q /z ')
    print '        - Copying complete'
    return(0)

def GenerateScoreCard(PackageMetadata):
    ScoreCard = []
    for Locations in PackageMetadata['Locations']:
        for DataSet in Locations['LocationVectorDataSets']:

            # Score Projection
            if DataSet['Projection'] == 'Unknown':
                ScoreCard.append((DataSet['Name'],'Projection','Fail'))
            else:
                ScoreCard.append((DataSet['Name'],'Projection','Pass'))

            # Score Geometry
            if DataSet['NumberOfFeatures'] > 0:
                # ScoreGeometry = str(int((int(DataSet['NumberOfFeatures']) - len(DataSet['GeometryErrors'])) / (int(DataSet['NumberOfFeatures'])) * 10))
                # ScoreGeometry = str(int(DataSet['NumberOfFeatures']) - len(DataSet['GeometryErrors'])) + ' of ' + str(DataSet['NumberOfFeatures'])
                Good = int(DataSet['NumberOfFeatures']) - len(DataSet['GeometryErrors'])
                Total = int(DataSet['NumberOfFeatures'])
                Percent = round(Good/Total,3) * 100
                ScoreGeometry = str(Percent) + r'% (' + str(Good) + ' of ' + str(Total) + ')'
            else:
                # No features in data set
                ScoreGeometry = 'N/A'
            ScoreCard.append((DataSet['Name'],'Geometry',ScoreGeometry))



    # ScoreCard = [('Projection','10.0'),('Geometry','7.3'),('Total','8.6')]
    return(ScoreCard)

def GenerateSummaryText(ValidationReport):
    SummaryText = 'Placehoder 0'
    return(SummaryText)

def DetailedValidation(InputLocation, LocationReport, PackageMetadata):
    print '  Validating:',InputLocation

    #### Debug
    # InputLocation = InputLocation + '\\DoesNotExist'



    # Initialize an empty location:
    ReportLocation = str(InputLocation)
    LocationSummary = ''
    LocationStatisticLocations = 0
    LocationStatisticFiles = 0
    LocationStatisticVectors = 0
    LocationStatisticRasters = 0
    LocationStatisticDatabases = 0
    LocationSpatialData = []
    LocationNonSpatialData = []
    LocationSpecialFiles = []
    LocationDirectories = []
    LocationFiles = []
    LocationVectorDataSets = []
    LocationRasterDataSets = []
    LocationDatabaseDataSets = []
    SubLocations = []

    if not gp.exists(InputLocation):
        gp.AddError("Workspace does not exist")
        # WRITE_OUTPUT = os.system(CPC_SCRIPT_WRITE_OUTPUT + ' "****INPUT_LOCATION DOES NOT EXIST" 3 1 1 ' + CPC_QA_LOG_FILE)
        PROCESS_STATUS = 98
        PROCESS_INFO   = "****INPUT_LOCATION DOES NOT EXIST: " + ReportLocation

        LocationSummary = 'This location could not be read or does not exist.'
        LocationSpatialData.append(('Not Available','',''))
        LocationNonSpatialData.append(('Not Available','',''))
        LocationSpecialFiles.append(('Not Available','',''))
        LocationDirectories.append(('Not Available',''))
        LocationFiles.append(('Not Available','',''))

        PackageMetadata['ProcessingErrors'].append('Input location does not exist: ' + ReportLocation)
    else:
        if os.path.isdir(InputLocation):
            # This is for "directories" only!
            ## print '    - Scanning', InputLocation
            DirectoryListing = os.listdir(ReportLocation)
            DirectoryListing.sort()

            for File in DirectoryListing:
                Current_File = os.path.join(ReportLocation, File)
                print '        -- Found', Current_File
                if os.path.isdir(Current_File) and Current_File[-4:].upper() <> '.GDB':
                    # Directory - Standard
                    LocationDirectories.append((File,'Directory'))
                    SubLocations.append(File)
                    LocationStatisticLocations = LocationStatisticLocations + 1
                    PackageMetadata['StatisticTotalLocations'] = PackageMetadata['StatisticTotalLocations'] + 1

                elif os.path.isdir(Current_File) and Current_File[-4:].upper() == '.GDB':
                    # Directory - ESRI File Geodatabase
                    LocationDirectories.append((File,'File Geodatabase'))
                    SubLocations.append(File)
                    LocationStatisticLocations = LocationStatisticLocations + 1
                    PackageMetadata['StatisticTotalLocations'] = PackageMetadata['StatisticTotalLocations'] + 1

                elif os.path.isfile(Current_File) and Current_File[-4:].upper() == '.MDB':
                    # MS Access Database File
                    LocationDirectories.append((File,'Access Database'))
                    SubLocations.append(File)
                    LocationStatisticFiles = LocationStatisticFiles + 1
                    LocationStatisticLocations = LocationStatisticLocations + 1
                    PackageMetadata['StatisticTotalLocations'] = PackageMetadata['StatisticTotalLocations'] + 1

                # elif os.path.isfile(Current_File) and Current_File[-4:].upper() == '.DWG':
                    # AutoCAD Drawing File
                    # SubLocations.append(File)
                    # LocationStatisticFiles = LocationStatisticFiles + 1
                    # LocationStatisticLocations = LocationStatisticLocations + 1
                    # PackageMetadata['StatisticTotalLocations'] = PackageMetadata['StatisticTotalLocations'] + 1

                else:
                    # Basic File
                    FileSize = os.path.getsize(Current_File)
                    LocationFiles.append((File,str(FileSize),'Misc File'))
                    LocationStatisticFiles = LocationStatisticFiles + 1
                    PackageMetadata['StatisticTotalFiles'] = PackageMetadata['StatisticTotalFiles'] + 1

    # Now use the GeoProcessor to get all of the contents of any type of location
    print '        -- Setting Input Location: ', InputLocation
    gp.workspace = InputLocation
    print '        -- Input Set to: ', gp.workspace
    print '        -- Refreshing the Catalog: ', InputLocation
    try:
        gp.RefreshCatalog(InputLocation)
    except:
        print ' ** Error refreshing the catalog! **'
    print '        -- Listing Datasets: ', InputLocation
    datasets = gp.listdatasets("*", "feature")
    # There appears to be a bug at 9.2 that lists CAD data sets twice if you use the "all" option
    print '        -- Checking all datasets'
    dataset = datasets.next()
    while dataset:
        print '    - Adding Feature Dataset: ', dataset
        LocationStatisticLocations = LocationStatisticLocations + 1
        PackageMetadata['StatisticTotalLocations'] = PackageMetadata['StatisticTotalLocations'] + 1
        SubLocations.append(dataset)
        dataset = datasets.next()
    print '        -- Finished Checking all Datasets'
    # datasets = gp.listdatasets("*", "CAD")
    # # There appears to be a bug at 9.2 that lists CAD data sets twice if you use the "all" option
    # dataset = datasets.next()
    # while dataset:
    #     print '    - Adding CAD Dataset: ', dataset
    #     LocationStatisticLocations = LocationStatisticLocations + 1
    #     PackageMetadata['StatisticTotalLocations'] = PackageMetadata['StatisticTotalLocations'] + 1
    #     SubLocations.append(dataset)
    #     dataset = datasets.next()

    rasters = gp.ListRasters("")
    raster = rasters.next()
    while raster:
        print '   - Raster: ', raster
        LocationStatisticRasters = LocationStatisticRasters + 1
        PackageMetadata['StatisticTotalRasters'] = PackageMetadata['StatisticTotalRasters'] + 1

        raster = rasters.next()

    tables = gp.ListTables("")
    table = tables.next()
    while table:
        tableSummary = ''
        print '   - Table: ', table
        tableName = table
        LocationStatisticDatabases = LocationStatisticDatabases + 1
        PackageMetadata['StatisticTotalDatabases'] = PackageMetadata['StatisticTotalDatabases'] + 1
        tableBasicFields = []
        tableDetailedAttributes = []
        tableOID = ''
        try:
            tableTotalFeatures = gp.GetCount_management(table)
            print '      -- ', tableTotalFeatures, 'rows'        

            # Find any ObjectID Fields
            Field_List = gp.ListFields(table, "", "OID")
            Current_Field = Field_List.next()
            NumberOfObjectIDFields = 0
            while Current_Field:
                print '      -- Object ID Field Found:',Current_Field.Name
                NumberOfObjectIDFields = NumberOfObjectIDFields + 1
                tableOID = Current_Field.Name
                Current_Field = Field_List.next()
            print '      -- Number of Object ID Fields:', NumberOfObjectIDFields        

            Field_List = gp.ListFields(table, "", "")
            Current_Field = Field_List.next()

            while Current_Field:
                print '      -- ',Current_Field.Name
                if Current_Field.AliasName <> Current_Field.Name:
                    CurrentFieldName = Current_Field.Name + ' (' + Current_Field.AliasName + ')'
                else:
                    CurrentFieldName = Current_Field.Name
                tableBasicFields.append(CurrentFieldName)
                tableDetailedAttributes.append((Current_Field.Name,Current_Field.AliasName,Current_Field.Type,Current_Field.Length,Current_Field.Scale,Current_Field.Precision))
                Current_Field = Field_List.next()

            tableSummary = tableSummary + 'There are ' + str(tableTotalFeatures) + ' rows with ' + str(len(tableBasicFields)) + ' attribute fields. '        

            LocationDatabaseDataSets.append(dict([
            ('Name',tableName),
            ('Summary',tableSummary),
            ('BasicAttributes',tableBasicFields),
            ('ObjectID',tableOID),
            ('NumberOfFeatures',tableTotalFeatures),
            ('DetailedAttributes',tableDetailedAttributes)
            ]))
        except:
            print ' ** Not a valid table.'

            LocationDatabaseDataSets.append(dict([
            ('Name',tableName),
            ('Summary', '** Not a valid table.'),
            ('BasicAttributes',tableBasicFields),
            ('ObjectID',tableOID),
            ('NumberOfFeatures',0),
            ('DetailedAttributes',tableDetailedAttributes)
            ]))
        
        
        table = tables.next()


    # Get a list of feature classes recognized by the GeoProcessor
    fcs = gp.listfeatureclasses("*", "all")
    fc = fcs.next()
    while fc:
        print '   - FC: ', fc
        LocationStatisticVectors = LocationStatisticVectors + 1
        PackageMetadata['StatisticTotalVectors'] = PackageMetadata['StatisticTotalVectors'] + 1
        fcName = fc

        fcBasicFields = []
        fcDetailedAttributes = []
        fcGeometryErrors = []
        # Features in the data set
        fcTotalFeatures = 0
        # Features scanned
        fcNumber_of_Features = 0
        fcMultiPart_List = []
        fcNoGeom_List = []
        fcOID = ''
        fcExtent = 'Undefined'
        fcSpatialReference = 'Undefined'


        if fc[-4:].upper() <> '.DWG':
            try:

                fcTotalFeatures = gp.GetCount_management(fc)
                print '      -- ', fcTotalFeatures, 'features'

                dsc = gp.describe(fc)

                fcPath = dsc.CatalogPath
                LocationSpatialData.append((str(fc),0,dsc.DataType))

                fcShapeType = dsc.ShapeType
                print '      -- ', dsc.ShapeType

                #   Assemble the summary strings by testing various options and seeing what sticks
                fcSummary = dsc.ShapeType + ' ' + dsc.DataType
                if dsc.HasM:
                    fcSummary = fcSummary + ' with M values'
                if dsc.HasZ:
                    fcSummary = fcSummary + ' with Z values'
                fcSummary = fcSummary + ' in ' + dsc.SpatialReference.Name + ' projection.'

                Field_List = gp.ListFields(fc, "", "")
                Current_Field = Field_List.next()

                while Current_Field:
                    print '      -- ',Current_Field.Name
                    if Current_Field.AliasName <> Current_Field.Name:
                        CurrentFieldName = Current_Field.Name + ' (' + Current_Field.AliasName + ')'
                    else:
                        CurrentFieldName = Current_Field.Name
                    fcBasicFields.append(CurrentFieldName)
                    fcDetailedAttributes.append((Current_Field.Name,Current_Field.AliasName,Current_Field.Type,Current_Field.Length,Current_Field.Scale,Current_Field.Precision))
                    Current_Field = Field_List.next()

                fcSummary = fcSummary + 'There are ' + str(fcTotalFeatures) + ' features with ' + str(len(fcBasicFields)) + ' attribute fields. '

                # Find any ObjectID Fields
                Field_List = gp.ListFields(fc, "", "OID")
                Current_Field = Field_List.next()
                NumberOfObjectIDFields = 0
                while Current_Field:
                    print '      -- Object ID Field Found:',Current_Field.Name
                    NumberOfObjectIDFields = NumberOfObjectIDFields + 1
                    fcOID = Current_Field.Name
                    Current_Field = Field_List.next()
                print '      -- Number of Object ID Fields:', NumberOfObjectIDFields


                # Find any Geometry Fields
                Field_List = gp.ListFields(fc, "", "Geometry")
                Current_Field = Field_List.next()
                NumberOfGeometryFields = 0
                while Current_Field:
                    print '      -- Geometry Field Found:',Current_Field.Name
                    NumberOfGeometryFields = NumberOfGeometryFields + 1
                    fcGeom = Current_Field.Name
                    Current_Field = Field_List.next()
                print '      -- Number of Geometry Fields:', NumberOfGeometryFields

                fcSpatialReference = dsc.SpatialReference.Name
                fcExtent = dsc.Extent

                CheckGeometryTable = PackageMetadata['CheckGeometryTable']

                print '      --  Checking geometry to',CheckGeometryTable
                try:
                    gp.CheckGeometry_management(fc, CheckGeometryTable)
                    #### Filter this SearchCursor to just be this data set
                    GeomErrors = gp.SearchCursor(CheckGeometryTable)
                    GeomError = GeomErrors.next()
                    while GeomError:
                        GeomFID = GeomError.GetValue("FEATURE_ID")
                        GeomProblem = GeomError.GetValue("PROBLEM")
                        fcGeometryErrors.append((str(GeomFID),GeomProblem))
                        GeomError = GeomErrors.next()
                    del GeomError
                    del GeomErrors
                except:
                    print '        -- Can not check geometry'

                # Initialize Geometry Scan

                print '      --  Scanning geometry'

                Rows = gp.SearchCursor(fc)
                Row  = Rows.Next()

                #### Weird error over 12,000 in PropertyLotAnno

                # while Row and fcNumber_of_Features < 10000:
                while Row:
                    if fcGeom.upper() == 'SHAPE':
                        Feature = Row.Shape
                    if fcGeom.upper() == 'GEOMETRY':
                        Feature = Row.Geometry
                    fcNumber_of_Features = fcNumber_of_Features + 1
                    if fcNumber_of_Features % 1000 == 0:
                        print '          --  ', fcNumber_of_Features, 'complete'
                    if str(Feature.FirstPoint) == '':
                        fcNoGeom_List.append(str(Row.GetValue(fcOID)))
                    if Feature.PartCount > 1:
                        fcMultiPart_List.append(str(Row.GetValue(fcOID)))
                    Row = Rows.Next()
                del Row
                del Rows

                print '      --  Scanning geometry complete'

                Field_List = gp.ListFields(fc, "", "")
                Current_Field = Field_List.next()
            except:
                print ' ** Badly constructed Feature Class'
                fcSummary = ' ** Badly constructed Feature Class'
                fcShapeType = ' ** Badly constructed Feature Class'
                LocationSpatialData.append((str(fc),0,'BAD'))
        else:
            # This is an AutoCAD root dataset
            fcPath = InputLocation + '\\' + fc
            fcSummary = 'DWG root dataset. '
            LocationSpatialData.append((str(fc),0,'DWG'))
            fcShapeType = 'DWG collection'

        LocationVectorDataSets.append(dict([
        ('Name',fcName),
        ('Summary',fcSummary),
        ('ShapeType', fcShapeType),
        ('BasicAttributes',fcBasicFields),
        ('ObjectID',fcOID),
        ('NumberOfFeatures',fcTotalFeatures),
        ('Extent',fcExtent),
        ('Projection',fcSpatialReference),
        ('GeometryErrors',fcGeometryErrors),
        ('NumberOfFeaturesScanned',fcNumber_of_Features),
        ('MultipartFeatures',fcMultiPart_List),
        ('NoGeometryFeatures',fcNoGeom_List),
        ('DetailedAttributes',fcDetailedAttributes)
        ]))
        fc = fcs.next()

    LocationReport.append(dict([
    ('LocationRelativePath',ReportLocation.replace(PackageMetadata['ReportDataRoot'],'')),
    ('LocationSummary',LocationSummary),
    ('LocationStatisticLocations',str(LocationStatisticLocations)),
    ('LocationStatisticFiles',str(LocationStatisticFiles)),
    ('LocationStatisticVectors',str(LocationStatisticVectors)),
    ('LocationStatisticRasters',str(LocationStatisticRasters)),
    ('LocationStatisticDatabases',str(LocationStatisticDatabases)),
    ('LocationSpatialData',LocationSpatialData),
    ('LocationNonSpatialData',LocationNonSpatialData),
    ('LocationSpecialFiles',LocationSpecialFiles),
    ('LocationDirectories',LocationDirectories),
    ('LocationFiles',LocationFiles),
    ('LocationVectorDataSets',LocationVectorDataSets),
    ('LocationRasterDataSets',LocationRasterDataSets),
    ('LocationDatabaseDataSets',LocationDatabaseDataSets)
    ]))

    for SubLocation in SubLocations:
        DetailedValidation(InputLocation + '\\' + SubLocation, LocationReport, PackageMetadata)

    return(LocationReport)


def ReviewDataPackageContents(StagingArea, OutputFile, PackageMetadata):
    REPORT_START_TIME_SEC = time.time()

    # Add the title and subtitle
    PackageMetadata['ReportHyperlink'] = 'file://' + OutputFile + '.html'
    PackageMetadata['ReportTitle'] = 'Clover Point GeoData Checker Report'
    PackageMetadata['ReportSubtitle'] = INPUT_LOCATION
    PackageMetadata['ReportRunDate'] = START_TIME_SQL
    PackageMetadata['ReportToolVersion'] = PROGRAM_VERSION
    PackageMetadata['ReportRunCommand'] = PYTHON_SCRIPT + ' ' + INPUT_LOCATION
    PackageMetadata['ReportToolLocation'] = PYTHON_SCRIPT
    PackageMetadata['ReportDataRoot'] = StagingArea + '\\Extracted'
    PackageMetadata['ReportTOCTitle'] = 'Table of Contents'
    PackageMetadata['ReportTOC'] = []

    ValidationReport = []
    LocationReport = []
    PackageMetadata['Locations'] = DetailedValidation(StagingArea + '\\Extracted', LocationReport, PackageMetadata)





    ScoreCard = GenerateScoreCard(PackageMetadata)
    PackageMetadata['SummaryScoreCard'] = ScoreCard

    SummaryText = GenerateSummaryText(PackageMetadata)
    PackageMetadata['SummaryText'] = SummaryText

    PackageMetadata['RESULT_VALIDATION'] = 'PASS'


    # This must be last and outside any loops to get a correct time back.
    REPORT_END_TIME_SEC = time.time()
    REPORT_TIME = REPORT_END_TIME_SEC - REPORT_START_TIME_SEC
    PackageMetadata['StatisticRunTime'] = str(int(REPORT_TIME / 60)) + ' minutes'

    return(0)


def ReportToScreen(DataReport):
    print DataReport
    return(0)

def ReportToTXT(DataReport, OutputFile):

    return(0)

def WriteFile(FileName, FileContents):
    try:
        file = open(FileName,'w')

        # write FileContents to the fileName
        file.write(FileContents)

        # Close the file
        file.close()
        PROCESS_STATUS = 0
        PROCESS_INFO   = "Created new file and put in the comment"
    except:
        # All failures are unexpected in this case...
        # Maybe an invalid drive name, read only file or something?
        PROCESS_STATUS = 600
        PROCESS_INFO   = "**ERROR: Can't write to file: " + FileName
        return -1

def ValidationReportToHTML(DataReport, OutputFile, PackageMetadata):

    return(0)

def ExtractionReportToHTML(OutputFile, PackageMetadata):
    HTML_Header = '''
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml" >
        <head>
        <title>
            GeoData Checker Report - Provided By Clover Point Cartographics
        </title>
        </head>
        <body>
        <style>
        * {font-family: "Verdana"; font-size: 1em;}
        body { background-color: white; }
        h1,h2,h3,h4,h5,h6 {text-indent:10px;font-style: normal; color: #404040; margin: 15px 0px; padding: 5px 0px;}
        h1,h2,h3 { background-color: #E6E6E6; border: 1px solid #404040;}
        h4,h5,h6 { background-color: white; border-bottom: 1px solid #404040;}
        h1 {font-size: 1.5em;}
        h2 {font-size: 1.4em;}
        h3 {font-size: 1.3em;}
        h4 {font-size: 1.2em;}
        h5 {font-size: 1.1em;}
        h6 {font-size: 1em;}
        table {margin-left: 30px; border: 1px solid gray; border-collapse:collapse; background-color: transparent;}
        th {padding: 3px 6px; background-color: #404040; color: white; border: 1px solid #404040;}
        td {padding: 3px 6px; background-color: white; border: 1px solid #404040;}
        strong,em {color: #404040}
        p{margin-left:30px;color: rgb(0,0,0);}
        p.small {line-height:90%}

       .scorecard {}
       .spatiallist {}
       .nonspatiallist {}
       .specialfilelist {}
       .directorylist {}
       .filelist {}
       .attributelist {}
        a:link {color: #0057cd;}

        <!-- CSS Subtype Information -->
        hr.subsection {color:SteelBlue}
        .summary {color:black;line-height:110%;font-weight:bold}
        .detail {color:grey;line-height:75%}
        .error {color:red;text-transform:uppercase;font-weight:bold}
        .warning {color: #b5a200;text-transform:uppercase;font-weight:bold}
        .recommendation {color:ForestGreen;text-transform:uppercase;font-weight:bold}
        </style>

            <div>'''

# '''<a href="mailto:someone@microsoft.com?cc=someoneelse@microsoft.com&bcc=andsomeoneelse2@microsoft.com&subject=ICIS%20Data%20Report&body=Please%20send%20more%20information%20related%20to%20report%20####">email the administrator</a>'''
    HTML_Title = '''
        <!-- Document Titles and Metadata -->

        <h1>''' + PackageMetadata['ReportTitle'] + '''</h1><br />
        <h2>''' + PackageMetadata['ReportSubtitle'] + '''</h2><br />
        <p>Data Extractor Version ''' + PackageMetadata['ReportToolVersion'] + '''</p>
        <p>For details visit the <a href="http://www.cloverpoint.com/">Clover Point web site</a> or
        ''' + '''<a href="mailto:mike.n@cloverpoint.com?subject=CP GDC%20Data%20Report&body=Please%20send%20more%20information%20related%20to%20the%20CPC%20GeoData%20Checker%20version%20''' + PROGRAM_VERSION + '%20' + LICENSE + '''.">email the developer</a>''' + '''.
        </p>

        <h2>Run Parameters</h2>
        <p><strong>''' + PackageMetadata['ReportRunDate'] + ' on ' + COMPUTERNAME + '''</strong></p>
        <p><strong>Script: </strong><code>''' + PackageMetadata['ReportToolLocation'] + '''</code></p>
        <p><strong>Root: </strong><var>''' + PackageMetadata['ReportDataRoot'] + '''</var></p>
        <p><strong>Input Location: </strong><var>''' + INPUT_LOCATION + '''</var></p>'''

    HTML_TOC_Code = '''
        <!-- Table of Contents -->
        <h2><a name="TOC">''' + PackageMetadata['ReportTOCTitle'] + '''</a></h2>'''

    HTML_Parts = []


    HTML_Parts.append('''
        <!-- Executitve Summary -->
        <h2><a name="Report Summary">Report Summary</a></h2>''')

    # HTML_Parts.append('''    <h3><a name="Executive Summary">Executive Summary</a></h3>
    #    <p class="summary">''' + PackageMetadata['SummaryText'] + '''</p>''')

    PackageMetadata['ReportTOC'].append(('Report Summary','Report Summary', 2))
    PackageMetadata['ReportTOC'].append(('Executive Summary','Executive Summary',3))


    # Compile Scorecard
    HTML_Parts.append('''
        <!-- Summary Scorecard -->
        <h3><a name="Report Summary Scorecard">Report Summary Scorecard</a></h3>
        <table class="scorecard">
        <tr><th>Name</th><th>Criteria</th><th>Grade</th></tr>
        ''')
    PackageMetadata['ReportTOC'].append(('Report Summary Scorecard','Report Summary Scorecard',3))

    for Grades in PackageMetadata['SummaryScoreCard']:
        HTML_Parts.append('''<tr><td>''' + Grades[0] + '</td><td>' + Grades[1] + '</td><td>' + Grades[2] + '''</td></tr>''')

    HTML_Parts.append('''</table>''')


    HTML_Parts.append('''
        <!-- Run Statistics -->
        <h3><a name="Summary Statistics">Summary Statistics</a></h3>
        <p class="detail">
        <ul>
        <li>Total Run Time: ''' + PackageMetadata['StatisticRunTime'] + '''</li>
        <li>Number of Locations: ''' + str(PackageMetadata['StatisticTotalLocations']) + '''</li>
        <li>Number of Files: ''' + str(PackageMetadata['StatisticTotalFiles']) + '''</li>
        <li>Number of Vector Datasets: ''' + str(PackageMetadata['StatisticTotalVectors']) + '''</li>
        <li>Number of Raster Datasets: ''' + str(PackageMetadata['StatisticTotalRasters']) + '''</li>
        <li>Number of Databases: ''' + str(PackageMetadata['StatisticTotalDatabases']) + '''</li>
        </ul>
        </p>''')
    PackageMetadata['ReportTOC'].append(('Summary Statistics', 'Summary Statistics', 3))

    HTML_Parts.append('''
        <!-- Report Notes -->
        <h3><a name="Report Notes">Report Notes</a></h3>''')
    PackageMetadata['ReportTOC'].append(('Report Notes', 'Report Notes', 3))



    # Report Errors
    for Errors in PackageMetadata['ReportErrors']:
        HTML_Parts.append('''<p class="error">''' + Errors + '''</p>''')

    # Report Warnings
    for Warnings in PackageMetadata['ReportWarnings']:
        HTML_Parts.append('''<p class="warning">''' + Warnings + '''</p>''')

    # Report Recommendations
    for Recommendation in PackageMetadata['ReportRecommendations']:
        HTML_Parts.append('''<p class="recommendation">''' + Recommendation + '''</p>''')

    if len(PackageMetadata['ReportErrors']) == 0 and len(PackageMetadata['ReportWarnings']) == 0 and len(PackageMetadata['ReportRecommendations']) == 0:
        HTML_Parts.append('''<p>No report notes.</p>''')


    HTML_Parts.append('''
        <!-- Processing Notes -->
        <h3><a name="Processing Notes">Processing Notes</a></h3>''')
    PackageMetadata['ReportTOC'].append(('Processing Notes', 'Processing Notes', 3))

    # Processing Errors
    for Errors in PackageMetadata['ProcessingErrors']:
        HTML_Parts.append('''<p class="error">''' + Errors + '''</p>''')

    # Processing Warnings
    for Warnings in PackageMetadata['ProcessingWarnings']:
        HTML_Parts.append('''<p class="warning">''' + Warnings + '''</p>''')

    # Processing Recommendations
    for Recommendation in PackageMetadata['ProcessingRecommendations']:
        HTML_Parts.append('''<p class="recommendation">''' + Recommendation + '''</p>''')

    if len(PackageMetadata['ProcessingErrors']) == 0 and len(PackageMetadata['ProcessingWarnings']) == 0 and len(PackageMetadata['ProcessingRecommendations']) == 0:
        HTML_Parts.append('''<p>No processing notes.</p>''')


    HTML_Parts.append('''
        <!-- Detailed Results -->
        <h2><a name="Detailed Results">Detailed Results</a></h2>''')
    PackageMetadata['ReportTOC'].append(('Detailed Results','Detailed Results', 2))

    Location_Number = 0
    for Location in PackageMetadata['Locations']:
        Location_Number = Location_Number + 1
        HTML_Parts.append('''
            <h3><a name="Location ''' + str(Location_Number) + '''">Relative Location: Root\\''' + Location['LocationRelativePath'] + '''</a></h3>
            <h4><a name="Location''' + str(Location_Number) + '''Summary">Summary</a></h4>
            <p class="summary">''' + Location['LocationSummary'] + '''</p>
            <p class="detail">Statistics:
            <ul>
            <li>''' + Location['LocationStatisticLocations'] + ''' directories</li>
            <li>''' + Location['LocationStatisticFiles'] + ''' files</li>
            <li>''' + Location['LocationStatisticVectors'] + ''' vector data sets</li>
            <li>''' + Location['LocationStatisticRasters'] + ''' raster data sets (not checked)</li>
            <li>''' + Location['LocationStatisticDatabases'] + ''' databases (not checked)</li>
            </ul>
            </p>''')
        PackageMetadata['ReportTOC'].append(('Location ' + str(Location_Number),'''Root\\''' + Location['LocationRelativePath'],3))
        PackageMetadata['ReportTOC'].append(('Location' + str(Location_Number) + 'Summary','Summary',4))



        HTML_Parts.append('''<h4><a name="Location''' + str(Location_Number) + '''Contents">Contents</h4>
                <h5>Spatial Data List</h5>
                <table class="spatiallist">
                <tr><th>Data</th><th>Size</th><th>Data Type</th></tr>''')
        PackageMetadata['ReportTOC'].append(('Location' + str(Location_Number) + 'Contents','Contents',4))

        for DataSet in Location['LocationSpatialData']:
            HTML_Parts.append('''<tr><td>''' + DataSet[0] + '''</td><td>''' + str(DataSet[1]) + '''</td><td>''' + DataSet[2] + '''</td></tr>''')

        if len(Location['LocationSpatialData']) == 0:
            HTML_Parts.append('''<tr><td>No Spatial Data</td><td></td><td></td></tr>''')


        HTML_Parts.append('''</table>
                <h5>Non-Spatial Data List</h5>
                <table class="nonspatiallist">
                <tr><th>File</th><th>Size</th><th>Data Type</th></tr>''')

        for DataSet in Location['LocationNonSpatialData']:
            HTML_Parts.append('''<tr><td>''' + DataSet[0] + '''</td><td>''' + DataSet[1] + '''</td><td>''' + DataSet[2] + '''</td></tr>''')

        if len(Location['LocationNonSpatialData']) == 0:
            HTML_Parts.append('''<tr><td>No Non-Spatial Data</td><td></td><td></td></tr>''')


        HTML_Parts.append('''</table>

                <h5>Special Files</h5>

                <table class="specialfilelist">
                <tr><th>File</th><th>Size</th><th>Data Type</th></tr>''')

        for DataSet in Location['LocationSpecialFiles']:
            HTML_Parts.append('''<tr><td>''' + DataSet[0] + '''</td><td>''' + DataSet[1] + '''</td><td>''' + DataSet[2] + '''</td></tr>''')

        if len(Location['LocationSpecialFiles']) == 0:
            HTML_Parts.append('''<tr><td>No Special Files</td><td></td><td></td></tr>''')


        HTML_Parts.append('''
                </table>
                <h5>Directory List</h5>
                <table class="directorylist">
                <tr><th>Directory</th><th>Type</th></tr>''')

        for DataSet in Location['LocationDirectories']:
            HTML_Parts.append('''<tr><td>''' + DataSet[0] + '''</td><td>''' + DataSet[1] + '''</td></tr>''')

        if len(Location['LocationDirectories']) == 0:
            HTML_Parts.append('''<tr><td>No Directories</td><td></td></tr>''')

        HTML_Parts.append('''
                </table>

                <h5>Complete File List</h5>

                <table class="filelist">
                <tr><th>File</th><th>Size</th><th>Data Type</th></tr>''')

        for DataSet in Location['LocationFiles']:
            HTML_Parts.append('''<tr><td>''' + DataSet[0] + '''</td><td>''' + DataSet[1] + '''</td><td>''' + DataSet[2] + '''</td></tr>''')

        if len(Location['LocationFiles']) == 0:
            HTML_Parts.append('''<tr><td>No Files</td><td></td><td></td></tr>''')

        HTML_Parts.append('''</table>
                <h4><a name="Location''' + str(Location_Number) + '''DataSets">Data Sets</a></h4>''')
        PackageMetadata['ReportTOC'].append(('Location' + str(Location_Number) + 'DataSets','Data Sets',4))


        for DataSet in Location['LocationVectorDataSets']:
            HTML_Parts.append('''<h5><a name="Location ''' + str(Location_Number) + DataSet['Name'] +  '''">''' + DataSet['Name'] + '''</a></h5>
                <h6>Summary</h6>
                <p class="summary">''' + DataSet['Summary'] + '''</p>''')

            # HTML_Parts.append('''    <h6>Score Card</h6>
            #     <h6>File Details</h6>''')


            HTML_Parts.append('''<h6>Basic Attributes</h6>''')

            HTML_Parts.append('''<ul>''')

            for Attribute in DataSet['BasicAttributes']:
                HTML_Parts.append('''<li>''' + Attribute + '''</li>''')

            HTML_Parts.append('''</ul>''')

            HTML_Parts.append('''<h6>Detailed Attributes</h6>''')

            HTML_Parts.append('''<table class="attributelist">
                <tr><th>Name</th><th>Alias</th><th>Type</th><th>Length</th><th>Scale</th><th>Precision</th></tr>''')

            for Attribute in DataSet['DetailedAttributes']:
                HTML_Parts.append('''<tr><td>''' + Attribute[0] + '''</td><td>''' + Attribute[1] + '''</td><td>''' + Attribute[2] + '''</td><td>''' + str(Attribute[3]) + '''</td><td>''' + str(Attribute[4]) + '''</td><td>''' + str(Attribute[5]) + '''</td></tr>''')

            HTML_Parts.append('''</table>''')

            # HTML_Parts.append('''<h6>Attribute Values</h6>'''

            HTML_Parts.append('''    <h6>Basic Geometry</h6>''')

            HTML_Parts.append('''<p>There are ''' + str(DataSet['NumberOfFeatures']) + ''' ''' + DataSet['ShapeType'] + ''' features.</p>''')
            HTML_Parts.append('''<p>The extent is ''' + str(DataSet['Extent']) + '''.</p>''')
            HTML_Parts.append('''<p>The projection is ''' + str(DataSet['Projection']) + '''.</p>''')

            HTML_Parts.append('''
                <h6>Detailed Geometry</h6>''')
            HTML_Parts.append('''<p>There are ''' + str(DataSet['NumberOfFeaturesScanned']) + ''' ''' + DataSet['ShapeType'] + ''' features that were scanned in detail.</p>''')
            HTML_Parts.append('''<p>There are ''' + str(len(DataSet['MultipartFeatures'])) + ''' multipart features.</p>''')
            if len(DataSet['MultipartFeatures']) > 0 :
                HTML_Parts.append('''<p> Feature IDs: ''' + str(DataSet['MultipartFeatures']) + '''</p>''')

            HTML_Parts.append('''<p>There are ''' + str(len(DataSet['NoGeometryFeatures'])) + ''' features without geometry.</p>''')

            if len(DataSet['NoGeometryFeatures']) > 0 :
                HTML_Parts.append('''<p> Feature IDs" ''' + str(DataSet['NoGeometryFeatures']) + '''</p>''')

            if len(DataSet['GeometryErrors']) == 0:
                HTML_Parts.append('''<p>There were no geometry errors found.</p>''')
            else:
                HTML_Parts.append('''<p>There are ''' + str(len(DataSet['GeometryErrors'])) + ''' geometry errors:</p><table><tr><th>Feature ID</th><th>Problem</th></tr>''')
                for GeomError in DataSet['GeometryErrors']:
                    HTML_Parts.append('''<tr><td>''' + str(GeomError[0]) + '''</td><td>''' + GeomError[1] + '''</td></tr>''')
                HTML_Parts.append('''</table>''')

            # HTML_Parts.append('''
            #     <h6>Relationships</h6>
            #     <h6>Metadata</h6>
            #     ''')


            PackageMetadata['ReportTOC'].append(('Location ' + str(Location_Number) + DataSet['Name'], DataSet['Name'], 5))



        # for DataSet in Location['LocationRasterDataSets']:
        #     HTML_Parts.append('''<h5>''' + DataSet['Name'] + '''</h5>
        #         <h6>Summary</h6>
        #         <h6>Score Card</h6>
        #         <h6>File Details</h6>
        #         <h6>Basic Attributes</h6>
        #         <h6>Detailed Attributes</h6>
        #         <h6>Attribute Values</h6>
        #         <h6>Basic Geometry</h6>
        #         <h6>Metadata</h6>
        #         ''')
        #    PackageMetadata['ReportTOC'].append(('Location ' + str(Location_Number) + DataSet['Name'], DataSet['Name'], 5))
        # if len(Location['LocationRasterDataSets']) == 0:
        #     HTML_Parts.append('''<p>No Raster Data Sets Found.</p>''')


        # for DataSet in Location['LocationDatabaseDataSets']:
        #     HTML_Parts.append('''<h5>''' + DataSet['Name'] + '''</h5>
        #         <h6>Summary</h6>
        #         <h6>Score Card</h6>
        #         <h6>File Details</h6>
        #         <h6>Basic Attributes</h6>
        #         <h6>Detailed Attributes</h6>
        #         <h6>Attribute Values</h6>
        #         <h6>Relationships</h6>
        #         <h6>Metadata</h6>
        #         ''')
        #     PackageMetadata['ReportTOC'].append(('Location ' + str(Location_Number) + DataSet['Name'], DataSet['Name'], 5))
        # if len(Location['LocationDatabaseDataSets']) == 0:
        #     HTML_Parts.append('''<p>No Databases Found.</p>''')


        for DataSet in Location['LocationDatabaseDataSets']:
            HTML_Parts.append('''<h5><a name="Location ''' + str(Location_Number) + DataSet['Name'] +  '''">''' + DataSet['Name'] + '''</a></h5>
                <h6>Summary</h6>
                <p class="summary">''' + DataSet['Summary'] + '''</p>''')

            # HTML_Parts.append('''    <h6>Score Card</h6>
            #     <h6>File Details</h6>''')


            HTML_Parts.append('''<h6>Basic Attributes</h6>''')

            HTML_Parts.append('''<ul>''')

            for Attribute in DataSet['BasicAttributes']:
                HTML_Parts.append('''<li>''' + Attribute + '''</li>''')

            HTML_Parts.append('''</ul>''')

            HTML_Parts.append('''<h6>Detailed Attributes</h6>''')

            HTML_Parts.append('''<table class="attributelist">
                <tr><th>Name</th><th>Alias</th><th>Type</th><th>Length</th><th>Scale</th><th>Precision</th></tr>''')

            for Attribute in DataSet['DetailedAttributes']:
                HTML_Parts.append('''<tr><td>''' + Attribute[0] + '''</td><td>''' + Attribute[1] + '''</td><td>''' + Attribute[2] + '''</td><td>''' + str(Attribute[3]) + '''</td><td>''' + str(Attribute[4]) + '''</td><td>''' + str(Attribute[5]) + '''</td></tr>''')

            HTML_Parts.append('''</table>''')

            # HTML_Parts.append('''<h6>Attribute Values</h6>'''

            # HTML_Parts.append('''
            #     <h6>Relationships</h6>
            #     <h6>Metadata</h6>
            #     ''')


            PackageMetadata['ReportTOC'].append(('Location ' + str(Location_Number) + DataSet['Name'], DataSet['Name'], 5))

        if len(Location['LocationDatabaseDataSets']) == 0:
             HTML_Parts.append('''<p>No Databases Found.</p>''')
        
        if len(Location['LocationVectorDataSets']) == 0 and len(Location['LocationRasterDataSets']) == 0 and len(Location['LocationDatabaseDataSets']):
            HTML_Parts.append('''<p>No Data Sets Found.</p>''')




    # Complete the report
    HTML_Parts.append('''<h1>End of Report</h1>''')



    TOC_Current_Level = 1
    for TOC_Element in PackageMetadata['ReportTOC']:
        while TOC_Current_Level < TOC_Element[2]:
            TOC_Current_Level = TOC_Current_Level + 1
            HTML_TOC_Code = HTML_TOC_Code + '''<ul>'''
        while TOC_Current_Level > TOC_Element[2]:
            TOC_Current_Level = TOC_Current_Level - 1
            HTML_TOC_Code = HTML_TOC_Code + '''</ul>'''

        HTML_TOC_Code = HTML_TOC_Code + '''<li><a href="#''' + TOC_Element[0] + '''">''' + TOC_Element[1] + '''</a></li>'''


    while TOC_Current_Level > 1:
        TOC_Current_Level = TOC_Current_Level - 1
        HTML_TOC_Code = HTML_TOC_Code + '''</ul>'''


    HTML_Div = ''
    for HTML_Code in HTML_Parts:
        HTML_Div = HTML_Div + HTML_Code

    HTML_Footer = '''</div><IFRAME NAME="CP_GDC" SRC="http://www.cloverpoint.com/CP_GDC/index.php?'''

    HTML_Footer = HTML_Footer + '''version=''' + PROGRAM_VERSION
    HTML_Footer = HTML_Footer + '''&license=''' + LICENSE
    HTML_Footer = HTML_Footer + '''&computer=''' + COMPUTERNAME
    #### HTML_Footer = HTML_Footer + '''&user=''' + USERNAME
    HTML_Footer = HTML_Footer + '''&user=PublicBetaUser'''
    HTML_Footer = HTML_Footer + '''&datetime=''' + START_TIME
    HTML_Footer = HTML_Footer + '''&rootpath=''' + INPUT_LOCATION
    HTML_Footer = HTML_Footer + '''&runtime=''' + str(PackageMetadata['StatisticRunTime'])
    HTML_Footer = HTML_Footer + '''&locations=''' + str(PackageMetadata['StatisticTotalLocations'])
    HTML_Footer = HTML_Footer + '''&files=''' + str(PackageMetadata['StatisticTotalFiles'])
    HTML_Footer = HTML_Footer + '''&vectors=''' + str(PackageMetadata['StatisticTotalVectors'])
    HTML_Footer = HTML_Footer + '''&rasters=''' + str(PackageMetadata['StatisticTotalRasters'])
    HTML_Footer = HTML_Footer + '''&databases=''' + str(PackageMetadata['StatisticTotalDatabases'])
    HTML_Footer = HTML_Footer + '''&comments=''' + 'No comments'



    HTML_Footer = HTML_Footer + '''" WIDTH=300 HEIGHT=100 SCROLLING="NO" FRAMEBORDER="0"></IFRAME></body></html>'''

    HTML_Document = HTML_Header + HTML_Title + HTML_TOC_Code + HTML_Div + HTML_Footer

    WriteFile(OutputFile, HTML_Document)

    #### or use OutputFile
    os.system('explorer ' + PackageMetadata['ReportHyperlink'])

    return(0)


def ReportToDOC(DataReport, OutputFile):

    return(0)

def ReportToDatabase(DatabaseConnection, DataReport):
    print ' To Database:', DatabaseConnection
    print ' Content:', DataReport
    return(0)

def ArchiveOldRecords(ProductionArea, ArchiveArea):

    return(0)

def AddNewRecords(StagingArea, ProductionArea):

    return(0)

def CompareAreas(StagingArea, ArchiveArea):
    DeltaReport = [('RESULT_DELTA','80%'),('TITLE','Delta Report'),('SUBTITLE','Test data only')]
    return(DeltaReport)

def CleanUpStagingArea(StagingArea):

    return(0)


#*********************************************************************************************************
#  Main Program
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# Run Main
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain():
    global PROCESS_INFO, PROCESS_STATUS

    ProcessingList = [INPUT_LOCATION]
    # print ProcessingList
    for DataPackage in ProcessingList:
        # Set up the Metadata construct (a dictionary)
        PackageMetadata = dict(
            ComputerName = COMPUTERNAME,
            CheckGeometryTable = '',
            ReportTitle = '',
            ReportSubtitle = '',
            ReportRunDate = '',
            ReportToolVersion = '',
            ReportRunCommand = '',
            ReportToolLocation = '',
            ReportDataRoot = '',
            ReportTOCTitle = '',
            SummaryText = '',
            SummaryScoreCard = '',
            StatisticRunTime = '',
            StatisticTotalLocations = 1,
            StatisticTotalFiles = 0,
            StatisticTotalVectors = 0,
            StatisticTotalRasters = 0,
            StatisticTotalDatabases = 0,
            ReportErrors = [],
            ReportWarnings = [],
            ReportRecommendations = [],
            ProcessingErrors = [],
            ProcessingWarnings = [],
            ProcessingRecommendations = [],
            ReportHyperlink = '',
            RESULT_VALIDATION = '',
            RESULT_DELTA = '',
            EXITCODE = '')
        StagingArea = CreateStagingArea(DataPackage, PackageMetadata)
        OutputFile = StagingArea + '\\ExtractionReport'
        ExtractDataPackage(DataPackage, StagingArea)
        ReviewDataPackageContents(StagingArea, OutputFile, PackageMetadata)
        # ReportToScreen(PackageMetadata)
        # ReportToTXT(PackageMetadata, OutputFile + '.txt')
        # ExtractionReportToHTML(OutputFile + '.html', PackageMetadata)
        ExtractionReportToHTML(OutputFile + '.html', PackageMetadata)
        # ReportToDOC(PackageMetadata, OutputFile + '.doc')
        # DatabaseConnection = 'ODBC Connection'
        # ReportToDatabase(DatabaseConnection, PackageMetadata)

        # if PackageMetadata['RESULT_VALIDATION'] == 'PASS':
        #     ArchiveOldRecords(ProductionArea, ArchiveArea)
        #     AddNewRecords(StagingArea, ProductionArea)
        #     DeltaReport = CompareAreas(StagingArea, ArchiveArea)
        #     # ReportToScreen(DeltaReport)
        #     OutputFile = StagingArea + '\\ComplianceReport'
        #     # ReportToTXT(DeltaReport, OutputFile)
        #     ValidationReportToHTML(DeltaReport, OutputFile + '.html', PackageMetadata)
        #     # ReportToDOC(DeltaReport, OutputFile)
        #     DatabaseConnection = 'ODBC Connection'
        #     ReportToDatabase(DatabaseConnection, DeltaReport)

        CleanUpStagingArea(StagingArea)

    return PROCESS_STATUS

#*********************************************************************************************************
#  Execution
#*********************************************************************************************************

PROCESS_STATUS   = 0
runMain()



#---------------------------------------------------------------------------------------------------------
#  Clean-up
#---------------------------------------------------------------------------------------------------------

# gp.CheckInExtension("spatial")
del gp
END_TIME       = time.ctime(time.time())
END_TIME_SEC   = time.time()
END_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
EXECUTION_TIME = END_TIME_SEC - START_TIME_SEC

# WRITE_OUTPUT = os.system(CPC_SCRIPT_WRITE_OUTPUT + ' "Complete ' + PYTHON_SCRIPT + '" 1 1 1 ' + CPC_QA_LOG_FILE)
print 'Complete ' + PYTHON_SCRIPT


# Report Program Execution Status back to the calling program
sys.exit(PROCESS_STATUS)

#*********************************************************************************************************
#  End of Program
#*********************************************************************************************************
