#=========================================================================================================
#                                          RepairGeometry
#=========================================================================================================
#
# Author(s)         :  Clover Point Cartographics (Mike Neal, Martin Davis, Jeff Kruys)
#
# Date              :  2012-10-25
#
# Purpose           :  This script checks polygons for SDE-invalid geometry (ie. inverted polygons where
#                      the hole that touches the outer shell is represented as a separate part of the 
#                      multi-part polygon, rather than as part of the outer shell). Also executes the 
#                      ESRI "RepairGeometry" tool to fix other invalid polygons.
#
# Arguments         :  argv(1) = Long Table feature class
#
# Example Input     :  X:\fullpath\StandardizeNoData.py X:\fullpath\TEI_Long_Tbl
#
# Outputs           :  No new outputs are created. The input feature class or table is modified (ie. the 
#                      geometry of SDE-invalid polygons is updated to become SDE-valid).
#
# Dependencies      :  None
#
# Expected Duration :  Approx. ______ for 200,000 records
#                      Approx. ______ for 24 million records
#
# History           :  2013-04-22 (JK): Removed function to lookup IP address (script crashes if there is
#                                       no internet connectivity)
#
#                   :  2013-06-19 (JK): Updated header comments.
#
#                   :  2014-03-20 (JK): Changed it so the Arc RepairGeometry tool is executed before running
#                                       the routine to fix SDE-invalid geometry. Update the ArcGIS version
#                                       check so it can handle 10.2 and above.
#
#=========================================================================================================
#=========================================================================================================
#  Process Status and Return Codes:
#      -1   - Unexpected error
#       0   - Completed Successful Run
#       1   - In progress
#       100 - Invalid Arguments
#       101 - ESRI License Error
#
#=========================================================================================================

#*********************************************************************************************************
#  Program Initialization
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
#  Python and System Modules
#---------------------------------------------------------------------------------------------------------

from __future__ import division

print '-- Importing Modules'

import sys
import string
import os
import stat # index of constants for os.stat()
import time
import socket
import zipfile

# import re
# import shutil
# import dircache
# import win32com.client
# import cpc_gis_common

#---------------------------------------------------------------------------------------------------------
#  Arguement Validation
#---------------------------------------------------------------------------------------------------------

print '-- Getting Run Parameters'

# Capture/Validate Parameters
#  Test that the right number of paramters were provided and validated
#  Note that 1 must be added because the Python script is an arguement (0) supplied to the
#  Python Environment.


# If long text parameters are required, use:  any_value = arcpy.GetParameterAsText(0)

if len(sys.argv) < 2 or len(sys.argv) > 2 :

    print "\nInvalid number of arguments (" + str(len(sys.argv) - 1) + " supplied) to:"
    print "    " + sys.argv[0] + "\n"
    print " Usage:   " + sys.argv[0] + " [ARG1]"
    print "    [ARG1]   : Feature Class to fix"

    sys.exit(100)

PYTHON_SCRIPT   = sys.argv[0]
INPUT_LOCATION  = sys.argv[1]


#---------------------------------------------------------------------------------------------------------
#  Standard/Global Variables/Constants
#---------------------------------------------------------------------------------------------------------
print '-- Setting Global Variables'

# Global Constants
COMPUTERNAME     = os.environ['COMPUTERNAME']
USERNAME         = os.environ['USERNAME']
START_TIME       = time.ctime(time.time())
START_TIME_SEC   = time.time()
START_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
BEEP             = chr(7)
#s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
#s.connect(("google.com",0))
#IP_ADDRESS       = s.getsockname()[0]
#s.close()
#del s

# Global Variables
#   The Processing_Variables is the collection of messaging and script specific information needed
#   throughout this program.  It is constantly updated, and generally includes all paths and configuration
#   information.  For parameters that are passed in to the script, they should generally be assigned
#   to a Processing_Variable through the Initialize() routine.

Processing_Variables = {}
RunStatus = {}



#---------------------------------------------------------------------------------------------------------
#  CPC Variables/Constants
#---------------------------------------------------------------------------------------------------------




#---------------------------------------------------------------------------------------------------------
#  Set up the ArcGIS Geoprocessing Environment
#---------------------------------------------------------------------------------------------------------

try:
    import arcgisscripting
    arcpy = arcgisscripting.create(93)
except:
    print '**** ArcGIS Failed to initialize a GeoProcessor Object'
    sys.exit(101)

print '-- Checking Script Pre-Conditions'

d = arcpy.GetInstallInfo('desktop')
if float(d['Version'][:4]) >= 10.1:
    print "Arc " +  d['Version'][:4] + " detected on this machine"
    del arcpy
    import arcpy

if arcpy.Exists(INPUT_LOCATION):
    print '    - Input dataset exists:', INPUT_LOCATION
else:
    print ' **** Input dataset does not exist:', INPUT_LOCATION
    sys.exit(100)

desc = arcpy.Describe(INPUT_LOCATION)
if desc.shapeType == 'Polygon':
    print '    - Input dataset has polygon geometry.'
else:
    print ' **** Input dataset does not have polygon geometry.'
    sys.exit(100)

#---------------------------------------------------------------------------------------------------------
#  Script Specific Variables/Constants
#---------------------------------------------------------------------------------------------------------




#*********************************************************************************************************
#  Routines
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# TemplateRoutine(Input_FC)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TemplateRoutine(Input_FC):
    CPMessage('    - Template Routine', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TemplateRoutine" 

    result = arcpy.GetCount_management(Input_FC)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Calculating individual values for ' + str(Number_of_Features) + ' features.', 4)
    rows = arcpy.SearchCursor(Input_FC)
    row = rows.Next()
    
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features), 4)

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    del row
    del rows

    return(Processing_Variables['PROCESS_STATUS'])

#=========================================================================================================
#  Standard Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# Initialize()
#
#       Set up all of the paths and variables that we need to run this script.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def Initialize():
    print '    - Initializing Script'

    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Script_Name'] = os.path.split(PYTHON_SCRIPT)[1]
    Processing_Variables['Log_Directory_Path'] = Processing_Variables['Script_Directory_Path'] + r'\log'

    Processing_Variables['Project_Base_Path']      = os.path.split(Processing_Variables['Script_Directory_Path'])[0]

    
    Processing_Variables['ToolBoxRun'] = 'FALSE'
    # Processing_Variables['SendRunLogToEmail'] = ['mneal@cloverpoint.com', 'chaos_mike@hotmail.com']
    Processing_Variables['Log_Level'] = 4
    Processing_Variables['PROCESS_INFO']   = "RUNNING " + PYTHON_SCRIPT
    Processing_Variables['PROCESS_STATUS'] = 1
    Processing_Variables['PROCESS_LOG'] = []
    Processing_Variables['RunLogFile']   = Processing_Variables['Log_Directory_Path'] + '\\' + Processing_Variables['Script_Name'] + '.' + COMPUTERNAME + '.log'

    Processing_Variables['Scratch_Path']           = Processing_Variables['Project_Base_Path'] + r'\Scratch'
    if not os.path.exists(Processing_Variables['Scratch_Path']):
        CPMessage('Creating Scratch folder ' + Processing_Variables['Scratch_Path'], 1)
        os.mkdir(Processing_Variables['Log_Directory_Path'])

    Processing_Variables['ScratchFGDB_Path']           = Processing_Variables['Scratch_Path'] + r'\Scratch.gdb'
    if not arcpy.Exists(Processing_Variables['ScratchFGDB_Path']):
        CPMessage('Creating Scratch FGDB ' + Processing_Variables['ScratchFGDB_Path'], 1)
        arcpy.CreateFileGDB_management(Processing_Variables['Scratch_Path'], "Scratch", "9.3")
        
    Processing_Variables['Spatial_Reference'] = Processing_Variables['Project_Base_Path'] + r'\Scripts\NAD_1983_BC_Environment_Albers.prj'

    #### Adding a FGDB Log will significantly slow down performance.
    # Processing_Variables['Log_FGDB_Table'] = Processing_Variables['Project_Base_Path'] + r"\Operational.gdb\Processing_Log"

    if not arcpy.Exists(Processing_Variables['Log_Directory_Path']):
        os.mkdir(Processing_Variables['Log_Directory_Path'])
        CPMessage(r'Created \Scripts\log folder', 1)
    
    CPMessage('Executing: ' + PYTHON_SCRIPT, 1)
    CPMessage('    Start Time: ' + START_TIME, 2)
    CPMessage('    Python ' + sys.version, 4)
    #CPMessage('    Computer ' + COMPUTERNAME + ' (' + IP_ADDRESS + ')', 4)
    CPMessage('    Computer ' + COMPUTERNAME, 4)
    

    
    # Load required toolboxes...
    # arcpy.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")
    
    # Set the Geoprocessing environment...
    arcpy.OverwriteOutput = 1
    # arcpy.cellSize = "MINOF"
    # arcpy.workspace = INPUT_LOCATION
    # arcpy.scratchworkspace = CPC_LOCAL_TEMP_SPACE
    # arcpy.XYResolution = "" 
    # arcpy.XYDomain = "" 
    # arcpy.MTolerance = "" 
    # arcpy.randomGenerator = "0 ACM599" 
    # arcpy.outputCoordinateSystem = "" 
    # arcpy.ZDomain = "" 
    # arcpy.configKeyword = "" 
    # arcpy.outputZFlag = "Same As Input" 
    # arcpy.qualifiedFieldNames = "true" 
    # arcpy.extent = "DEFAULT" 
    # arcpy.XYTolerance = "" 
    # arcpy.MDomain = "" 
    # arcpy.spatialGrid1 = "0" 
    # arcpy.outputZValue = "" 
    # arcpy.outputMFlag = "Same As Input" 
    # arcpy.geographicTransformations = "" 
    # arcpy.spatialGrid2 = "0" 
    # arcpy.ZResolution = "" 
    # arcpy.spatialGrid3 = "0" 
    # arcpy.MResolution = "" 
    # arcpy.ZTolerance = ""
    
    # arcpy.CreateFolder_management()
    
    
    
    
    # CPC Naming Standard: {variable}_Path is the fully qualified path, otherwise it is just the name of the item.
  
    # Processing Variables
    
    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Working_Directory_Path'] = INPUT_LOCATION
    
    return()




#---------------------------------------------------------------------------------------------------------
# SetupWorkspace()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SetupWorkspace():
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "SetupWorkspace"  

    CPMessage('    - Setting Up Workspace', 2)


    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# CPMessage(Message)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CPMessage(Message, MessageLevel):
    # Based on the logging level, add messages to the message queues.
    #   Text Run Log   : Implimented
    #   Console        : Implimented
    #   GeoProcessor   : Implimented
    #   Processing Log : Implimented (with e-mail)
    #   SQL Database   : Not Implimented ####
    #   FGDB           : Implimented
    
    # Message Levels:
    #   0    : Fatal Errors
    #   1    : Top Level Program
    #   2    : Main routines and events
    #   3    : Warnings
    #   4    : General Information
    #   5    : Debug level detail
    
    # Send to the screen
    if MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToScreen(Message, MessageLevel)    
    
    # Send to the Processing Log
    if 'SendRunLogToEmail' in Processing_Variables:
        Processing_Variables['PROCESS_LOG'].append((Message, MessageLevel))
    
    # Send to the Log File
    if 'RunLogFile' in Processing_Variables and MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToFile(Message, Processing_Variables['RunLogFile'])

    # Send to the GeoProcessor Message Queue
    #### At ArcGIS 9.3 SP1 there is a bug that also causes any GeoProcessor messages to be echoed to the console.
    #    Added the ToolBoxRun parameter to allow the script to add messages to the GeoProcessor Messaging properly
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and Processing_Variables['ToolBoxRun'] == 'TRUE':
        if MessageLevel == 0:
            arcpy.AddError(Message)
        if MessageLevel == 1 or MessageLevel == 2:
            arcpy.AddMessage(Message)
        if MessageLevel == 3:
            arcpy.AddWarning(Message)

    # Send to the FGDB Processing Log
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and 'Log_FGDB_Table' in Processing_Variables:
        rows = arcpy.InsertCursor(Processing_Variables['Log_FGDB_Table'])

        row = rows.NewRow()
        row.SetValue("Computer_Name", COMPUTERNAME)
        row.SetValue("User_Name", USERNAME)
        row.SetValue("Script_Name", PYTHON_SCRIPT)
        row.SetValue("Event_Time", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
        row.SetValue("Message_Level", MessageLevel)
        row.SetValue("Message", Message)
        row.SetValue("Start_Time", START_TIME_SQL)

        rows.InsertRow(row)
        del row
        del rows
    

    return(1)


#---------------------------------------------------------------------------------------------------------
# WriteOutputToScreen(Output_Comment, Header_Style):
#
#   Prints the Output_Comment to the screen with the appropriate formatting
#       0 : "******" Used for Error Sections
#       1 : "======" Used for the Program's Header and Footer
#       2 : "------" Used for New Sections
#       3+:          No delimitation
#   
#---------------------------------------------------------------------------------------------------------
def WriteOutputToScreen(Output_Comment, Header_Style):
    if Header_Style == 0:
        # Errors get beeps, too!
        print BEEP * 3 + "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 2:
        print "\n" + "-" * 79 + "\n"
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"
        
    #print Output_Comment

    arcpy.AddMessage(Output_Comment)
    
    if Header_Style == 0:
        print "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"


#---------------------------------------------------------------------------------------------------------
# WriteOutputToFile(Output_Comment, Run_Log)
#   Record to Log 
#
#
#---------------------------------------------------------------------------------------------------------
def WriteOutputToFile(Output_Comment, Run_Log):
# Run logs are not compatible with concurrency - only one process is expected in debug mode at a time!
# First try to append to an existing file - if that fails create a new file.
    ReturnValue = 1
    try:
        if os.path.isfile(Run_Log) == 1:
            # try to open up the fileName 
            file = open(Run_Log,'a') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Appended comment to the file"
        else:
            # try to open up the fileName 
            file = open(Run_Log,'w') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created new file and put in the comment"
    except:
        # All failures are unexpected in this case...  
        # Maybe an invalid drive name, read only file or something?
        Processing_Variables['PROCESS_STATUS'] = -1
        Processing_Variables['PROCESS_INFO']   = "**ERROR: Can't write to file: " + Run_Log   
        print BEEP + BEEP + BEEP
        ReturnValue = -1
    return(ReturnValue)
        

#---------------------------------------------------------------------------------------------------------
# SendRunLogToEmail()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SendRunLogToEmail():
    import smtplib
    try:
        Seconds_Since_Last_Message = time.time() - time.mktime(time.strptime(RunStatus['LastStatusEmailSent'], "%Y-%m-%d %H:%M:%S" ))
    except:
        # No message ever sent, or bad data, so send one now.
        Seconds_Since_Last_Message = -1

    if Seconds_Since_Last_Message < 120:
        Last_Message_Time = str(int(Seconds_Since_Last_Message)) + ' seconds.'
    elif Seconds_Since_Last_Message < 3600:
        Last_Message_Time = str(round(Seconds_Since_Last_Message / 60, 1)) + ' minutes.'
    else:
        Last_Message_Time = str(round(Seconds_Since_Last_Message / 3600, 2)) + ' hours.'


    if int(RunStatus['MinimumMessageInterval']) < 120:
        Minimum_Message_Time = str(int(RunStatus['MinimumMessageInterval'])) + ' seconds.'
    elif Seconds_Since_Last_Message < 3600:
        Minimum_Message_Time = str(round(int(RunStatus['MinimumMessageInterval']) / 60, 1)) + ' minutes.'
    else:
        Minimum_Message_Time = str(round(int(RunStatus['MinimumMessageInterval']) / 3600, 2)) + ' hours.'


    CPMessage('    Time since last message sent: ' + Last_Message_Time, 4)        
    CPMessage('    Minimum message interval is : ' + Minimum_Message_Time, 4)
    
    for mailto in Processing_Variables['SendRunLogToEmail']:
        if int(RunStatus['MinimumMessageInterval']) < Seconds_Since_Last_Message or Seconds_Since_Last_Message == -1:
            CPMessage('  * Sending log to ' + mailto, 4)
            mailfrom = 'GIS_Python@cloverpoint.com'
            # mailto = 'mneal@cloverpoint.com'
            mailsubject = 'Run Log: ' + os.path.split(PYTHON_SCRIPT)[1] + ' (' + COMPUTERNAME + '\\' + USERNAME + ')'
            mailbody = "This is an automatic run log.\n\n"
            
            
            mailbody = mailbody + '\n\n  ** Note: The minimum time between messages for this monitor is: ' + Minimum_Message_Time + '.\n'

            for LogEntry in Processing_Variables['PROCESS_LOG']:
                mailbody = mailbody + '\n' + LogEntry[0]

            mailbody = mailbody + '\n\nEnd of Line.\n'

            mailmsg = ("From: " + mailfrom + "\r\nTo: " + mailto + "\r\nSubject: " + mailsubject + "\r\n" + mailbody)

            mailserver = smtplib.SMTP('shawmail.gv.shawcable.net')
            # mailserver = smtplib.SMTP('localhost')

            # SMTP.login(user, password)

            mailserver.sendmail(mailfrom, mailto, mailmsg)
            mailserver.quit()
            RunStatus['LastStatusEmailSent'] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        else:
            CPMessage('  ** NOT Sending log to ' + mailto + ' due to short interval since last message.', 3)
            
    
    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# SanitizeElapsedTime(number_of_seconds)
#
# routine to convert a raw number of seconds into a human-friendly statement of duration
#---------------------------------------------------------------------------------------------------------

def SanitizeElapsedTime(dtInput):

    if dtInput < 120:
        strElapsedTime = str(int(dtInput)) + ' seconds.'
        
    elif dtInput < 3600:
        strElapsedTime = str(int(dtInput)) + ' seconds (' + str(round(dtInput / 60, 1)) + ' minutes).'
        
    else:
        strElapsedTime = str(int(dtInput)) + ' seconds (' + str(round(dtInput / 3600, 2)) + ' hours).'
        
    return strElapsedTime

#=========================================================================================================
#  Special Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# HasTouchingHole(geometry_object)
#
# test whether an Arc geometry has a hole that touches the outer shell
#---------------------------------------------------------------------------------------------------------
def HasTouchingHole(geom):

    if d['Version'] == '9.3':

        partcount = geom.PartCount
        partnum = 0
        while partnum < partcount:
            part = geom.GetPart(partnum)
            # Step through each vertex in the geometry, handling hole separators
            #
            shellDict = {}
            # load shell pnts into dict
            p = part.next()
            while p:
                shellDict[ (p.X, p.Y) ] = 1
                p = part.next()

            # for any holes, check if any vertex is in shell  
            p = part.next()
            while p:
                pntTup = (p.X, p.Y)
                if pntTup in shellDict:
                    return True
                p = part.next()
                if not p:
                    p = part.next()
            partnum += 1

    else:
        # Step through each part of the geometry
        #
        for part in geom:

            # Step through each vertex in the geometry, handling hole separators
            #
            shellDict = {}
            # load shell pnts into dict
            p = part.next()
            while p:
                shellDict[ (p.X, p.Y) ] = 1
                p = part.next()

            # for any holes, check if any vertex is in shell  
            p = part.next()
            while p:
                pntTup = (p.X, p.Y)
                if pntTup in shellDict:
                    return True
                p = part.next()
                if not p:
                    p = part.next()

    return False

#---------------------------------------------------------------------------------------------------------
# HasHole(geometry_object)
#
# test whether an Arc geometry has any kind of hole        
#---------------------------------------------------------------------------------------------------------
def HasHole(geom):

    if d['Version'] == '9.3':
        # Step through each part of the geometry
        #
        partcount = geom.PartCount
        partnum = 0
        while partnum < partcount:
            # Step through each vertex in the geometry, handling hole separators
            #
            part = geom.GetPart(partnum)
            p = part.next()
            while p:
                p = part.next()

            # if geometry part has holes, p will be non-null  
            p = part.next()
            if p:
                return True
            partnum += 1

    else:
        # Step through each part of the geometry
        #
        for part in geom:
            # Step through each vertex in the geometry, handling hole separators
            #
            p = part.next()
            while p:
                p = part.next()

            # if geometry part has holes, p will be non-null  
            p = part.next()
            if p:
                return True

    return False

#---------------------------------------------------------------------------------------------------------
# FixSDEInvalidPolygons93()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def FixSDEInvalidPolygons93():
    CPMessage('        - Fixing SDE-invalid polygon geometry', 4)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "FixSDEInvalidPolygons93" 

    Temp_FC = Processing_Variables['ScratchFGDB_Path'] + r'\temp1'
    Union_FC = Processing_Variables['ScratchFGDB_Path'] + r'\temp2'
    for fc in [Temp_FC,Union_FC]:
        if arcpy.Exists(fc):
            while not arcpy.TestSchemaLock(fc):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + fc, 3)
                time.sleep(10)            
            arcpy.Delete_management(fc)

    result = arcpy.GetCount_management(INPUT_LOCATION)
    Number_of_Features = int(result.GetOutput(0))

    CurrentFieldList = [f.name for f in arcpy.ListFields(INPUT_LOCATION)]
    if 'TEIS_ID' not in CurrentFieldList:
        CPMessage('*** Specified input dataset does not contain a TEIS_ID field.',0)
        sys.exit()

    desc = arcpy.Describe(INPUT_LOCATION)
    ShapeFieldName = desc.ShapeFieldName

    Number_Processed = 0
    Number_Repaired = 0
    Nullify_MEAN_ASP_Dict = {}
    dtCalcStart = time.time()
    CPMessage('        - Processing input feature class for ' + str(Number_of_Features) + ' features', 4)
    rows = arcpy.UpdateCursor(INPUT_LOCATION)
    row = rows.Next()
    while row:
        geom = row.getValue(ShapeFieldName)
        if HasHole(geom):
            if HasTouchingHole(geom):
                Number_Repaired += 1
                CPMessage('                - SDE-invalid polygon #' + str(Number_Repaired) + ' detected at TEIS_ID = ' + str(row.getValue("TEIS_ID")),4)
                #CPMessage('                    - Exporting polygon to temp feature class',4)
                arcpy.CreateFeatureclass_management(Processing_Variables['ScratchFGDB_Path'],'temp1','POLYGON','#','DISABLED','DISABLED',Processing_Variables['Spatial_Reference'])

                desc = arcpy.Describe(Temp_FC)
                InsertShapeFieldName = desc.ShapeFieldName
                insert_rows = arcpy.InsertCursor(Temp_FC)
                insert_row = insert_rows.NewRow()
                insert_row.setValue(InsertShapeFieldName, geom)
                insert_rows.insertRow(insert_row)
                del insert_row
                del insert_rows

                CPMessage('                    - Repairing polygon using Union tool',4)
                arcpy.Union_analysis(Temp_FC,Union_FC)
                desc = arcpy.Describe(Union_FC)
                UnionShapeFieldName = desc.ShapeFieldName
                union_row_counter = 0
                union_rows = arcpy.SearchCursor(Union_FC)
                union_row = union_rows.Next()
                while union_row:
                    union_row_counter +=1
                    fixed_geometry = row.getValue(UnionShapeFieldName)
                    union_row = union_rows.Next()
                del union_row
                del union_rows
                if union_row_counter == 1:
                    #CPMessage('                    - Updating geometry of original polygon',4)
                    row.setValue(ShapeFieldName,fixed_geometry)
                    rows.updateRow(row)
                else:
                    CPMessage('                  *** Unexpected Union output. Unable to repair geometry. Investigate this polygon.',4)
                    sys.exit()
                for fc in [Temp_FC,Union_FC]:
                    arcpy.Delete_management(fc)

        Number_Processed += 1
        if Number_Processed == 100 or Number_Processed % 10000 == 0:
            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)
        row = rows.Next()
    dtCalcNow = time.time()
    dtCalcElapsed = dtCalcNow - dtCalcStart
    CPMessage('            - Processed all ' + str(Number_Processed) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed)', 4)
            
    del row
    del rows

    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# FixSDEInvalidPolygons101()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def FixSDEInvalidPolygons101():
    CPMessage('        - Fixing SDE-invalid polygon geometry', 4)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "FixSDEInvalidPolygons101" 

    Temp_FC = Processing_Variables['ScratchFGDB_Path'] + r'\temp1'
    Union_FC = Processing_Variables['ScratchFGDB_Path'] + r'\temp2'
    for fc in [Temp_FC,Union_FC]:
        if arcpy.Exists(fc):
            while not arcpy.TestSchemaLock(fc):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + fc, 3)
                time.sleep(10)            
            arcpy.Delete_management(fc)

    Number_of_Features = int(arcpy.GetCount_management(INPUT_LOCATION).getOutput(0))

    CurrentFieldList = [f.name for f in arcpy.ListFields(INPUT_LOCATION)]
    if 'TEIS_ID' not in CurrentFieldList:
        CPMessage('*** Specified input dataset does not contain a TEIS_ID field.',0)
        sys.exit()

    CursorFields = ['TEIS_ID','SHAPE@']

    Number_Processed = 0
    Number_Repaired = 0
    Nullify_MEAN_ASP_Dict = {}
    dtCalcStart = time.time()
    CPMessage('        - Processing input feature class for ' + str(Number_of_Features) + ' features', 4)
    exec_string = '''with arcpy.da.UpdateCursor(INPUT_LOCATION,CursorFields) as cursor:
    for row in cursor:
        if HasHole(row[1]):
            if HasTouchingHole(row[1]):
                Number_Repaired += 1
                CPMessage('                - SDE-invalid polygon #' + str(Number_Repaired) + ' detected at TEIS_ID = ' + str(row[0]),4)
                #CPMessage('                    - Exporting polygon to temp feature class',4)
                arcpy.CreateFeatureclass_management(Processing_Variables['ScratchFGDB_Path'],'temp1','POLYGON','#','DISABLED','DISABLED',Processing_Variables['Spatial_Reference'])
                with arcpy.da.InsertCursor(Temp_FC,['SHAPE@']) as insert_cursor:
                    insert_cursor.insertRow([row[1]])
                del insert_cursor
                CPMessage('                    - Repairing polygon using Union tool',4)
                arcpy.Union_analysis(Temp_FC,Union_FC)
                union_row_counter = 0
                with arcpy.da.SearchCursor(Union_FC,['SHAPE@']) as union_cursor:
                    for union_row in union_cursor:
                        union_row_counter +=1
                        fixed_geometry = union_row[0]
                if union_row_counter == 1:
                    #CPMessage('                    - Updating geometry of original polygon',4)
                    row[1] = fixed_geometry
                    cursor.updateRow(row)
                else:
                    CPMessage('            ** Unexpected Union output. Unable to repair geometry. Investigate this polygon.',4)
                for fc in [Temp_FC,Union_FC]:
                    arcpy.Delete_management(fc)

        Number_Processed += 1
        if Number_Processed == 100 or Number_Processed % 10000 == 0:
            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)
    dtCalcNow = time.time()
    dtCalcElapsed = dtCalcNow - dtCalcStart
    CPMessage('            - Processed all ' + str(Number_Processed) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed)', 4)'''
            
    exec(exec_string)
    del row
    del cursor

    return(Processing_Variables['PROCESS_STATUS'])


#*********************************************************************************************************
#  Main Program
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# Run Main for ArcGIS 9.3 and 10.0
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain93():

    Initialize()

    SetupWorkspace()
    
    try:
        CPMessage('    - Running RepairGeometry tool', 2)
        arcpy.RepairGeometry_management(INPUT_LOCATION, "DELETE_NULL") #2010-11-16: Jeff K. changed "KEEP_NULL". 2011-03-01: changed it back.
    except:
        CPMessage('** ERROR REPAIRING GEOMETRY. ***', 3)

    FixSDEInvalidPolygons93()

    return Processing_Variables['PROCESS_STATUS']

#---------------------------------------------------------------------------------------------------------
# Run Main for ArcGIS 10.1
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain101():

    Initialize()

    SetupWorkspace()
    
    try:
        CPMessage('    - Running RepairGeometry tool', 2)
        arcpy.RepairGeometry_management(INPUT_LOCATION, "DELETE_NULL") #2010-11-16: Jeff K. changed "KEEP_NULL". 2011-03-01: changed it back.
    except:
        CPMessage('** ERROR REPAIRING GEOMETRY. ***', 3)

    FixSDEInvalidPolygons101()

    return Processing_Variables['PROCESS_STATUS']

#*********************************************************************************************************
#  Execution
#*********************************************************************************************************


# if float(d['Version']) < 9.3:
#     print '** Unsupported ArcGIS version (' + str(d['Version']) + ') running on this machine.'
# elif float(d['Version']) <= 10.0:
#     print '    - ArcGIS 9.3 or 10.0 detected on this machine.'
#     runMain93()
#     d = arcpy.GetInstallInfo('desktop')
if float(d['Version'][:4]) >= 10.1:
    print '    - ArcGIS 10.1 detected on this machine.'
    runMain101()

#---------------------------------------------------------------------------------------------------------
#  Clean-up
#---------------------------------------------------------------------------------------------------------


END_TIME       = time.ctime(time.time())
END_TIME_SEC   = time.time()
END_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
EXECUTION_TIME = END_TIME_SEC - START_TIME_SEC
if EXECUTION_TIME < 120:
    EXECUTION_TIME_STRING = str(int(EXECUTION_TIME)) + ' seconds.'
elif EXECUTION_TIME < 3600:
    EXECUTION_TIME_STRING = str(round(EXECUTION_TIME / 60, 1)) + ' minutes.'
else:
    EXECUTION_TIME_STRING = str(round(EXECUTION_TIME / 3600, 2)) + ' hours.'


CPMessage('    End Time: ' + END_TIME, 2)
CPMessage('    Execution Time: ' + EXECUTION_TIME_STRING, 4)
CPMessage('Complete: ' + PYTHON_SCRIPT, 1)
CPMessage('Next Script: ValidateData' + '/n' +'Preconditions for next step: None ',1)
#*********************************************************************************************************
#  End of Program
#*********************************************************************************************************
