#=========================================================================================================
#                                          StandardizeNoData
#=========================================================================================================
#
# Author(s)         :  Clover Point Cartographics (Mike Neal, Jeff Kruys)
#
# Date              :  2008-06-11
#
# Purpose           :  This script standardizes the representation of null values throughout the input
#                      table. It sets all NULL values in text fields to an empty string, and all invalid
#                      zero values in numeric fields to NULL. If the input is a feature class, the script
#                      also runs the RepairGeometry tool with the DELETE_NULL option, which deletes any
#                      records with null geometry. The input can be a feature class or a table.
#
# Arguments         :  argv(1) = 
#                      argv(2) = 
#
# Example Input     :  X:\fullpath\StandardizeNoData.py X:\fullpath\TEI_Long_Tbl
#
# Outputs           :  No new outputs are created. The input feature class or table is modified.
#
# Dependencies      :  None
#
# Expected Duration :  Approx. 2 hours for 200,000 records
#                      Approx. 6 days for 24 million records
#
# History           :  2012-09-17 (JK)
#                      Rewrote to use cursor-based logic rather than SelectByAttribute and CalculateField.
#                      In each cursor loop, all fields are processed.
#                      Script detects if local machine is running 10.1 and if so, uses faster Data Access
#                      Module cursors. 
#
#                   :  2012-10-12 (JK): Update to CPMessage routine: only append messages to PROCESS_LOG
#                                       if the e-mail function is being used (otherwise it can fill memory).
#                                       Replaced all "gp" references to "arcpy" to save having to test
#                                       current Arc version during every CPMessage call.
#                   :  2012-10-18 (JK): Add test to check if input dataset has geometry before trying to 
#                                       repair geometry.
#                   :  2012-10-26 (JK): Removed the repair geometry function from this script. That now 
#                                       takes place in the new RepairGeometry.py script.
#                   :  2013-05-21 (JK): Script used str() function on text fields to test if field was 
#                                       Null or single space, but this caused error if field value contained
#                                       Unicode characters. Removed the str().
#
#=========================================================================================================
#=========================================================================================================
#  Process Status and Return Codes:
#      -1   - Unexpected error
#       0   - Completed Successful Run
#       1   - In progress
#       100 - Invalid Arguments
#       101 - ESRI License Error
#
#=========================================================================================================

#*********************************************************************************************************
#  Program Initialization
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
#  Python and System Modules
#---------------------------------------------------------------------------------------------------------

from __future__ import division

print '-- Importing Modules'

import sys
import string
import os
import stat # index of constants for os.stat()
import time
import socket
import zipfile

# import re
# import shutil
# import dircache
# import win32com.client
# import cpc_gis_common

#---------------------------------------------------------------------------------------------------------
#  Arguement Validation
#---------------------------------------------------------------------------------------------------------

print '-- Getting Run Parameters'

# Capture/Validate Parameters
#  Test that the right number of paramters were provided and validated
#  Note that 1 must be added because the Python script is an arguement (0) supplied to the
#  Python Environment.


# If long text parameters are required, use:  any_value = arcpy.GetParameterAsText(0)

if len(sys.argv) < 2 or len(sys.argv) > 2 :

    print "\nInvalid number of arguments (" + str(len(sys.argv) - 1) + " supplied) to:"
    print "    " + sys.argv[0] + "\n"
    print " Usage:   " + sys.argv[0] + " [ARG1]"
    print "    [ARG1]   : Feature Class to fix"

    sys.exit(100)

PYTHON_SCRIPT   = sys.argv[0]
INPUT_LOCATION  = sys.argv[1]


#---------------------------------------------------------------------------------------------------------
#  Standard/Global Variables/Constants
#---------------------------------------------------------------------------------------------------------
print '-- Setting Global Variables'

# Global Constants
COMPUTERNAME     = os.environ['COMPUTERNAME']
USERNAME         = os.environ['USERNAME']
START_TIME       = time.ctime(time.time())
START_TIME_SEC   = time.time()
START_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
BEEP             = chr(7)
#s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
#s.connect(("google.com",0))
#IP_ADDRESS       = s.getsockname()[0]
#s.close()
#del s

# Global Variables
#   The Processing_Variables is the collection of messaging and script specific information needed
#   throughout this program.  It is constantly updated, and generally includes all paths and configuration
#   information.  For parameters that are passed in to the script, they should generally be assigned
#   to a Processing_Variable through the Initialize() routine.

Processing_Variables = {}
RunStatus = {}



#---------------------------------------------------------------------------------------------------------
#  CPC Variables/Constants
#---------------------------------------------------------------------------------------------------------




#---------------------------------------------------------------------------------------------------------
#  Set up the ArcGIS Geoprocessing Environment
#---------------------------------------------------------------------------------------------------------

try:
    import arcgisscripting
    arcpy = arcgisscripting.create(93)
except:
    print '**** ArcGIS Failed to initialize a GeoProcessor Object'
    sys.exit(101)

print '-- Checking Script Pre-Conditions'

if arcpy.Exists(INPUT_LOCATION):
    print '    - Input Location Exists:', INPUT_LOCATION
else:
    print ' **** Input Location Does Not Exist:', INPUT_LOCATION
    sys.exit(100)

d = arcpy.GetInstallInfo('desktop')
if float(d['Version'][:4]) >= 10.1:
    print "Arc " +  d['Version'][:4] + " detected on this machine"
    del arcpy
    import arcpy

#---------------------------------------------------------------------------------------------------------
#  Script Specific Variables/Constants
#---------------------------------------------------------------------------------------------------------




#*********************************************************************************************************
#  Routines
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# TemplateRoutine(Input_FC)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TemplateRoutine(Input_FC):
    CPMessage('    - Template Routine', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TemplateRoutine" 

    result = arcpy.GetCount_management(Input_FC)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Calculating individual values for ' + str(Number_of_Features) + ' features.', 4)
    rows = arcpy.SearchCursor(Input_FC)
    row = rows.Next()
    
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features), 4)

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    del row
    del rows

    return(Processing_Variables['PROCESS_STATUS'])

#=========================================================================================================
#  Standard Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# Initialize93()
#
#       Set up all of the paths and variables that we need to run this script.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def Initialize93():
    print '    - Initializing Script'

    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Script_Name'] = os.path.split(PYTHON_SCRIPT)[1]
    Processing_Variables['Log_Directory_Path'] = Processing_Variables['Script_Directory_Path'] + r'\log'

    Processing_Variables['Project_Base_Path']      = os.path.split(Processing_Variables['Script_Directory_Path'])[0]

    
    Processing_Variables['ToolBoxRun'] = 'FALSE'
    # Processing_Variables['SendRunLogToEmail'] = ['mneal@cloverpoint.com', 'chaos_mike@hotmail.com']
    Processing_Variables['Log_Level'] = 4
    Processing_Variables['PROCESS_INFO']   = "RUNNING " + PYTHON_SCRIPT
    Processing_Variables['PROCESS_STATUS'] = 1
    Processing_Variables['PROCESS_LOG'] = []
    Processing_Variables['RunLogFile']   = Processing_Variables['Log_Directory_Path'] + '\\' + Processing_Variables['Script_Name'] + '.' + COMPUTERNAME + '.log'

    #### Adding a FGDB Log will significantly slow down performance.
    # Processing_Variables['Log_FGDB_Table'] = Processing_Variables['Project_Base_Path'] + r"\Operational.gdb\Processing_Log"

    if not arcpy.Exists(Processing_Variables['Log_Directory_Path']):
        os.mkdir(Processing_Variables['Log_Directory_Path'])
        CPMessage(r'Created \Scripts\log folder', 1)
    
    CPMessage('Executing: ' + PYTHON_SCRIPT, 1)
    CPMessage('    Start Time: ' + START_TIME, 2)
    CPMessage('    Python ' + sys.version, 4)
    #CPMessage('    Computer ' + COMPUTERNAME + ' (' + IP_ADDRESS + ')', 4)
    CPMessage('    Computer ' + COMPUTERNAME, 4)
    

    
    # Load required toolboxes...
    # arcpy.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")
    
    # Set the Geoprocessing environment...
    arcpy.OverwriteOutput = 1
    # arcpy.cellSize = "MINOF"
    # arcpy.workspace = INPUT_LOCATION
    # arcpy.scratchworkspace = CPC_LOCAL_TEMP_SPACE
    # arcpy.XYResolution = "" 
    # arcpy.XYDomain = "" 
    # arcpy.MTolerance = "" 
    # arcpy.randomGenerator = "0 ACM599" 
    # arcpy.outputCoordinateSystem = "" 
    # arcpy.ZDomain = "" 
    # arcpy.configKeyword = "" 
    # arcpy.outputZFlag = "Same As Input" 
    # arcpy.qualifiedFieldNames = "true" 
    # arcpy.extent = "DEFAULT" 
    # arcpy.XYTolerance = "" 
    # arcpy.MDomain = "" 
    # arcpy.spatialGrid1 = "0" 
    # arcpy.outputZValue = "" 
    # arcpy.outputMFlag = "Same As Input" 
    # arcpy.geographicTransformations = "" 
    # arcpy.spatialGrid2 = "0" 
    # arcpy.ZResolution = "" 
    # arcpy.spatialGrid3 = "0" 
    # arcpy.MResolution = "" 
    # arcpy.ZTolerance = ""
    
    # arcpy.CreateFolder_management()
    
    
    
    
    # CPC Naming Standard: {variable}_Path is the fully qualified path, otherwise it is just the name of the item.
  
    # Processing Variables
    
    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Working_Directory_Path'] = INPUT_LOCATION
    
    
    
    Processing_Variables['Processing_FGDB'] = 'Processing.gdb'
    Processing_Variables['Processing_Path'] = Processing_Variables['Working_Directory_Path'] + '\\' + Processing_Variables['Processing_FGDB']
    
    
    # Processing_Variables['Densify_Script'] = os.path.split(PYTHON_SCRIPT)[0] + '\\densify.py'
   
    return(1)


#---------------------------------------------------------------------------------------------------------
# Initialize101()
#
#       Set up all of the paths and variables that we need to run this script.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def Initialize101():
    print '    - Initializing Script'

    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Script_Name'] = os.path.split(PYTHON_SCRIPT)[1]
    Processing_Variables['Log_Directory_Path'] = Processing_Variables['Script_Directory_Path'] + r'\log'

    Processing_Variables['Project_Base_Path']      = os.path.split(Processing_Variables['Script_Directory_Path'])[0]

    
    Processing_Variables['ToolBoxRun'] = 'FALSE'
    # Processing_Variables['SendRunLogToEmail'] = ['mneal@cloverpoint.com', 'chaos_mike@hotmail.com']
    Processing_Variables['Log_Level'] = 4
    Processing_Variables['PROCESS_INFO']   = "RUNNING " + PYTHON_SCRIPT
    Processing_Variables['PROCESS_STATUS'] = 1
    Processing_Variables['PROCESS_LOG'] = []
    Processing_Variables['RunLogFile']   = Processing_Variables['Log_Directory_Path'] + '\\' + Processing_Variables['Script_Name'] + '.' + COMPUTERNAME + '.log'

    #### Adding a FGDB Log will significantly slow down performance.
    # Processing_Variables['Log_FGDB_Table'] = Processing_Variables['Project_Base_Path'] + r"\Operational.gdb\Processing_Log"

    if not arcpy.Exists(Processing_Variables['Log_Directory_Path']):
        os.mkdir(Processing_Variables['Log_Directory_Path'])
        CPMessage(r'Created \Scripts\log folder', 1)
    
    CPMessage('Executing: ' + PYTHON_SCRIPT, 1)
    CPMessage('    Start Time: ' + START_TIME, 2)
    CPMessage('    Python ' + sys.version, 4)
    #CPMessage('    Computer ' + COMPUTERNAME + ' (' + IP_ADDRESS + ')', 4)
    CPMessage('    Computer ' + COMPUTERNAME, 4)
    

    
    # Load required toolboxes...
    # arcpy.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")
    
    # Set the Geoprocessing environment...
    arcpy.OverwriteOutput = 1
    # arcpy.cellSize = "MINOF"
    # arcpy.workspace = INPUT_LOCATION
    # arcpy.scratchworkspace = CPC_LOCAL_TEMP_SPACE
    # arcpy.XYResolution = "" 
    # arcpy.XYDomain = "" 
    # arcpy.MTolerance = "" 
    # arcpy.randomGenerator = "0 ACM599" 
    # arcpy.outputCoordinateSystem = "" 
    # arcpy.ZDomain = "" 
    # arcpy.configKeyword = "" 
    # arcpy.outputZFlag = "Same As Input" 
    # arcpy.qualifiedFieldNames = "true" 
    # arcpy.extent = "DEFAULT" 
    # arcpy.XYTolerance = "" 
    # arcpy.MDomain = "" 
    # arcpy.spatialGrid1 = "0" 
    # arcpy.outputZValue = "" 
    # arcpy.outputMFlag = "Same As Input" 
    # arcpy.geographicTransformations = "" 
    # arcpy.spatialGrid2 = "0" 
    # arcpy.ZResolution = "" 
    # arcpy.spatialGrid3 = "0" 
    # arcpy.MResolution = "" 
    # arcpy.ZTolerance = ""
    
    # arcpy.CreateFolder_management()
    
    
    
    
    # CPC Naming Standard: {variable}_Path is the fully qualified path, otherwise it is just the name of the item.
  
    # Processing Variables
    
    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Working_Directory_Path'] = INPUT_LOCATION
    
    
    
    Processing_Variables['Processing_FGDB'] = 'Processing.gdb'
    Processing_Variables['Processing_Path'] = Processing_Variables['Working_Directory_Path'] + '\\' + Processing_Variables['Processing_FGDB']
    
    
    # Processing_Variables['Densify_Script'] = os.path.split(PYTHON_SCRIPT)[0] + '\\densify.py'
   
    return(1)




#---------------------------------------------------------------------------------------------------------
# SetupWorkspace93()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SetupWorkspace93():
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "SetupWorkspace"  

    CPMessage('    - Setting Up Workspace', 2)

  
    if not arcpy.Exists(Processing_Variables['Processing_Path']):
        CPMessage('        -- Creating Processing FGDB', 4)
        try:
            arcpy.CreateFileGDB_management(Processing_Variables['Working_Directory_Path'], Processing_Variables['Processing_FGDB'])
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created Processing FGDB"
        except:
            Processing_Variables['PROCESS_STATUS'] = -1
            Processing_Variables['PROCESS_INFO']   = "Error Creating Processing FGDB"
            CPMessage('** Can not find or create ' + Processing_Variables['Processing_Path'], 0)


    # Create/overwrite the temporary table/FC
    # CPMessage('        -- Initializing Temporary Table.', 4)
    # if arcpy.Exists(Temp_FC):
    #     while not arcpy.TestSchemaLock(Temp_FC):
    #         CPMessage('** Schema Locked - waiting 10 seconds: ' + Temp_FC, 3)
    #         time.sleep(10)
    #     arcpy.Delete_management(Temp_FC)
    # arcpy.CreateTable_management(os.path.split(Temp_FC)[0], os.path.split(Temp_FC)[1], Template_Table)



    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# SetupWorkspace101()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SetupWorkspace101():
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "SetupWorkspace"  

    CPMessage('    - Setting Up Workspace', 2)

  
    if not arcpy.Exists(Processing_Variables['Processing_Path']):
        CPMessage('        -- Creating Processing FGDB', 4)
        try:
            arcpy.CreateFileGDB_management(Processing_Variables['Working_Directory_Path'], Processing_Variables['Processing_FGDB'])
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created Processing FGDB"
        except:
            Processing_Variables['PROCESS_STATUS'] = -1
            Processing_Variables['PROCESS_INFO']   = "Error Creating Processing FGDB"
            CPMessage('** Can not find or create ' + Processing_Variables['Processing_Path'], 0)


    # Create/overwrite the temporary table/FC
    # CPMessage('        -- Initializing Temporary Table.', 4)
    # if arcpy.Exists(Temp_FC):
    #     while not arcpy.TestSchemaLock(Temp_FC):
    #         CPMessage('** Schema Locked - waiting 10 seconds: ' + Temp_FC, 3)
    #         time.sleep(10)
    #     arcpy.Delete_management(Temp_FC)
    # arcpy.CreateTable_management(os.path.split(Temp_FC)[0], os.path.split(Temp_FC)[1], Template_Table)



    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# CPMessage(Message)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CPMessage(Message, MessageLevel):
    # Based on the logging level, add messages to the message queues.
    #   Text Run Log   : Implimented
    #   Console        : Implimented
    #   GeoProcessor   : Implimented
    #   Processing Log : Implimented (with e-mail)
    #   SQL Database   : Not Implimented ####
    #   FGDB           : Implimented
    
    # Message Levels:
    #   0    : Fatal Errors
    #   1    : Top Level Program
    #   2    : Main routines and events
    #   3    : Warnings
    #   4    : General Information
    #   5    : Debug level detail
    
    # Send to the screen
    if MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToScreen(Message, MessageLevel)    
    
    # Send to the Processing Log
    if 'SendRunLogToEmail' in Processing_Variables:
        Processing_Variables['PROCESS_LOG'].append((Message, MessageLevel))
    
    # Send to the Log File
    if 'RunLogFile' in Processing_Variables and MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToFile(Message, Processing_Variables['RunLogFile'])

    # Send to the GeoProcessor Message Queue
    #### At ArcGIS 9.3 SP1 there is a bug that also causes any GeoProcessor messages to be echoed to the console.
    #    Added the ToolBoxRun parameter to allow the script to add messages to the GeoProcessor Messaging properly
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and Processing_Variables['ToolBoxRun'] == 'TRUE':
        if MessageLevel == 0:
            arcpy.AddError(Message)
        if MessageLevel == 1 or MessageLevel == 2:
            arcpy.AddMessage(Message)
        if MessageLevel == 3:
            arcpy.AddWarning(Message)

    # Send to the FGDB Processing Log
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and 'Log_FGDB_Table' in Processing_Variables:
        rows = arcpy.InsertCursor(Processing_Variables['Log_FGDB_Table'])

        row = rows.NewRow()
        row.SetValue("Computer_Name", COMPUTERNAME)
        row.SetValue("User_Name", USERNAME)
        row.SetValue("Script_Name", PYTHON_SCRIPT)
        row.SetValue("Event_Time", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
        row.SetValue("Message_Level", MessageLevel)
        row.SetValue("Message", Message)
        row.SetValue("Start_Time", START_TIME_SQL)

        rows.InsertRow(row)
        del row
        del rows
    

    return(1)


#---------------------------------------------------------------------------------------------------------
# WriteOutputToScreen(Output_Comment, Header_Style):
#
#   Prints the Output_Comment to the screen with the appropriate formatting
#       0 : "******" Used for Error Sections
#       1 : "======" Used for the Program's Header and Footer
#       2 : "------" Used for New Sections
#       3+:          No delimitation
#   
#---------------------------------------------------------------------------------------------------------
def WriteOutputToScreen(Output_Comment, Header_Style):
    if Header_Style == 0:
        # Errors get beeps, too!
        print BEEP * 3 + "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 2:
        print "\n" + "-" * 79 + "\n"
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"
        
    #print Output_Comment

    arcpy.AddMessage(Output_Comment)
    
    if Header_Style == 0:
        print "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"


#---------------------------------------------------------------------------------------------------------
# WriteOutputToFile(Output_Comment, Run_Log)
#   Record to Log 
#
#
#---------------------------------------------------------------------------------------------------------
def WriteOutputToFile(Output_Comment, Run_Log):
# Run logs are not compatible with concurrency - only one process is expected in debug mode at a time!
# First try to append to an existing file - if that fails create a new file.
    ReturnValue = 1
    try:
        if os.path.isfile(Run_Log) == 1:
            # try to open up the fileName 
            file = open(Run_Log,'a') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Appended comment to the file"
        else:
            # try to open up the fileName 
            file = open(Run_Log,'w') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created new file and put in the comment"
    except:
        # All failures are unexpected in this case...  
        # Maybe an invalid drive name, read only file or something?
        Processing_Variables['PROCESS_STATUS'] = -1
        Processing_Variables['PROCESS_INFO']   = "**ERROR: Can't write to file: " + Run_Log   
        print BEEP + BEEP + BEEP
        ReturnValue = -1
    return(ReturnValue)
        

#---------------------------------------------------------------------------------------------------------
# SendRunLogToEmail()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SendRunLogToEmail():
    import smtplib
    try:
        Seconds_Since_Last_Message = time.time() - time.mktime(time.strptime(RunStatus['LastStatusEmailSent'], "%Y-%m-%d %H:%M:%S" ))
    except:
        # No message ever sent, or bad data, so send one now.
        Seconds_Since_Last_Message = -1

    if Seconds_Since_Last_Message < 120:
        Last_Message_Time = str(int(Seconds_Since_Last_Message)) + ' seconds.'
    elif Seconds_Since_Last_Message < 3600:
        Last_Message_Time = str(round(Seconds_Since_Last_Message / 60, 1)) + ' minutes.'
    else:
        Last_Message_Time = str(round(Seconds_Since_Last_Message / 3600, 2)) + ' hours.'


    if int(RunStatus['MinimumMessageInterval']) < 120:
        Minimum_Message_Time = str(int(RunStatus['MinimumMessageInterval'])) + ' seconds.'
    elif Seconds_Since_Last_Message < 3600:
        Minimum_Message_Time = str(round(int(RunStatus['MinimumMessageInterval']) / 60, 1)) + ' minutes.'
    else:
        Minimum_Message_Time = str(round(int(RunStatus['MinimumMessageInterval']) / 3600, 2)) + ' hours.'


    CPMessage('    Time since last message sent: ' + Last_Message_Time, 4)        
    CPMessage('    Minimum message interval is : ' + Minimum_Message_Time, 4)
    
    for mailto in Processing_Variables['SendRunLogToEmail']:
        if int(RunStatus['MinimumMessageInterval']) < Seconds_Since_Last_Message or Seconds_Since_Last_Message == -1:
            CPMessage('  * Sending log to ' + mailto, 4)
            mailfrom = 'GIS_Python@cloverpoint.com'
            # mailto = 'mneal@cloverpoint.com'
            mailsubject = 'Run Log: ' + os.path.split(PYTHON_SCRIPT)[1] + ' (' + COMPUTERNAME + '\\' + USERNAME + ')'
            mailbody = "This is an automatic run log.\n\n"
            
            
            mailbody = mailbody + '\n\n  ** Note: The minimum time between messages for this monitor is: ' + Minimum_Message_Time + '.\n'

            for LogEntry in Processing_Variables['PROCESS_LOG']:
                mailbody = mailbody + '\n' + LogEntry[0]

            mailbody = mailbody + '\n\nEnd of Line.\n'

            mailmsg = ("From: " + mailfrom + "\r\nTo: " + mailto + "\r\nSubject: " + mailsubject + "\r\n" + mailbody)

            mailserver = smtplib.SMTP('shawmail.gv.shawcable.net')
            # mailserver = smtplib.SMTP('localhost')

            # SMTP.login(user, password)

            mailserver.sendmail(mailfrom, mailto, mailmsg)
            mailserver.quit()
            RunStatus['LastStatusEmailSent'] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        else:
            CPMessage('  ** NOT Sending log to ' + mailto + ' due to short interval since last message.', 3)
            
    
    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# SanitizeElapsedTime(number_of_seconds)
#
# routine to convert a raw number of seconds into a human-friendly statement of duration
#---------------------------------------------------------------------------------------------------------

def SanitizeElapsedTime(dtInput):

    if dtInput < 120:
        strElapsedTime = str(int(dtInput)) + ' seconds.'
        
    elif dtInput < 3600:
        strElapsedTime = str(int(dtInput)) + ' seconds (' + str(round(dtInput / 60, 1)) + ' minutes).'
        
    else:
        strElapsedTime = str(int(dtInput)) + ' seconds (' + str(round(dtInput / 3600, 2)) + ' hours).'
        
    return strElapsedTime

#=========================================================================================================
#  Special Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# FixNumericAndStringFields93()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def FixNumericAndStringFields93():
    CPMessage('        - Fixing null values in numeric and string fields', 4)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "FixNumericAndStringFields93" 

    result = arcpy.GetCount_management(INPUT_LOCATION)
    Number_of_Features = int(result.GetOutput(0))

    CurrentFieldList = arcpy.ListFields(INPUT_LOCATION)


    Number_Processed = 0
    rows = arcpy.UpdateCursor(INPUT_LOCATION)
    row = rows.Next()
    Nullify_MEAN_ASP_Dict = {}
    dtCalcStart = time.time()
    CPMessage('        - Processing input table for all fields except MEAN_ASP', 4)
    while row:
        update_flag = False

        # handle the slope fields separately from the rest of the fields: if one of a slope field pair is zero and the other is greater than zero, keep the zero; otherwise nullify it
        SlopeFieldPairs = [['SLPLL_1','SLPUL_1'],['SLPLL_2','SLPUL_2'],['SLPLL_1A','SLPUL_1A'],['SLPLL_1B','SLPUL_1B'],['SLPLL_2A','SLPUL_2A'],['SLPLL_2B','SLPUL_2B'],['SLPLL_3A','SLPUL_3A'],['SLPLL_3B','SLPUL_3B']]
        for SlopeFieldPair in SlopeFieldPairs:
            if SlopeFieldPair[0] in [f.name for f in CurrentFieldList] and SlopeFieldPair[1] in [f.name for f in CurrentFieldList]:
                LowerSlope = str(row.getValue(SlopeFieldPair[0]))
                UpperSlope = str(row.getValue(SlopeFieldPair[1]))
                if LowerSlope == '0' and UpperSlope == '0':
                    row.setNull(LowerSlope)
                    row.setNull(UpperSlope)
                    update_flag = True
                elif LowerSlope == 'None' and UpperSlope == '0':
                    row.setNull(UpperSlope)
                    update_flag = True
                elif LowerSlope == '0' and UpperSlope == 'None':
                    row.setNull(LowerSlope)
                    update_flag = True

        for FixItem in CurrentFieldList:
            if FixItem.name in ['TEIS_ID','SLPLL_1', 'SLPUL_1', 'SLPLL_2', 'SLPUL_2', 'SLPLL_1A', 'SLPUL_1A', 'SLPLL_1B', 'SLPUL_1B', 'SLPLL_2A', 'SLPUL_2A', 'SLPLL_2B', 'SLPUL_2B', 'SLPLL_3A', 'SLPUL_3A', 'SLPLL_3B', 'SLPUL_3B']:
                pass  ## zero can be valid in these numeric fields

            elif FixItem.name == 'MEAN_ASP':
                # in this cursor loop, build a dictionary of unique MEAN_ASP values for each BAPID; later, for each BAPID, if all MEAN_ASP's are zero or 
                # null, we will nullify them, but if some are zeros and others are actual values, we will not modify any of them for that BAPID
                BAPID = row.getValue("BAPID")
                if BAPID not in Nullify_MEAN_ASP_Dict:
                    Nullify_MEAN_ASP_Dict[BAPID] = True
                MEAN_ASP = str(row.getValue("MEAN_ASP"))
                if MEAN_ASP not in ['0','None']:
                    Nullify_MEAN_ASP_Dict[BAPID] = False

            # fix SDEC-based numeric fields 'TREE_C1', 'TREE_C2', 'TREE_C3', 'SHRUB_C1', 'SHRUB_C2', 'SHRUB_C3' (set to null only if SDEC_x is also zero or null)
            elif FixItem.name in ['TREE_C1','SHRUB_C1','TREE_C2','SHRUB_C2','TREE_C3','SHRUB_C3']:
                FixItemValue = row.getValue(FixItem.name)
                if FixItemValue == 0:
                    SDEC_fieldname = 'SDEC_' + FixItem.name[-1:]
                    SDEC = str(row.getValue(SDEC_fieldname))
                    if SDEC not in ['1','2','3','4','5','6','7','8','9','10']:
                        row.setNull(FixItem.name)
                        update_flag = True

            elif FixItem.Type in ['Integer', 'SmallInteger']:
                FixItemValue = str(row.getValue(FixItem.name))
                if FixItemValue == '0':
                    row.setNull(FixItem.name)
                    update_flag = True

            elif FixItem.Type not in ['SmallInteger', 'Integer', 'Single', 'Double', 'OID', 'Geometry','Date']:
                FixItemValue = row.getValue(FixItem.name)   ## normally we would "str()" this value, but POLY_COM sometimes contains unicode characters which causes error
                if FixItemValue in [None,' ']:
                    row.setValue(FixItem.name, '')
                    update_flag = True

        if update_flag:
            rows.UpdateRow(row)

        Number_Processed += 1
        if Number_Processed == 100 or (0 < Number_Processed <= 10000 and Number_Processed % 1000 == 0) or (10000 < Number_Processed and Number_Processed % 10000 == 0):
            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)

        row = rows.Next()
       
    del row
    del rows

    CurrentFieldNameList = []
    for field in CurrentFieldList:
        CurrentFieldNameList.append(field.name)
    if 'MEAN_ASP' in CurrentFieldNameList:
        Number_Processed = 0
        rows = arcpy.UpdateCursor(INPUT_LOCATION)
        row = rows.Next()
        dtCalcStart = time.time()
        CPMessage('        - Processing input table for MEAN_ASP field', 4)
        while row:
            update_flag = False
            BAPID = row.getValue("BAPID")
            # check the dictionary we built in the previous cursor loop for whether MEAN_ASP should be nullified for this BAPID
            if Nullify_MEAN_ASP_Dict[BAPID]:
                row.setNull("MEAN_ASP")
                rows.UpdateRow(row)
            Number_Processed += 1
            if Number_Processed == 100 or (0 < Number_Processed <= 10000 and Number_Processed % 1000 == 0) or (10000 < Number_Processed and Number_Processed % 10000 == 0):
                dtCalcNow = time.time()
                dtCalcElapsed = dtCalcNow - dtCalcStart
                dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
                CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)

            row = rows.Next()

        del row
        del rows


    return(Processing_Variables['PROCESS_STATUS'])



#---------------------------------------------------------------------------------------------------------
# FixNumericAndStringFields101()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def FixNumericAndStringFields101():
    CPMessage('        - Fixing null values in numeric and string fields', 4)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "FixNumericAndStringFields101" 

    Number_of_Features = int(arcpy.GetCount_management(INPUT_LOCATION).getOutput(0))

    CurrentFieldList = arcpy.ListFields(INPUT_LOCATION)
    CurrentFieldNameList = []
    for CurrentField in CurrentFieldList:
        if CurrentField.type not in ['Blob','Geometry','Guid','OID','Raster','Date']:
            CurrentFieldNameList.append(str(CurrentField.name))

    FieldName_Dict = {}
    x = 0
    for CurrentFieldName in CurrentFieldNameList:
        FieldName_Dict[CurrentFieldName] = x
        x += 1

    Number_Processed = 0
    Nullify_MEAN_ASP_Dict = {}
    dtCalcStart = time.time()
    CPMessage('        - Processing input table for all fields except MEAN_ASP', 4)
    exec_string = '''with arcpy.da.UpdateCursor(INPUT_LOCATION, CurrentFieldNameList) as cursor:
    for row in cursor:
        update_flag = False

        # handle the slope fields separately from the rest of the fields: if one of a slope field pair is zero and the other is greater than zero, keep the zero; otherwise nullify the zero(es)
        SlopeFieldPairs = [['SLPLL_1','SLPUL_1'],['SLPLL_2','SLPUL_2'],['SLPLL_1A','SLPUL_1A'],['SLPLL_1B','SLPUL_1B'],['SLPLL_2A','SLPUL_2A'],['SLPLL_2B','SLPUL_2B'],['SLPLL_3A','SLPUL_3A'],['SLPLL_3B','SLPUL_3B']]
        for SlopeFieldPair in SlopeFieldPairs:
            if SlopeFieldPair[0] in CurrentFieldNameList and SlopeFieldPair[0] in CurrentFieldNameList:
                LowerSlope = str(row[FieldName_Dict[SlopeFieldPair[0]]])
                UpperSlope = str(row[FieldName_Dict[SlopeFieldPair[1]]])
                if LowerSlope == '0' and UpperSlope == '0':
                    row[FieldName_Dict[SlopeFieldPair[0]]] = None
                    row[FieldName_Dict[SlopeFieldPair[1]]] = None
                    update_flag = True
                elif LowerSlope == 'None' and UpperSlope == '0':
                    row[FieldName_Dict[SlopeFieldPair[1]]] = None
                    update_flag = True
                elif LowerSlope == '0' and UpperSlope == 'None':
                    row[FieldName_Dict[SlopeFieldPair[0]]] = None
                    update_flag = True

        for FixItem in CurrentFieldList:
            if FixItem.name in ['TEIS_ID','SLPLL_1', 'SLPUL_1', 'SLPLL_2', 'SLPUL_2', 'SLPLL_1A', 'SLPUL_1A', 'SLPLL_1B', 'SLPUL_1B', 'SLPLL_2A', 'SLPUL_2A', 'SLPLL_2B', 'SLPUL_2B', 'SLPLL_3A', 'SLPUL_3A', 'SLPLL_3B', 'SLPUL_3B']:
                pass  ## don't (further) modify these fields

            if FixItem.name == 'MEAN_ASP':
                # in this cursor loop, build a dictionary of unique MEAN_ASP values for each BAPID; in the next cursor loop (below), for each BAPID, if all MEAN_ASP's are zero or 
                # null, we will nullify them, but if some are zeros and others are actual values, we will not modify any of them for that BAPID
                BAPID = row[FieldName_Dict['BAPID']]
                if BAPID not in Nullify_MEAN_ASP_Dict:
                    Nullify_MEAN_ASP_Dict[BAPID] = True
                MEAN_ASP = str(row[FieldName_Dict['MEAN_ASP']])
                if MEAN_ASP not in ['0','None']:
                    Nullify_MEAN_ASP_Dict[BAPID] = False

            # fix SDEC-based numeric fields 'TREE_C1', 'TREE_C2', 'TREE_C3', 'SHRUB_C1', 'SHRUB_C2', 'SHRUB_C3' (set to null only if SDEC_x is also zero or null)
            elif FixItem.name in ['TREE_C1','SHRUB_C1','TREE_C2','SHRUB_C2','TREE_C3','SHRUB_C3']:
                FixItemValue = row[FieldName_Dict[str(FixItem.name)]]
                if FixItemValue == 0:
                    SDEC_fieldname = 'SDEC_' + FixItem.name[-1:]
                    SDEC = str(row[FieldName_Dict[SDEC_fieldname]])
                    if SDEC not in ['1','2','3','4','5','6','7','8','9','10']:
                        row[FieldName_Dict[str(FixItem.name)]] = None
                        update_flag = True

            elif FixItem.type in ['Integer', 'SmallInteger']:
                FixItemValue = str(row[FieldName_Dict[str(FixItem.name)]])
                if FixItemValue == '0':
                    row[FieldName_Dict[str(FixItem.name)]] = None
                    update_flag = True

            elif FixItem.type not in ['SmallInteger', 'Integer', 'Single', 'Double', 'OID', 'Geometry','Date']:
                if row[FieldName_Dict[str(FixItem.name)]] in [' ',None]:   ## normally we would "str()" this value, but POLY_COM sometimes contains unicode characters which causes error
                    row[FieldName_Dict[str(FixItem.name)]] = ''
                    update_flag = True

        if update_flag:
            cursor.updateRow(row)

        Number_Processed += 1
        if Number_Processed == 100 or (0 < Number_Processed <= 10000 and Number_Processed % 1000 == 0) or (10000 < Number_Processed and Number_Processed % 10000 == 0):
            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)'''
            
    exec(exec_string)
    del row
    del cursor

    if 'MEAN_ASP' in CurrentFieldNameList:
        Number_Processed = 0
        dtCalcStart = time.time()
        CPMessage('        - Processing input table for MEAN_ASP field', 4)
        exec_string = '''with arcpy.da.UpdateCursor(INPUT_LOCATION, CurrentFieldNameList) as cursor:
    for row in cursor:
        update_flag = False
        BAPID = row[FieldName_Dict['BAPID']]
        # check the dictionary we built in the previous cursor loop for whether MEAN_ASP should be nullified for this BAPID
        if Nullify_MEAN_ASP_Dict[BAPID]:
            row[FieldName_Dict['MEAN_ASP']] = None
            cursor.updateRow(row)
        Number_Processed += 1
        if Number_Processed == 100 or (0 < Number_Processed <= 10000 and Number_Processed % 1000 == 0) or (10000 < Number_Processed and Number_Processed % 10000 == 0):
            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features) + ' records (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)'''

        exec(exec_string)
        del row
        del cursor


    return(Processing_Variables['PROCESS_STATUS'])


#*********************************************************************************************************
#  Main Program
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# Run Main for ArcGIS 9.3 and 10.0
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain93():

    Initialize93()

    FixNumericAndStringFields93()

    return Processing_Variables['PROCESS_STATUS']

#---------------------------------------------------------------------------------------------------------
# Run Main for ArcGIS 10.1
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain101():

    Initialize101()

    FixNumericAndStringFields101()

    return Processing_Variables['PROCESS_STATUS']

#*********************************************************************************************************
#  Execution
#*********************************************************************************************************


d = arcpy.GetInstallInfo('desktop')
if len(d['Version'].split('.')) > 2:
    dv = d['Version'].rsplit('.',len(d['Version'].split('.'))-2)[0]
    del arcpy
    import arcpy
    runMain101()
else:
    dv = d['Version']
    if float(dv) <10.1:
        runMain93()
    else:
        print '** Unsupported ArcGIS version (' + str(d['Version']) + ') running on this machine.'




#---------------------------------------------------------------------------------------------------------
#  Clean-up
#---------------------------------------------------------------------------------------------------------


END_TIME       = time.ctime(time.time())
END_TIME_SEC   = time.time()
END_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
EXECUTION_TIME = END_TIME_SEC - START_TIME_SEC
if EXECUTION_TIME < 120:
    EXECUTION_TIME_STRING = str(int(EXECUTION_TIME)) + ' seconds.'
elif EXECUTION_TIME < 3600:
    EXECUTION_TIME_STRING = str(round(EXECUTION_TIME / 60, 1)) + ' minutes.'
else:
    EXECUTION_TIME_STRING = str(round(EXECUTION_TIME / 3600, 2)) + ' hours.'


CPMessage('    End Time: ' + END_TIME, 2)
CPMessage('    Execution Time: ' + EXECUTION_TIME_STRING, 4)
CPMessage('Complete: ' + PYTHON_SCRIPT, 1)
CPMessage('Next Script: RepairGeometry (if needed) or ValidateData' + '/n' +'Preconditions for next step: RepairGeometry: None ValidateData: None ',1)


#*********************************************************************************************************
#  End of Program
#*********************************************************************************************************
