#=========================================================================================================
#                                      ValidateUserDefined
#=========================================================================================================
#
# Author(s)         :  Clover Point Cartographics (Mike Neal)
#
# Date              :  2009-04-08
#
# Purpose           :  This script uses domains and ranges defined by the user in TEI_Usr_Dfn_Fields to
#                      validate the data in TEI_Usr_Dfn_Data. The user will have to populate both tables
#                      before executing this script.
#
# Arguments         :  argv(1) = INPUT_DATA_TABLE       (User-Defined Data table to be validated)
#                      argv(2) = INPUT_FIELD_DEF_TABLE  (User-Defined Field Definitions table to be validated against)
#
# Example Input     :  X:\fullpath\ValidateUserDefined.py X:\fullpath\TEI_User_Defined_Fields X:\fullpath\TEI_User_Defined_Field_Def
#
# Outputs           :  Three new tables in the Operational_Data.gdb file geodatabase, listing the errors
#                      found in the TEI_Usr_Dfn_Data input table.
#
#                           TEI_Usr_Dfn_Data_DomainErrors
#                           TEI_Usr_Dfn_Data_MandatoryFieldErrors
#                           TEI_Usr_Dfn_Data_RangeErrors
#
#                      In each output table the FEATURE_ID field should be used to link to the TEIS_ID
#                      field of the TEI_Usr_Dfn_Data table.
#              
# Dependencies      :  Must have TEI_Usr_Dfn_Fields and TEI_Usr_Dfn_Data in the Operational_Data.gdb file
#                      geodatabase.
#              
# Expected Duration :  < 5 minutes
#
# History           :  2012-10-12 (JK): Update to CPMessage routine: only append messages to PROCESS_LOG
#                                       if the e-mail function is being used (otherwise it can fill memory)
#
#=========================================================================================================
#=========================================================================================================
#  Process Status and Return Codes:
#      -1   - Unexpected error
#       0   - Completed Successful Run
#       1   - In progress
#       100 - Invalid Arguments
#       101 - ESRI License Error
#
#=========================================================================================================

#*********************************************************************************************************
#  Program Initialization
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
#  Python and System Modules
#---------------------------------------------------------------------------------------------------------

from __future__ import division

import sys
import string
import os
import time


#---------------------------------------------------------------------------------------------------------
#  Arguement Validation
#---------------------------------------------------------------------------------------------------------

# Capture/Validate Parameters
#  Test that the right number of paramters were provided and validated
#  Note that 1 must be added because the Python script is an arguement (0) supplied to the
#  Python Environment.


# If long text parameters are required, use:  any_value = gp.GetParameterAsText(0)

if len(sys.argv) < 3 or len(sys.argv) > 3 :

    print "\nInvalid number of arguments (" + str(len(sys.argv) - 1) + " supplied) to:"
    print "    " + sys.argv[0] + "\n"
    print " Usage:   " + sys.argv[0] + " [INPUT_UDF_TABLE] [INPUT_UDF_DEF_TABLE]"
    print "    [INPUT_UDF_TABLE]      : User-Defined Data table to be validated\n"
    print "    [INPUT_UDF_DEF_TABLE]  : User-Defined Field Definitions table to be validated against\n"

    sys.exit(100)

PYTHON_SCRIPT         = sys.argv[0]
INPUT_DATA_TABLE      = sys.argv[1]
INPUT_FIELD_DEF_TABLE = sys.argv[2]

#---------------------------------------------------------------------------------------------------------
#  Standard/Global Variables/Constants
#---------------------------------------------------------------------------------------------------------

# Global Constants
COMPUTERNAME     = os.environ['COMPUTERNAME']
USERNAME         = os.environ['USERNAME']
START_TIME       = time.ctime(time.time())
START_TIME_SEC   = time.time()
START_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
BEEP             = chr(7)


# Global Variables
#   The Processing_Variables is the collection of messaging and script specific information needed
#   throughout this program.  It is constantly updated, and generally includes all paths and configuration
#   information.  For parameters that are passed in to the script, they should generally be assigned
#   to a Processing_Variable through the Initialize() routine.

Processing_Variables = {}



#---------------------------------------------------------------------------------------------------------
#  CPC Variables/Constants
#---------------------------------------------------------------------------------------------------------




#---------------------------------------------------------------------------------------------------------
#  Set up the ArcGIS Geoprocessing Environment
#---------------------------------------------------------------------------------------------------------



# For ArcGIS 9.3:
try:
    import arcgisscripting
    gp = arcgisscripting.create(9.3)

# Get any required extensions
# if gp.CheckExtension("spatial") == "available":
#     gp.CheckOutExtension("spatial")
# else:
#     sys.exit(101)

except:
    print '**** ArcGIS Failed to initialize a GeoProcessor Object'
    sys.exit(101)




#---------------------------------------------------------------------------------------------------------
#  Script Specific Variables/Constants
#---------------------------------------------------------------------------------------------------------



#*********************************************************************************************************
#  Routines
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# TemplateRoutine(Input_FC)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TemplateRoutine(Input_FC):
    CPMessage('    - Template Routine', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TemplateRoutine" 

    result = gp.GetCount_management(Input_FC)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Calculating individual values for' + str(Number_of_Features) + ' features.', 4)
    rows = gp.SearchCursor(Input_FC)
    row = rows.Next()
    
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    del row
    del rows

    return(Processing_Variables['PROCESS_STATUS'])

#=========================================================================================================
#  Standard Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# Initialize()
#
#       Set up all of the paths and variables that we need to run this script.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def Initialize():
    print '    - Initializing Script'
    
    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Script_Name'] = os.path.split(PYTHON_SCRIPT)[1]
    Processing_Variables['Log_Directory_Path'] = Processing_Variables['Script_Directory_Path'] + r'\log'

    Processing_Variables['Project_Base_Path']      = os.path.split(Processing_Variables['Script_Directory_Path'])[0]
    
    Processing_Variables['ToolBoxRun'] = 'FALSE'
    # Processing_Variables['SendRunLogToEmail'] = ['mneal@cloverpoint.com', 'chaos_mike@hotmail.com']
    Processing_Variables['Log_Level'] = 4
    Processing_Variables['PROCESS_INFO']   = "RUNNING " + PYTHON_SCRIPT
    Processing_Variables['PROCESS_STATUS'] = 1
    Processing_Variables['PROCESS_LOG'] = []
    Processing_Variables['MessageQueueFile']   = Processing_Variables['Log_Directory_Path'] + '\\' + Processing_Variables['Script_Name'] + '.' + COMPUTERNAME + '.log'

    #### Adding a FGDB Log will significantly slow down performance.
    # Processing_Variables['Log_FGDB_Table'] = Processing_Variables['Project_Base_Path'] + r"\Operational.gdb\Processing_Log"

    if not gp.Exists(Processing_Variables['Log_Directory_Path']):
        os.mkdir(Processing_Variables['Log_Directory_Path'])
        CPMessage(r'Created \Scripts\log folder', 1)

    CPMessage('Executing: ' + PYTHON_SCRIPT, 1)
    CPMessage('    Start Time: ' + START_TIME, 2)
    CPMessage('    Python ' + sys.version, 4)

    # Load required toolboxes...
    # gp.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")
    
    # Set the Geoprocessing environment...
    gp.OverwriteOutput = 1
    # gp.cellSize = "MINOF"
    # gp.workspace = INPUT_LOCATION
    
    
    
    # CPC Naming Standard: {variable}_Path is the fully qualified path, otherwise it is just the name of the item.
  
    # Processing Variables
   
    Processing_Variables['OperationalFGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Operational_Data.gdb'
    Processing_Variables['Feature_Dataset_Path']   = Processing_Variables['OperationalFGDB_Path'] + r'\TEIS_Feature_Dataset'
    Processing_Variables['Scratch_Path']           = Processing_Variables['Project_Base_Path'] + r'\Scratch'
    Processing_Variables['ScratchFGDB_Path']       = Processing_Variables['Scratch_Path'] + r'\Scratch.gdb'
    gp.scratchworkspace = Processing_Variables['ScratchFGDB_Path']
    
    Processing_Variables['Tools_and_Templates_FGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Tools_and_Templates.gdb'
    
    Processing_Variables['TEIS_Metadata_Items']    = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\TEIS_Metadata_Items'
    Processing_Variables['TEIS_Metadata_Tables']   = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\TEIS_Metadata_Tables'

    Processing_Variables['Template_ErrorTable']    = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\Template_ErrorTable'
  
    Processing_Variables['DomainErrorTable']         = INPUT_DATA_TABLE + r'_DomainErrors'
    Processing_Variables['RangeErrorTable']          = INPUT_DATA_TABLE + r'_RangeErrors'
    Processing_Variables['MandatoryFieldErrorTable'] = INPUT_DATA_TABLE + r'_MandatoryFieldErrors'
    Processing_Variables['User_Defined_Fields']      = INPUT_DATA_TABLE
    Processing_Variables['User_Defined_Field_Def']   = INPUT_FIELD_DEF_TABLE

    if gp.Exists(INPUT_DATA_TABLE):
        CPMessage(' - Input Exists:' + INPUT_DATA_TABLE, 4)
    else:
        CPMessage('** ERROR: INPUT DOES NOT EXIST: ' + INPUT_DATA_TABLE, 0)
        sys.exit(100)

    if gp.Exists(INPUT_FIELD_DEF_TABLE):
        CPMessage(' - Input Exists:' + INPUT_FIELD_DEF_TABLE, 4)
    else:
        CPMessage('** ERROR: INPUT DOES NOT EXIST: ' + INPUT_FIELD_DEF_TABLE, 0)
        sys.exit(100)
   
    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# CPMessage(Message)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CPMessage(Message, MessageLevel):
    # Based on the logging level, add messages to the message queues.
    #   Text Run Log   : Implimented
    #   Console        : Implimented
    #   GeoProcessor   : Implimented
    #   Processing Log : Implimented (with e-mail)
    #   FGDB           : Implimented
    
    # Message Levels:
    #   0    : Fatal Errors
    #   1    : Top Level Program
    #   2    : Main routines and events
    #   3    : Warnings
    #   4    : General Information
    #   5    : Debug level detail
    
    # Send to the screen
    if MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToScreen(Message, MessageLevel)    
    
        # Send to the Processing Log
        #### The log is extremely large with this script and may overflow memory if all levels are appended.
        if 'SendRunLogToEmail' in Processing_Variables:
            Processing_Variables['PROCESS_LOG'].append((Message, MessageLevel))
    
    # Send to the Log File
    if 'MessageQueueFile' in Processing_Variables and MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToFile(Message, Processing_Variables['MessageQueueFile'])

    # Send to the GeoProcessor Message Queue
    #### At ArcGIS 9.3 SP1 there is a bug that also causes any GeoProcessor messages to be echoed to the console.
    #    Added the ToolBoxRun parameter to allow the script to add messages to the GeoProcessor Messaging properly
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and Processing_Variables['ToolBoxRun'] == 'TRUE':
        if MessageLevel == 0:
            gp.AddError(Message)
        if MessageLevel == 1 or MessageLevel == 2:
            gp.AddMessage(Message)
        if MessageLevel == 3:
            gp.AddWarning(Message)

    # Send to the FGDB Processing Log
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and 'Log_FGDB_Table' in Processing_Variables:
        rows = gp.InsertCursor(Processing_Variables['Log_FGDB_Table'])

        row = rows.NewRow()
        row.SetValue("Computer_Name", COMPUTERNAME)
        row.SetValue("User_Name", USERNAME)
        row.SetValue("Script_Name", PYTHON_SCRIPT)
        row.SetValue("Event_Time", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
        row.SetValue("Message_Level", MessageLevel)
        row.SetValue("Message", Message)
        row.SetValue("Start_Time", START_TIME_SQL)
        
        rows.InsertRow(row)
        del row
        del rows
    

    return(1)


#---------------------------------------------------------------------------------------------------------
# WriteOutputToScreen(Output_Comment, Header_Style):
#
#   Prints the Output_Comment to the screen with the appropriate formatting
#       0 : "******" Used for Error Sections
#       1 : "======" Used for the Program's Header and Footer
#       2 : "------" Used for New Sections
#       3+:          No delimitation
#   
#---------------------------------------------------------------------------------------------------------
def WriteOutputToScreen(Output_Comment, Header_Style):
    if Header_Style == 0:
        # Errors get beeps, too!
        print BEEP * 3 + "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 2:
        print "\n" + "-" * 79 + "\n"
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"
        
    #print Output_Comment
    gp.AddMessage(Output_Comment)
    
    if Header_Style == 0:
        print "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"


#---------------------------------------------------------------------------------------------------------
# WriteOutputToFile(Output_Comment, Run_Log)
#   Record to Log 
#
#
#---------------------------------------------------------------------------------------------------------
def WriteOutputToFile(Output_Comment, Run_Log):
# Run logs are not compatible with concurrency - only one process is expected in debug mode at a time!
# First try to append to an existing file - if that fails create a new file.
    ReturnValue = 1
    try:
        if os.path.isfile(Run_Log) == 1:
            # try to open up the fileName 
            file = open(Run_Log,'a') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Appended comment to the file"
        else:
            # try to open up the fileName 
            file = open(Run_Log,'w') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created new file and put in the comment"
    except:
        # All failures are unexpected in this case...  
        # Maybe an invalid drive name, read only file or something?
        Processing_Variables['PROCESS_STATUS'] = -1
        Processing_Variables['PROCESS_INFO']   = "**ERROR: Can't write to file: " + Run_Log   
        print BEEP + BEEP + BEEP
        ReturnValue = -1
    return(ReturnValue)
        

#---------------------------------------------------------------------------------------------------------
# SetupWorkspace()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SetupWorkspace():
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "SetupWorkspace"  

    CPMessage('    - Setting Up Workspace', 2)
    
    if not gp.Exists(Processing_Variables['ScratchFGDB_Path']):
        CPMessage('        -- Creating Scratch FGDB', 4)
        try:
            gp.CreateFileGDB_management(os.path.split(Processing_Variables['ScratchFGDB_Path'])[0], os.path.split(Processing_Variables['ScratchFGDB_Path'])[1])
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created Scratch FGDB"
        except:
            Processing_Variables['PROCESS_STATUS'] = -1
            Processing_Variables['PROCESS_INFO']   = "Error Creating Scratch FGDB"
            CPMessage('** Can not find or create ' + Processing_Variables['ScratchFGDB_Path'], 0)
            sys.exit(100)
    else:
        CPMessage('        -- Scratch FGDB exists, nothing to do.', 4)
       
    if gp.Exists(Processing_Variables['OperationalFGDB_Path']):
        CPMessage('        -- Operational FGDB exists.', 4)
    else:
        CPMessage(' **** Operational FGDB does not exist: ' + Processing_Variables['OperationalFGDB_Path'], 0)
        sys.exit(100)

    if gp.Exists(Processing_Variables['Tools_and_Templates_FGDB_Path']):
        CPMessage('        -- Tools and Templates FGDB exists.', 4)
    else:
        CPMessage(' **** Tools and Templates FGDB does not exist: ' + Processing_Variables['Tools_and_Templates_FGDB_Path'], 0)
        sys.exit(100)


    # Create/overwrite the old error tables
    CPMessage('        -- Initializing Domain Error Table.', 4)
    if gp.Exists(Processing_Variables['DomainErrorTable']):
        while not gp.TestSchemaLock(Processing_Variables['DomainErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['DomainErrorTable'], 3)
            time.sleep(10)        
        gp.Delete_management(Processing_Variables['DomainErrorTable'])
    gp.CreateTable_management(os.path.split(Processing_Variables['DomainErrorTable'])[0], os.path.split(Processing_Variables['DomainErrorTable'])[1], Processing_Variables['Template_ErrorTable'])

    CPMessage('        -- Initializing Range Error Table.', 4)
    if gp.Exists(Processing_Variables['RangeErrorTable']):
        while not gp.TestSchemaLock(Processing_Variables['RangeErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['RangeErrorTable'], 3)
            time.sleep(10)    
        gp.Delete_management(Processing_Variables['RangeErrorTable'])
    gp.CreateTable_management(os.path.split(Processing_Variables['RangeErrorTable'])[0], os.path.split(Processing_Variables['RangeErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    CPMessage('        -- Initializing Mandatory Field Error Table.', 4)
    if gp.Exists(Processing_Variables['MandatoryFieldErrorTable']):
        while not gp.TestSchemaLock(Processing_Variables['MandatoryFieldErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['MandatoryFieldErrorTable'], 3)
            time.sleep(10)    
        gp.Delete_management(Processing_Variables['MandatoryFieldErrorTable'])
    gp.CreateTable_management(os.path.split(Processing_Variables['MandatoryFieldErrorTable'])[0], os.path.split(Processing_Variables['MandatoryFieldErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   

    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# SendRunLogToEmail()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SendRunLogToEmail(mailto):
    import smtplib
    CPMessage('  * Sending log to ' + mailto, 2)
    mailfrom = 'GIS_Python@cloverpoint.com'
    # mailto = 'mneal@cloverpoint.com'
    mailsubject = 'Run Log: ' + os.path.split(PYTHON_SCRIPT)[1] + ' (' + COMPUTERNAME + '\\' + USERNAME + ')'
    mailbody = "This is an automatic run log.\n\n"
    
    for LogEntry in Processing_Variables['PROCESS_LOG']:
        mailbody = mailbody + '\n' + LogEntry[0]

    mailmsg = ("From: " + mailfrom + "\r\nTo: " + mailto + "\r\nSubject: " + mailsubject + "\r\n" + mailbody)

    mailserver = smtplib.SMTP('shawmail.gv.shawcable.net')

    # SMTP.login(user, password)

    mailserver.sendmail(mailfrom, mailto, mailmsg)
    mailserver.quit()
    return(Processing_Variables['PROCESS_STATUS'])



#=========================================================================================================
#  Special Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# LoadTableToPython(Input_Table)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def LoadTableToPython(Input_Table, PK):
    # A Table is a dictionary with the PK as the dictionary key, and each field/value pair as a sub-dictionary.
    CPMessage('    - Load Table To Python', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "LoadTableToPython" 

    result = gp.GetCount_management(Input_Table)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    Field_List = gp.ListFields(Input_Table)
    
    CPMessage('        - There are ' + str(len(Field_List)) + ' fields.', 4)
    CPMessage('        - Retrieving values for ' + str(Number_of_Features) + ' rows.', 4)
    
    #### rows = gp.SearchCursor(Input_Table, "", "", "", "STE_SHORT_TABLE_SP")
    rows = gp.SearchCursor(Input_Table)
    row = rows.Next()
    
    TableData = {}
    
    while row:
        Current_Row = {}
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed) + ' of ' + str(Number_of_Features), 4)
        
        for Current_Field in Field_List:
            CPMessage('            - Getting ' + str(Current_Field.name), 5) 
            try:
                Current_Row[Current_Field.name] = str(row.GetValue(Current_Field.name))
            except:
                Current_Row[Current_Field.name] = "Data Error - Check database entry."
    
        CPMessage('            - Adding ' + Current_Row[PK] + ': ' + str(Current_Row), 5)
        TableData[Current_Row[PK]] = Current_Row        
        
        
        Number_Processed = Number_Processed + 1
        row = rows.Next()
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    del row
    del rows

    return(TableData)


#---------------------------------------------------------------------------------------------------------
# ExtractDomainsAndRanges()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def ExtractDomainsAndRanges(User_Defined_Field_Def):
    CPMessage('    - Loading Domains and Ranges to Memory', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "LoadDomains" 

    OriginalWorkspace = gp.Workspace
    
    Domains = {}
    Ranges = {}
    Case_Restrictions = {}
    Mandatory_Fields = []
    
    # First check that TEIS_User_Defined_Field_Def is populated properly
    
    for CurrentRow in User_Defined_Field_Def:
        if User_Defined_Field_Def[CurrentRow]['COLM_NAME'] in ['',' ','None']:
            CPMessage('        *** TEIS_User_Defined_Field_Def table incomplete', 4)
            CPMessage('        *** Values missing in COLM_NAME field', 4)
            sys.exit(101)
        if User_Defined_Field_Def[CurrentRow]['VALUE_TYP'].upper() not in ['C','N','TEXT','FLOAT','DOUBLE','SHORT','LONG']:
            CPMessage('        *** TEIS_User_Defined_Field_Def table invalid', 4)
            CPMessage('        *** Values in VALUE_TYPE field must be one of C, N, TEXT, FLOAT, DOUBLE, SHORT, or LONG', 4)
            CPMessage('        *** Bad value: ' + User_Defined_Field_Def[CurrentRow]['VALUE_TYP'].upper(), 4)
            CPMessage('        *** See ESRI ArcGIS Help documentation for explanations of field types', 4)
            sys.exit(101)
        if User_Defined_Field_Def[CurrentRow]['REQUIRED'].upper() not in ['Y','N','',' ','None']:
            CPMessage('        *** TEIS_User_Defined_Field_Def table invalid', 4)
            CPMessage('        *** Values in REQUIRED field must be one of Y, N, or <Null>', 4)
            CPMessage('        *** See ESRI ArcGIS Help documentation for explanations of field types', 4)
            sys.exit(101)
        if User_Defined_Field_Def[CurrentRow]['VALUE_TYP'].upper() == 'TEXT' and int(User_Defined_Field_Def[CurrentRow]['LENGTH']) < 1:
            CPMessage('        *** TEIS_User_Defined_Field_Def invalid', 4)
            CPMessage('        *** LENGTH must be specified for user-defined fields of type TEXT', 4)
            CPMessage('        *** See ESRI ArcGIS Help documentation for explanations of field types', 4)
            sys.exit(101)
        if User_Defined_Field_Def[CurrentRow]['CASE'].upper() not in ['',' ','None','NONE','U','L','M']:
            CPMessage('        *** TEIS_User_Defined_Field_Def table invalid', 4)
            CPMessage('        *** Values in CASE field must be one of L, M, U, or <Null>', 4)
            sys.exit(101)
        if User_Defined_Field_Def[CurrentRow]['VALUE'] not in ['',' ','None'] and (User_Defined_Field_Def[CurrentRow]['CASE'].upper() == 'U' and User_Defined_Field_Def[CurrentRow]['VALUE'] != User_Defined_Field_Def[CurrentRow]['VALUE'].upper()) or (User_Defined_Field_Def[CurrentRow]['CASE'].upper() == 'L' and User_Defined_Field_Def[CurrentRow]['VALUE'] != User_Defined_Field_Def[CurrentRow]['VALUE'].lower()):
            CPMessage('        *** TEIS_User_Defined_Field_Def table invalid', 4)
            CPMessage('        *** Specified valid value "' + User_Defined_Field_Def[CurrentRow]['Value'] + '" does not conform to specified case restriction', 4)
            sys.exit(101)
        if str(User_Defined_Field_Def[CurrentRow]['MIN_VALUE']) not in ['',' ','None'] and str(User_Defined_Field_Def[CurrentRow]['MAX_VALUE']) not in ['',' ','None']:
            if int(User_Defined_Field_Def[CurrentRow]['MIN_VALUE']) > int(User_Defined_Field_Def[CurrentRow]['MAX_VALUE']):
                CPMessage('        *** TEIS_User_Defined_Field_Def table invalid', 4)
                CPMessage('        *** Values in CASE field must be one of L, M, U, or <Null>', 4)
                sys.exit(101)
            
    # Load each specified valid value into a Domains dictionary keyed on Field Name
    for CurrentRow in User_Defined_Field_Def:
        Current_Field = User_Defined_Field_Def[CurrentRow]['COLM_NAME']
        Domains[Current_Field] = []

    for CurrentRow in User_Defined_Field_Def:
        
        Current_Field = User_Defined_Field_Def[CurrentRow]['COLM_NAME']
        Current_Value = User_Defined_Field_Def[CurrentRow]['VALUE']
        Current_Range_Minimum = User_Defined_Field_Def[CurrentRow]['MIN_VALUE']
        Current_Range_Maximum = User_Defined_Field_Def[CurrentRow]['MAX_VALUE']
        
        if Current_Value not in ['',' ','None'] and Current_Value not in Domains[Current_Field]:
            Domains[Current_Field].append(Current_Value)
        
        if User_Defined_Field_Def[CurrentRow]['REQUIRED'] == 'Y' and Current_Field not in Mandatory_Fields:
            Mandatory_Fields.append(Current_Field)
        
        if Current_Range_Minimum not in ['',' ','None',None] and Current_Range_Maximum not in ['',' ','None',None]:
            Ranges[Current_Field] = {'Minimum':  int(Current_Range_Minimum), 'Maximum': int(Current_Range_Maximum)}

    # Domains is a dictionary of keys with a field name and values as a list.
    Processing_Variables['Domains'] = Domains
    Processing_Variables['Mandatory_Fields'] = Mandatory_Fields
    Processing_Variables['Ranges'] = Ranges
    #CPMessage('Domains: ' + str(Domains), 4)
    #CPMessage('Mandatory Fields: ' + str(Mandatory_Fields), 4)
    #CPMessage('Ranges: ' + str(Ranges), 4)
    
    gp.Workspace = OriginalWorkspace
    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# CheckDomains()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CheckDomains():
    CPMessage('    - Checking Domains', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "CheckDomains" 


    # Get the Field List
    Field_List = gp.ListFields(INPUT_DATA_TABLE)

    Check_Fields = []
    Mandatory_Check_Fields = []
    
    for Current_Field in Field_List:

        if Current_Field.Name in Processing_Variables['Domains'] and len(Processing_Variables['Domains'][Current_Field.Name]) != 0:
            Check_Fields.append(Current_Field.Name)
            CPMessage('            - Checking Field: ' + str(Current_Field.Name), 4)
        else:
            CPMessage('            - Not Checking Field: ' + str(Current_Field.Name), 5)

    CPMessage('        - Checking Fields: ' + str(Check_Fields), 5)

    # Find the OID Field
    Object_ID_Field = gp.ListFields(INPUT_DATA_TABLE, "*", "OID")[0].Name
    # Use the TEIS_ID field instead
    Object_ID_Field = "TEIS_ID"

    result = gp.GetCount_management(INPUT_DATA_TABLE)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Checking individual values for ' + str(Number_of_Features) + ' features.', 4)
    rows = gp.SearchCursor(INPUT_DATA_TABLE)
    row = rows.Next()
    Error_List = []
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)
        
        for Field_Name in Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            #### Testing numeric domains here!
            Test_Value = str(row.GetValue(Field_Name))
            Test_Domain_Name = Processing_Variables['Domains'][Field_Name]
            # print ' TEST DOMAIN:', Processing_Variables['Domains'][Test_Domain_Name]
            CPMessage('            - Testing: ' + Field_Name + ' with value ' + str(Test_Value), 5)

            if Test_Value in Processing_Variables['Domains'][Field_Name] or str(Test_Value) == 'None' or str(Test_Value) == '' or str(Test_Value) == ' ' or len(Processing_Variables['Domains'][Field_Name]) == 0:
                # The value is null or in the allowed list
                CPMessage('            - Value OK : ' + str(Test_Value), 5)
                pass
            else:
                Current_OID = row.GetValue(Object_ID_Field)
                CPMessage('            - Invalid Value : ' + str(Test_Value) + ' in ' + Field_Name + ' at OID ' + str(Current_OID) + '.', 4)
                Error_List.append((Current_OID, 'Invalid Value in ' + Field_Name + ' (' + str(Test_Value) + ').'))

        Mandatory_Field_Name = ''
        for Mandatory_Field_Name in Mandatory_Check_Fields:
            Test_Value = str(row.GetValue(Mandatory_Field_Name))
            CPMessage('            - Testing: ' + Mandatory_Field_Name + ' for mandatory value', 5)
            if str(Test_Value) not in ['None','',' ']:
                CPMessage('            - Value OK : ' + str(Test_Value), 5)
                pass
            else:
                Current_OID = row.GetValue(Object_ID_Field)
                CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name + ' at OID ' + str(Current_OID) + '.', 4)
                Error_List.append((Current_OID, 'Mandatory value missing in ' + Mandatory_Field_Name + '.'))
            
        Number_Processed = Number_Processed + 1
        row = rows.Next()
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    del row
    del rows

    # Write the Error List out
    CPMessage('        - Writing Error Table with ' + str(len(Error_List)) + ' errors.', 3)
    rows = gp.InsertCursor(Processing_Variables['DomainErrorTable'])
    
    Number_Processed = 0
    
    for Current_Error in Error_List:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)        
        row = rows.NewRow()
        row.SetValue('CLASS', INPUT_DATA_TABLE)
        row.SetValue('FEATURE_ID', Current_Error[0])
        row.SetValue('PROBLEM', Current_Error[1])
        # print INPUT_DATA_TABLE, Current_Error[0], Current_Error[1]
        Number_Processed = Number_Processed + 1
        rows.InsertRow(row)
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    
    # A new row is not created if there are no errors to be written.
    try:
        del row
    except:
        pass
        
    del rows
        
    return(Processing_Variables['PROCESS_STATUS'])
    


#---------------------------------------------------------------------------------------------------------
# CheckRanges()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CheckRanges():
    CPMessage('    - Checking Ranges', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "CheckRanges" 


    # Get the Field List
    Field_List = gp.ListFields(INPUT_DATA_TABLE)

    Check_Fields = []
    
    CPMessage('        - Ranges: ' + str(Processing_Variables['Ranges'].keys()), 5)
    
    for Current_Field in Field_List:
        if Current_Field.Name.upper() in Processing_Variables['Ranges']:
            Check_Fields.append(Current_Field.Name)
            CPMessage('            - Checking Field: ' + str(Current_Field.Name), 4)
        else:
            CPMessage('            - Not Checking Field: ' + str(Current_Field.Name), 5)

    CPMessage('        - Checking Fields: ' + str(Check_Fields), 5)

    # Find the OID Field
    Object_ID_Field = gp.ListFields(INPUT_DATA_TABLE, "*", "OID")[0].Name

    result = gp.GetCount_management(INPUT_DATA_TABLE)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Checking individual values for ' + str(Number_of_Features) + ' features.', 4)
    rows = gp.SearchCursor(INPUT_DATA_TABLE)
    row = rows.Next()
    Error_List = []
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)
        
        for Field_Name in Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            Test_Value = row.GetValue(Field_Name)
            
            Test_Range_Name = Field_Name
            # print ' TEST Range:', Processing_Variables['Ranges'][Field_Name]
            CPMessage('            - Testing: ' + Field_Name + ' with value ' + str(Test_Value), 5)
            
            Test_Validation = 'Number'
            
            if str(Test_Value) in ['None', '', ' ']:
                Test_Validation = 'Empty'
                CPMessage('            - Value OK : NULL (' + str(Test_Value) + ').', 5)
            if Test_Validation == 'Number':
                try:
                    Test_Value = int(Test_Value)
                except:
                    Test_Validation = 'NotNumber'
                    Current_OID = row.GetValue(Object_ID_Field)
                    CPMessage('            - Invalid Value (Not an Integer): ' + str(Test_Value) + ' in ' + Field_Name + ' at OID ' + str(Current_OID) + '.', 4)
                    Error_List.append((Current_OID, 'Invalid Value (not an integer) in ' + Field_Name + ' (' + str(Test_Value) + ').'))



            if Test_Validation == 'Number':
                #if int(Processing_Variables['Ranges'][Test_Range_Name]['Minimum']) <= int(Test_Value) and int(Test_Value) <= int(Processing_Variables['Ranges'][Test_Range_Name]['Maximum']):
                if (Processing_Variables['Ranges'][Test_Range_Name]['Minimum'] <= Test_Value <= Processing_Variables['Ranges'][Test_Range_Name]['Maximum']):
                    # The value is in the allowed range
                    CPMessage('            - Value OK : ' + str(Test_Value), 5)
                    pass
                else:
                    Current_OID = row.GetValue(Object_ID_Field)
                    CPMessage('            - Invalid Value : ' + str(Test_Value) + ' in ' + Field_Name + ' at OID ' + str(Current_OID) + '.', 4)
                    Error_List.append((Current_OID, 'Invalid Value (out of range) in ' + Field_Name + ' (' + str(Test_Value) + ').'))
        
        

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    del row
    del rows

    # Write the Error List out
    CPMessage('        - Writing Error Table with ' + str(len(Error_List)) + ' errors.', 3)
    rows = gp.InsertCursor(Processing_Variables['RangeErrorTable'])
    
    Number_Processed = 0
    
    for Current_Error in Error_List:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)        
        row = rows.NewRow()
        row.SetValue('CLASS', INPUT_DATA_TABLE)
        row.SetValue('FEATURE_ID', Current_Error[0])
        row.SetValue('PROBLEM', Current_Error[1])
        # print INPUT_DATA_TABLE, Current_Error[0], Current_Error[1]
        Number_Processed = Number_Processed + 1
        rows.InsertRow(row)
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    
    # A new row is not created if there are no errors to be written.
    try:
        del row
    except:
        pass
        
    del rows
        
    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# CheckMandatoryFields()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CheckMandatoryFields():
    CPMessage('    - Checking Mandatory Fields', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "CheckMandatoryFields" 


    # Get the Field List
    Field_List = gp.ListFields(INPUT_DATA_TABLE)

    Mandatory_Check_Fields = []
    
    for Current_Field in Field_List:

        if Current_Field.Name in Processing_Variables['Mandatory_Fields']:
            Mandatory_Check_Fields.append(Current_Field.Name)
            CPMessage('        - Checking Mandatory Field: ' + str(Current_Field.Name), 4)

    # Find the OID Field
    Object_ID_Field = gp.ListFields(INPUT_DATA_TABLE, "*", "OID")[0].Name
    # Use the TEIS_ID field instead
    Object_ID_Field = "TEIS_ID"

    result = gp.GetCount_management(INPUT_DATA_TABLE)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Checking individual values for ' + str(Number_of_Features) + ' features.', 4)
    rows = gp.SearchCursor(INPUT_DATA_TABLE)
    row = rows.Next()
    Error_List = []
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)
        
        Mandatory_Field_Name = ''
        for Mandatory_Field_Name in Mandatory_Check_Fields:
            Test_Value = str(row.GetValue(Mandatory_Field_Name))
            CPMessage('            - Testing: ' + Mandatory_Field_Name + ' for mandatory value', 5)
            if str(Test_Value) not in ['None','',' ']:
                CPMessage('            - Value OK : ' + str(Test_Value), 5)
                pass
            else:
                Current_OID = row.GetValue(Object_ID_Field)
                CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name + ' at OID ' + str(Current_OID) + '.', 4)
                Error_List.append((Current_OID, 'Mandatory value missing in ' + Mandatory_Field_Name + '.'))

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    del row
    del rows

    # Write the Error List out
    CPMessage('        - Writing Error Table with ' + str(len(Error_List)) + ' errors.', 3)
    rows = gp.InsertCursor(Processing_Variables['MandatoryFieldErrorTable'])
    
    Number_Processed = 0
    
    for Current_Error in Error_List:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)        
        row = rows.NewRow()
        row.SetValue('CLASS', INPUT_DATA_TABLE)
        row.SetValue('FEATURE_ID', Current_Error[0])
        row.SetValue('PROBLEM', Current_Error[1])
        # print INPUT_DATA_TABLE, Current_Error[0], Current_Error[1]
        Number_Processed = Number_Processed + 1
        rows.InsertRow(row)
    CPMessage('            - Processed all ' + str(Number_Processed), 4)
    
    # A new row is not created if there are no errors to be written.
    try:
        del row
    except:
        pass
        
    del rows
        
    return(Processing_Variables['PROCESS_STATUS'])

#*********************************************************************************************************
#  Main Program
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# Run Main
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain():

    Initialize()
    SetupWorkspace()
    
    User_Defined_Field_Def = LoadTableToPython(INPUT_FIELD_DEF_TABLE,'OBJECTID')

    ExtractDomainsAndRanges(User_Defined_Field_Def)

    CheckDomains()
    
    CheckRanges()    

    CheckMandatoryFields()
    
    return Processing_Variables['PROCESS_STATUS']

#*********************************************************************************************************
#  Execution
#*********************************************************************************************************


runMain()



#---------------------------------------------------------------------------------------------------------
#  Clean-up
#---------------------------------------------------------------------------------------------------------


END_TIME       = time.ctime(time.time())
END_TIME_SEC   = time.time()
END_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
EXECUTION_TIME = END_TIME_SEC - START_TIME_SEC
if EXECUTION_TIME < 120:
    EXECUTION_TIME_STRING = str(int(EXECUTION_TIME)) + ' seconds.'
elif EXECUTION_TIME < 3600:
    EXECUTION_TIME_STRING = str(round(EXECUTION_TIME / 60, 1)) + ' minutes.'
else:
    EXECUTION_TIME_STRING = str(round(EXECUTION_TIME / 3600, 2)) + ' hours.'


CPMessage('    End Time: ' + END_TIME, 2)
CPMessage('    Execution Time: ' + EXECUTION_TIME_STRING, 4)
CPMessage('Complete: ' + PYTHON_SCRIPT, 1)

# gp.CheckInExtension("spatial")
del gp

# Report Program Execution Status back to the calling program


if 'SendRunLogToEmail' in Processing_Variables:
    for Email in Processing_Variables['SendRunLogToEmail']:
        SendRunLogToEmail(Email)

#*********************************************************************************************************
#  End of Program
#*********************************************************************************************************
