#=========================================================================================================
#                                      ValidateData
#=========================================================================================================
#
# Author(s)         :  Clover Point Cartographics (Mike Neal, Jeff Kruys)
#
# Date              :  2012-08-30
#
# Purpose           :  This script uses the TEIS_Metadata_* and Domain_* tables in the Tools_and_Templates.gdb
#                      file geodatabase to validate an input feature class or table against defined
#                      attribute domains and ranges.
#              
# Arguments         :  argv(1) = INPUT_FC                               (Feature class to be validated) 
#                   :  argv(2) = {MandatoryFields|NoMandatoryFields}    (Option to check mandatory fields, or not) 
#                   :  argv(3) = {CheckGeometry|NoCheckGeometry}        (Option to check geometry, or not) 
#
# Example Input     :  X:\fullpath\ValidateData.py X:\fullpath\TEIS_Master_Long_Tbl NoMandatoryFields NoCheckGeometry
#
# Outputs           :  Seven new tables in the Operational_Data.gdb file geodatabase. These tables list the
#                      errors found in the input table.
#              
#                           input_table_name_CriticalRangeErrors
#                           input_table_name_DomainErrors
#                           input_table_name_GeometryErrors
#                           input_table_name_MandatoryFieldErrors
#                           input_table_name_RangeErrors
#                           input_table_name_RowErrors
#                           input_table_name_SummaryErrors
#
#                      In each output table the FEATURE_ID field should be used to link to the TEIS_ID
#                      field of the Long Table feature class input (or to the OBJECTID field of a Project
#                      Details table input).
#
# Dependencies      :  Must have TEIS_Metadata_* and Domain_* tables in Tools_and_Templates.gdb
#              
# Expected Duration :  < 10 minutes for 100,000 records
#                   :  ~7 days for 24,000,000 records in ArcGIS 9.3.1 or 10.0
#                   :  ~12 hours for approximately 24,000,000 records in ArcGIS 10.1 (with NoMandatoryFields and NoCheckGeometry)
#
# History           :  2012-08-30 (JK): Script created, based on ValidateDomains, ValidateRows etc. but using
#                                       cursors to validate everything one row at a time, rather than requiring
#                                       the running of several different scripts. Also new parallel code added
#                                       that uses Data Access module of arcpy, which gives faster cursor
#                                       operations (in Arc 10.1 and up).
#                   :  2012-10-11 (JK): Fixed bug that prevented row errors from being detected.
#                                       Reworked Arc 10.1 code to use edit session to keep multiple insert
#                                       cursors open simultaneously, rather than creating and deleting an
#                                       insert cursor object with each record to write. Much faster.
#                   :  2012-10-12 (JK): Update to CPMessage routine: only append messages to PROCESS_LOG
#                                       if the e-mail function is being used (otherwise it can fill memory).
#                                       Replaced all "gp" references to "arcpy" to save having to test
#                                       current Arc version during every CPMessage call.
#                   :  2013-01-02 (JK): Revised GEOP_SCM# logic to allow either 
#                                       GEOP_SCM# = GEOP_INZ# + GEOP_SCM#A + GEOP_SCM#B  (if GEOP_INZ# is not blank)
#                                       or        = GEOP_SCM#A + GEOP_SCM#B + GEOP_SCM#C (if GEOP_INZ# is blank)
#                   :  2013-04-22 (JK): Removed references to TEIS_Master_Project_Details table; set the edit workspace
#                                       to the location of the input data rather than Operational_Data.gdb
#                   :  2013-06-19 (JK): Added stra() function which replaces unicode characters with a question mark,
#                                       and replaced all str() functions throughout script with stra().
#
#=========================================================================================================
#=========================================================================================================
#  Process Status and Return Codes:
#      -1   - Unexpected error
#       0   - Completed Successful Run
#       1   - In progress
#       100 - Invalid Arguments
#       101 - ESRI License Error
#
#=========================================================================================================

#*********************************************************************************************************
#  Program Initialization
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
#  Python and System Modules
#---------------------------------------------------------------------------------------------------------

from __future__ import division

import sys
import string
import os
import time


#---------------------------------------------------------------------------------------------------------
#  Arguement Validation
#---------------------------------------------------------------------------------------------------------

# Capture/Validate Parameters
#  Test that the right number of paramters were provided and validated
#  Note that 1 must be added because the Python script is an arguement (0) supplied to the
#  Python Environment.


# If long text parameters are required, use:  any_value = arcpy.GetParameterAsText(0)

if len(sys.argv) <> 4:
    print "\nInvalid number of arguments (" + str(len(sys.argv) - 1) + " supplied) to:"
    print "    " + sys.argv[0] + "\n"
    print "Usage:   " + sys.argv[0] + " [INPUT_FC] {MandatoryFields|NoMandatoryFields} {CheckGeometry|NoCheckGeometry}"
    print " [INPUT_FC]                          : Feature class to be validated (input feature class in a FGDB)."
    print " {MandatoryFields|NoMandatoryFields} : Specify 'MandatoryFields' if mandatory fields should be validated"
    print " {CheckGeometry|NoCheckGeometry}     : Specify 'CheckGeometry' if the CheckGeometry function is needed"
    sys.exit(100)
elif sys.argv[2] not in ["MandatoryFields","NoMandatoryFields","true","false"] or sys.argv[3] not in ["CheckGeometry","NoCheckGeometry","true","false"]:
    print "\nInvalid arguments (" + str(len(sys.argv) - 1) + " supplied) to:"
    print "    " + sys.argv[0] + "\n"
    print "Usage:   " + sys.argv[0] + " [INPUT_FC] {MandatoryFields|NoMandatoryFields} {CheckGeometry|NoCheckGeometry}"
    print " [INPUT_FC]                          : Feature class to be validated (input feature class in a FGDB)."
    print " {MandatoryFields|NoMandatoryFields} : Specify 'MandatoryFields' if mandatory fields should be validated"
    print " {CheckGeometry|NoCheckGeometry}     : Specify 'CheckGeometry' if the CheckGeometry function is needed"
    sys.exit(100)
    
PYTHON_SCRIPT       = sys.argv[0]
INPUT_FOR_TESTING   = sys.argv[1]
if sys.argv[2] in ['NoMandatoryFields',"false"]:
    CHECK_MANDATORY_FIELDS = False
else:
    CHECK_MANDATORY_FIELDS = True
if sys.argv[3] in ['NoCheckGeometry',"false"]:
    CHECK_GEOMETRY = False
else:
    CHECK_GEOMETRY = True

#---------------------------------------------------------------------------------------------------------
#  Standard/Global Variables/Constants
#---------------------------------------------------------------------------------------------------------

# Global Constants
COMPUTERNAME     = os.environ['COMPUTERNAME']
USERNAME         = os.environ['USERNAME']
START_TIME       = time.ctime(time.time())
START_TIME_SEC   = time.time()
START_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
BEEP             = chr(7)


# Global Variables
#   The Processing_Variables is the collection of messaging and script specific information needed
#   throughout this program.  It is constantly updated, and generally includes all paths and configuration
#   information.  For parameters that are passed in to the script, they should generally be assigned
#   to a Processing_Variable through the Initialize() routine.

Processing_Variables = {}



#---------------------------------------------------------------------------------------------------------
#  CPC Variables/Constants
#---------------------------------------------------------------------------------------------------------




#---------------------------------------------------------------------------------------------------------
#  Set up the ArcGIS Geoprocessing Environment
#---------------------------------------------------------------------------------------------------------

try:
    print 'Importing arcgisscripting'
    import arcgisscripting
    arcpy = arcgisscripting.create(93)
except:
    print '**** ArcGIS Failed to initialize a GeoProcessor Object'
    sys.exit(101)

print 'Getting ArcGIS version'
d = arcpy.GetInstallInfo('desktop')
if float(d['Version'][:4]) >= 10.1:
    print 'Dropping arcgisscripting'
    del arcpy
    print 'Importing arcpy'
    import arcpy

#---------------------------------------------------------------------------------------------------------
#  Script Specific Variables/Constants
#---------------------------------------------------------------------------------------------------------



#*********************************************************************************************************
#  Routines
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# TemplateRoutine93(Input_FC)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TemplateRoutine93(Input_FC):
    CPMessage('    - Template Routine', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TemplateRoutine" 

    result = arcpy.GetCount_management(Input_FC)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Calculating individual values for' + str(Number_of_Features) + ' features.', 4)
    rows = arcpy.SearchCursor(Input_FC)
    row = rows.Next()
    
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + str(Number_Processed), 4)

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    del row
    del rows

    return(Processing_Variables['PROCESS_STATUS'])

#=========================================================================================================
#  Standard Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# stra(value)
#
#       Converts any value to a string. Replaces Python's stra() which incurs an error if the value
#       happens to have a unicode character in it. Unicode characters will be replaced with ?.
#
#---------------------------------------------------------------------------------------------------------
def stra(value):
    return unicode(value).encode('ascii','replace')

#---------------------------------------------------------------------------------------------------------
# Initialize93()
#
#       Set up all of the paths and variables that we need to run this script.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def Initialize93():
    print '    - Initializing Script'
    
    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Script_Name'] = os.path.split(PYTHON_SCRIPT)[1]
    Processing_Variables['Log_Directory_Path'] = Processing_Variables['Script_Directory_Path'] + r'\log'

    Processing_Variables['Project_Base_Path']      = os.path.split(Processing_Variables['Script_Directory_Path'])[0]

    
    Processing_Variables['ToolBoxRun'] = 'FALSE'
    # Processing_Variables['SendRunLogToEmail'] = ['mneal@cloverpoint.com', 'chaos_mike@hotmail.com']
    Processing_Variables['Log_Level'] = 4
    Processing_Variables['PROCESS_INFO']   = "RUNNING " + PYTHON_SCRIPT
    Processing_Variables['PROCESS_STATUS'] = 1
    Processing_Variables['PROCESS_LOG'] = []
    Processing_Variables['MessageQueueFile']   = Processing_Variables['Log_Directory_Path'] + '\\' + Processing_Variables['Script_Name'] + '.' + COMPUTERNAME + '.log'

    #### Adding a FGDB Log will significantly slow down performance.
    # Processing_Variables['Log_FGDB_Table'] = Processing_Variables['Project_Base_Path'] + r"\Operational.gdb\Processing_Log"

    if not arcpy.Exists(Processing_Variables['Log_Directory_Path']):
        os.mkdir(Processing_Variables['Log_Directory_Path'])
        CPMessage(r'Created \Scripts\log folder', 1)

    CPMessage('Executing: ' + PYTHON_SCRIPT, 1)
    CPMessage('    Start Time: ' + START_TIME, 2)
    CPMessage('    Python ' + sys.version, 4)

    # Load required toolboxes...
    # arcpy.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")
    
    # Set the Geoprocessing environment...
    arcpy.OverwriteOutput = 1
    # arcpy.cellSize = "MINOF"
    # arcpy.workspace = INPUT_LOCATION
    
    
    
    # CPC Naming Standard: {variable}_Path is the fully qualified path, otherwise it is just the name of the item.
  
    # Processing Variables
   
    Processing_Variables['OperationalFGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Operational_Data.gdb'
    Processing_Variables['Scratch_Path']           = Processing_Variables['Project_Base_Path'] + r'\Scratch'
    Processing_Variables['ScratchFGDB_Path']           = Processing_Variables['Scratch_Path'] + r'\Scratch.gdb'
    arcpy.scratchworkspace = Processing_Variables['ScratchFGDB_Path']
    
    Processing_Variables['Tools_and_Templates_FGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Tools_and_Templates.gdb'
    
    Processing_Variables['TEIS_Metadata_Items'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\TEIS_Metadata_Items'
    Processing_Variables['TEIS_Metadata_Tables'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\TEIS_Metadata_Tables'

    Processing_Variables['Template_ErrorTable'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\Template_ErrorTable'
    Processing_Variables['Template_ErrorSummary'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\Template_ErrorSummary'

    Processing_Variables['DomainErrorTable']         = INPUT_FOR_TESTING + r'_DomainErrors'
    Processing_Variables['RangeErrorTable']          = INPUT_FOR_TESTING + r'_RangeErrors'
    Processing_Variables['CriticalRangeErrorTable']  = INPUT_FOR_TESTING + r'_CriticalRangeErrors'
    Processing_Variables['RowErrorTable']            = INPUT_FOR_TESTING + r'_RowErrors'
    Processing_Variables['GeometryErrorTable']       = INPUT_FOR_TESTING + r'_GeometryErrors'
    Processing_Variables['MandatoryFieldErrorTable'] = INPUT_FOR_TESTING + r'_MandatoryFieldErrors'
    Processing_Variables['SummaryErrorTable']        = INPUT_FOR_TESTING + r'_SummaryErrors'
  
    Processing_Variables['Albers_PRJ']           = Processing_Variables['Project_Base_Path'] + r'\Scripts\NAD_1983_BC_Environment_Albers.prj'

    if arcpy.Exists(INPUT_FOR_TESTING):
        CPMessage(' - Input Exists: ' + INPUT_FOR_TESTING, 4)
    else:
        CPMessage('** ERROR: INPUT DOES NOT EXIST: ' + INPUT_FOR_TESTING, 0)
        sys.exit(100)

   
    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# Initialize101()
#
#       Set up all of the paths and variables that we need to run this script.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def Initialize101():
    print '    - Initializing Script'
    
    Processing_Variables['Script_Directory_Path'] = os.path.split(PYTHON_SCRIPT)[0]
    Processing_Variables['Script_Name'] = os.path.split(PYTHON_SCRIPT)[1]
    Processing_Variables['Log_Directory_Path'] = Processing_Variables['Script_Directory_Path'] + r'\log'

    Processing_Variables['Project_Base_Path']      = os.path.split(Processing_Variables['Script_Directory_Path'])[0]

    
    Processing_Variables['ToolBoxRun'] = 'FALSE'
    # Processing_Variables['SendRunLogToEmail'] = ['mneal@cloverpoint.com', 'chaos_mike@hotmail.com']
    Processing_Variables['Log_Level'] = 4
    Processing_Variables['PROCESS_INFO']   = "RUNNING " + PYTHON_SCRIPT
    Processing_Variables['PROCESS_STATUS'] = 1
    Processing_Variables['PROCESS_LOG'] = []
    Processing_Variables['MessageQueueFile']   = Processing_Variables['Log_Directory_Path'] + '\\' + Processing_Variables['Script_Name'] + '.' + COMPUTERNAME + '.log'

    #### Adding a FGDB Log will significantly slow down performance.
    # Processing_Variables['Log_FGDB_Table'] = Processing_Variables['Project_Base_Path'] + r"\Operational.gdb\Processing_Log"

    if not arcpy.Exists(Processing_Variables['Log_Directory_Path']):
        os.mkdir(Processing_Variables['Log_Directory_Path'])
        CPMessage(r'Created \Scripts\log folder', 1)

    CPMessage('Executing: ' + PYTHON_SCRIPT, 1)
    CPMessage('    Start Time: ' + START_TIME, 2)
    CPMessage('    Python ' + sys.version, 4)

    # Load required toolboxes...
    # arcpy.AddToolbox("C:/Program Files/ArcGIS/ArcToolbox/Toolboxes/Data Management Tools.tbx")
    
    # Set the Geoprocessing environment...
    arcpy.OverwriteOutput = 1
    # arcpy.cellSize = "MINOF"
    # arcpy.env.workspace = INPUT_LOCATION
    
    
    
    # CPC Naming Standard: {variable}_Path is the fully qualified path, otherwise it is just the name of the item.
  
    # Processing Variables
   
    Processing_Variables['OperationalFGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Operational_Data.gdb'
    Processing_Variables['Scratch_Path']           = Processing_Variables['Project_Base_Path'] + r'\Scratch'
    Processing_Variables['ScratchFGDB_Path']           = Processing_Variables['Scratch_Path'] + r'\Scratch.gdb'
    arcpy.env.scratchWorkspace = Processing_Variables['ScratchFGDB_Path']
    
    Processing_Variables['Tools_and_Templates_FGDB_Path']   = Processing_Variables['Project_Base_Path'] + r'\Tools_and_Templates.gdb'
    
    Processing_Variables['TEIS_Metadata_Items'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\TEIS_Metadata_Items'
    Processing_Variables['TEIS_Metadata_Tables'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\TEIS_Metadata_Tables'

    Processing_Variables['Template_ErrorTable'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\Template_ErrorTable'
    Processing_Variables['Template_ErrorSummary'] = Processing_Variables['Tools_and_Templates_FGDB_Path'] + r'\Template_ErrorSummary'

    Processing_Variables['DomainErrorTable']         = INPUT_FOR_TESTING + r'_DomainErrors'
    Processing_Variables['RangeErrorTable']          = INPUT_FOR_TESTING + r'_RangeErrors'
    Processing_Variables['CriticalRangeErrorTable']  = INPUT_FOR_TESTING + r'_CriticalRangeErrors'
    Processing_Variables['RowErrorTable']            = INPUT_FOR_TESTING + r'_RowErrors'
    Processing_Variables['GeometryErrorTable']       = INPUT_FOR_TESTING + r'_GeometryErrors'
    Processing_Variables['MandatoryFieldErrorTable'] = INPUT_FOR_TESTING + r'_MandatoryFieldErrors'
    Processing_Variables['SummaryErrorTable']        = INPUT_FOR_TESTING + r'_SummaryErrors'
  
    Processing_Variables['Albers_PRJ']           = Processing_Variables['Project_Base_Path'] + r'\Scripts\NAD_1983_BC_Environment_Albers.prj'

    if arcpy.Exists(INPUT_FOR_TESTING):
        CPMessage(' - Input Exists:' + INPUT_FOR_TESTING, 4)
    else:
        CPMessage('** ERROR: INPUT DOES NOT EXIST: ' + INPUT_FOR_TESTING, 0)
        sys.exit(100)

   
    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# CPMessage(Message)
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CPMessage(Message, MessageLevel):
    # Based on the logging level, add messages to the message queues.
    #   Text Run Log   : Implemented
    #   Console        : Implemented
    #   GeoProcessor   : Implemented
    #   Processing Log : Implemented (with e-mail)
    #   FGDB           : Implemented
    
    # Message Levels:
    #   0    : Fatal Errors
    #   1    : Top Level Program
    #   2    : Main routines and events
    #   3    : Warnings
    #   4    : General Information
    #   5    : Debug level detail
    
    # Send to the screen
    if MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToScreen(Message, MessageLevel)    
    
        # Send to the Processing Log
        #### The log is extremely large with this script and may overflow memory if all levels are appended.
        if 'SendRunLogToEmail' in Processing_Variables:
            Processing_Variables['PROCESS_LOG'].append((Message, MessageLevel))
    
    # Send to the Log File
    if 'MessageQueueFile' in Processing_Variables and MessageLevel <= Processing_Variables['Log_Level']:
        WriteOutputToFile(Message, Processing_Variables['MessageQueueFile'])

    # Send to the GeoProcessor Message Queue
    #### At ArcGIS 9.3 SP1 there is a bug that also causes any GeoProcessor messages to be echoed to the console.
    #    Added the ToolBoxRun parameter to allow the script to add messages to the GeoProcessor Messaging properly
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and Processing_Variables['ToolBoxRun'] == 'TRUE':
        if MessageLevel == 0:
            arcpy.AddError(Message)
        if MessageLevel == 1 or MessageLevel == 2:
            arcpy.AddMessage(Message)
        if MessageLevel == 3:
            arcpy.AddWarning(Message)

    # Send to the FGDB Processing Log
    if MessageLevel < 4 and MessageLevel <= Processing_Variables['Log_Level'] and 'Log_FGDB_Table' in Processing_Variables:
        rows = arcpy.InsertCursor(Processing_Variables['Log_FGDB_Table'])

        row = rows.NewRow()
        row.SetValue("Computer_Name", COMPUTERNAME)
        row.SetValue("User_Name", USERNAME)
        row.SetValue("Script_Name", PYTHON_SCRIPT)
        row.SetValue("Event_Time", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
        row.SetValue("Message_Level", MessageLevel)
        row.SetValue("Message", Message)
        row.SetValue("Start_Time", START_TIME_SQL)
        
        rows.InsertRow(row)
        del row
        del rows
    

    return(1)


#---------------------------------------------------------------------------------------------------------
# WriteOutputToScreen(Output_Comment, Header_Style):
#
#   Prints the Output_Comment to the screen with the appropriate formatting
#       0 : "******" Used for Error Sections
#       1 : "======" Used for the Program's Header and Footer
#       2 : "------" Used for New Sections
#       3+:          No delimitation
#   
#---------------------------------------------------------------------------------------------------------
def WriteOutputToScreen(Output_Comment, Header_Style):
    if Header_Style == 0:
        # Errors get beeps, too!
        print BEEP * 3 + "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 2:
        print "\n" + "-" * 79 + "\n"
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"
        
    #print Output_Comment
    arcpy.AddMessage(Output_Comment)
    
    if Header_Style == 0:
        print "\n" * 2 + "*" * 79 + "\n" * 2
    if Header_Style == 1:
        print "\n" + "=" * 79 + "\n"


#---------------------------------------------------------------------------------------------------------
# WriteOutputToFile(Output_Comment, Run_Log)
#   Record to Log 
#
#
#---------------------------------------------------------------------------------------------------------
def WriteOutputToFile(Output_Comment, Run_Log):
# Run logs are not compatible with concurrency - only one process is expected in debug mode at a time!
# First try to append to an existing file - if that fails create a new file.
    ReturnValue = 1
    try:
        if os.path.isfile(Run_Log) == 1:
            # try to open up the fileName 
            file = open(Run_Log,'a') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Appended comment to the file"
        else:
            # try to open up the fileName 
            file = open(Run_Log,'w') 

            # write String to the fileName
            file.write(COMPUTERNAME + " " + START_TIME_SQL + ' (' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ') : ' + Output_Comment + '\n')

            # Close the file
            file.close()
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created new file and put in the comment"
    except:
        # All failures are unexpected in this case...  
        # Maybe an invalid drive name, read only file or something?
        Processing_Variables['PROCESS_STATUS'] = -1
        Processing_Variables['PROCESS_INFO']   = "**ERROR: Can't write to file: " + Run_Log   
        print BEEP + BEEP + BEEP
        ReturnValue = -1
    return(ReturnValue)
        

#---------------------------------------------------------------------------------------------------------
# SetupWorkspace93()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SetupWorkspace93():
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "SetupWorkspace93"  

    CPMessage('    - Setting Up Workspace', 2)
    
    if arcpy.Exists(INPUT_FOR_TESTING):
        CPMessage('        -- User-specified input feature class exists.', 4)
    else:
        CPMessage(' **** User-specified input feature class does not exist: ' + INPUT_FOR_TESTING, 0)
        sys.exit(100)

    if not arcpy.Exists(Processing_Variables['ScratchFGDB_Path']):
        CPMessage('        -- Creating Scratch FGDB', 4)
        try:
            arcpy.CreateFileGDB_management(os.path.split(Processing_Variables['ScratchFGDB_Path'])[0], os.path.split(Processing_Variables['ScratchFGDB_Path'])[1], '9.3')
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created Scratch FGDB"
        except:
            Processing_Variables['PROCESS_STATUS'] = -1
            Processing_Variables['PROCESS_INFO']   = "Error Creating Scratch FGDB"
            CPMessage('** Can not find or create ' + Processing_Variables['ScratchFGDB_Path'], 0)
            sys.exit(100)
    else:
        CPMessage('        -- Scratch FGDB exists, nothing to do.', 4)
       
    if arcpy.Exists(Processing_Variables['OperationalFGDB_Path']):
        CPMessage('        -- Operational FGDB exists.', 4)
    else:
        CPMessage(' **** Operational FGDB does not exist: ' + Processing_Variables['OperationalFGDB_Path'], 0)
        sys.exit(100)

    if arcpy.Exists(Processing_Variables['Tools_and_Templates_FGDB_Path']):
        CPMessage('        -- Tools and Templates FGDB exists.', 4)
    else:
        CPMessage(' **** Tools and Templates FGDB does not exist: ' + Processing_Variables['Tools_and_Templates_FGDB_Path'], 0)
        sys.exit(100)


    # Create/overwrite the old error tables
    CPMessage('        -- Initializing Domain Error Table.', 4)
    if arcpy.Exists(Processing_Variables['DomainErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['DomainErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['DomainErrorTable'], 3)
            time.sleep(10)        
        arcpy.Delete_management(Processing_Variables['DomainErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['DomainErrorTable'])[0], os.path.split(Processing_Variables['DomainErrorTable'])[1], Processing_Variables['Template_ErrorTable'])

    CPMessage('        -- Initializing Range Error Table.', 4)
    if arcpy.Exists(Processing_Variables['RangeErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['RangeErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['RangeErrorTable'], 3)
            time.sleep(10)    
        arcpy.Delete_management(Processing_Variables['RangeErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['RangeErrorTable'])[0], os.path.split(Processing_Variables['RangeErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    CPMessage('        -- Initializing Critical Range Error Table.', 4)
    if arcpy.Exists(Processing_Variables['CriticalRangeErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['CriticalRangeErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['CriticalRangeErrorTable'], 3)
            time.sleep(10)    
        arcpy.Delete_management(Processing_Variables['CriticalRangeErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['CriticalRangeErrorTable'])[0], os.path.split(Processing_Variables['CriticalRangeErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    CPMessage('        -- Initializing Row Error Table.', 4)
    if arcpy.Exists(Processing_Variables['RowErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['RowErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['RowErrorTable'], 3)
            time.sleep(10)      
        arcpy.Delete_management(Processing_Variables['RowErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['RowErrorTable'])[0], os.path.split(Processing_Variables['RowErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    if CHECK_MANDATORY_FIELDS:
        CPMessage('        -- Initializing Mandatory Field Error Table.', 4)
        if arcpy.Exists(Processing_Variables['MandatoryFieldErrorTable']):
            while not arcpy.TestSchemaLock(Processing_Variables['MandatoryFieldErrorTable']):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['MandatoryFieldErrorTable'], 3)
                time.sleep(10)        
            arcpy.Delete_management(Processing_Variables['MandatoryFieldErrorTable'])
        arcpy.CreateTable_management(os.path.split(Processing_Variables['MandatoryFieldErrorTable'])[0], os.path.split(Processing_Variables['MandatoryFieldErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    if CHECK_GEOMETRY:
        CPMessage('        -- Initializing Geometry Error Table.', 4)
        if arcpy.Exists(Processing_Variables['GeometryErrorTable']):
            while not arcpy.TestSchemaLock(Processing_Variables['GeometryErrorTable']):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['GeometryErrorTable'], 3)
                time.sleep(10)        
            arcpy.Delete_management(Processing_Variables['GeometryErrorTable'])

    CPMessage('        -- Initializing Summary Error Table.', 4)
    if arcpy.Exists(Processing_Variables['SummaryErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['SummaryErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['SummaryErrorTable'], 3)
            time.sleep(10)
        arcpy.Delete_management(Processing_Variables['SummaryErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['SummaryErrorTable'])[0], os.path.split(Processing_Variables['SummaryErrorTable'])[1], Processing_Variables['Template_ErrorSummary'])
    


    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# SetupWorkspace101()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SetupWorkspace101():
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "SetupWorkspace101"  

    CPMessage('    - Setting Up Workspace', 2)
    
    if arcpy.Exists(INPUT_FOR_TESTING):
        CPMessage('        -- User-specified input feature class exists.', 4)
    else:
        CPMessage(' **** User-specified input feature class does not exist: ' + INPUT_FOR_TESTING, 0)
        sys.exit(100)

    if not arcpy.Exists(Processing_Variables['ScratchFGDB_Path']):
        CPMessage('        -- Creating Scratch FGDB', 4)
        try:
            arcpy.CreateFileGDB_management(os.path.split(Processing_Variables['ScratchFGDB_Path'])[0], os.path.split(Processing_Variables['ScratchFGDB_Path'])[1], '9.3')
            Processing_Variables['PROCESS_STATUS'] = 1
            Processing_Variables['PROCESS_INFO']   = "Created Scratch FGDB"
        except:
            Processing_Variables['PROCESS_STATUS'] = -1
            Processing_Variables['PROCESS_INFO']   = "Error Creating Scratch FGDB"
            CPMessage('** Can not find or create ' + Processing_Variables['ScratchFGDB_Path'], 0)
            sys.exit(100)
    else:
        CPMessage('        -- Scratch FGDB exists, nothing to do.', 4)
       
    if arcpy.Exists(Processing_Variables['OperationalFGDB_Path']):
        CPMessage('        -- Operational FGDB exists.', 4)
    else:
        CPMessage(' **** Operational FGDB does not exist: ' + Processing_Variables['OperationalFGDB_Path'], 0)
        sys.exit(100)

    if arcpy.Exists(Processing_Variables['Tools_and_Templates_FGDB_Path']):
        CPMessage('        -- Tools and Templates FGDB exists.', 4)
    else:
        CPMessage(' **** Tools and Templates FGDB does not exist: ' + Processing_Variables['Tools_and_Templates_FGDB_Path'], 0)
        sys.exit(100)


    # Create/overwrite the old error tables
    CPMessage('        -- Initializing Domain Error Table.', 4)
    if arcpy.Exists(Processing_Variables['DomainErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['DomainErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['DomainErrorTable'], 3)
            time.sleep(10)        
        arcpy.Delete_management(Processing_Variables['DomainErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['DomainErrorTable'])[0], os.path.split(Processing_Variables['DomainErrorTable'])[1], Processing_Variables['Template_ErrorTable'])

    CPMessage('        -- Initializing Range Error Table.', 4)
    if arcpy.Exists(Processing_Variables['RangeErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['RangeErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['RangeErrorTable'], 3)
            time.sleep(10)    
        arcpy.Delete_management(Processing_Variables['RangeErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['RangeErrorTable'])[0], os.path.split(Processing_Variables['RangeErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    CPMessage('        -- Initializing Critical Range Error Table.', 4)
    if arcpy.Exists(Processing_Variables['CriticalRangeErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['CriticalRangeErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['CriticalRangeErrorTable'], 3)
            time.sleep(10)    
        arcpy.Delete_management(Processing_Variables['CriticalRangeErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['CriticalRangeErrorTable'])[0], os.path.split(Processing_Variables['CriticalRangeErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    CPMessage('        -- Initializing Row Error Table.', 4)
    if arcpy.Exists(Processing_Variables['RowErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['RowErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['RowErrorTable'], 3)
            time.sleep(10)      
        arcpy.Delete_management(Processing_Variables['RowErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['RowErrorTable'])[0], os.path.split(Processing_Variables['RowErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    if CHECK_MANDATORY_FIELDS:
        CPMessage('        -- Initializing Mandatory Field Error Table.', 4)
        if arcpy.Exists(Processing_Variables['MandatoryFieldErrorTable']):
            while not arcpy.TestSchemaLock(Processing_Variables['MandatoryFieldErrorTable']):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['MandatoryFieldErrorTable'], 3)
                time.sleep(10)        
            arcpy.Delete_management(Processing_Variables['MandatoryFieldErrorTable'])
        arcpy.CreateTable_management(os.path.split(Processing_Variables['MandatoryFieldErrorTable'])[0], os.path.split(Processing_Variables['MandatoryFieldErrorTable'])[1], Processing_Variables['Template_ErrorTable'])
   
    if CHECK_GEOMETRY:
        CPMessage('        -- Initializing Geometry Error Table.', 4)
        if arcpy.Exists(Processing_Variables['GeometryErrorTable']):
            while not arcpy.TestSchemaLock(Processing_Variables['GeometryErrorTable']):
                CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['GeometryErrorTable'], 3)
                time.sleep(10)        
            arcpy.Delete_management(Processing_Variables['GeometryErrorTable'])
   
    CPMessage('        -- Initializing Summary Error Table.', 4)
    if arcpy.Exists(Processing_Variables['SummaryErrorTable']):
        while not arcpy.TestSchemaLock(Processing_Variables['SummaryErrorTable']):
            CPMessage('** Schema Locked - waiting 10 seconds: ' + Processing_Variables['SummaryErrorTable'], 3)
            time.sleep(10)
        arcpy.Delete_management(Processing_Variables['SummaryErrorTable'])
    arcpy.CreateTable_management(os.path.split(Processing_Variables['SummaryErrorTable'])[0], os.path.split(Processing_Variables['SummaryErrorTable'])[1], Processing_Variables['Template_ErrorSummary'])
    


    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# SendRunLogToEmail()
#
#       Test if we need to create the workspace file geodatabases, and then do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def SendRunLogToEmail(mailto):
    import smtplib
    CPMessage('  * Sending log to ' + mailto, 2)
    mailfrom = 'GIS_Python@cloverpoint.com'
    # mailto = 'mneal@cloverpoint.com'
    mailsubject = 'Run Log: ' + os.path.split(PYTHON_SCRIPT)[1] + ' (' + COMPUTERNAME + '\\' + USERNAME + ')'
    mailbody = "This is an automatic run log.\n\n"
    
    for LogEntry in Processing_Variables['PROCESS_LOG']:
        mailbody = mailbody + '\n' + LogEntry[0]

    mailmsg = ("From: " + mailfrom + "\r\nTo: " + mailto + "\r\nSubject: " + mailsubject + "\r\n" + mailbody)

    mailserver = smtplib.SMTP('shawmail.gv.shawcable.net')

    # SMTP.login(user, password)

    mailserver.sendmail(mailfrom, mailto, mailmsg)
    mailserver.quit()
    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# SanitizeElapsedTime(number_of_seconds)
#
# routine to convert a raw number of seconds into a human-friendly statement of duration
#---------------------------------------------------------------------------------------------------------

def SanitizeElapsedTime(dtInput):

    if dtInput < 120:
        strElapsedTime = stra(int(dtInput)) + ' seconds.'
        
    elif dtInput < 3600:
        strElapsedTime = stra(int(dtInput)) + ' seconds (' + stra(round(dtInput / 60, 1)) + ' minutes).'
        
    else:
        strElapsedTime = stra(int(dtInput)) + ' seconds (' + stra(round(dtInput / 3600, 2)) + ' hours).'
        
    return strElapsedTime


#=========================================================================================================
#  Special Routines
#=========================================================================================================

#---------------------------------------------------------------------------------------------------------
# LoadEverything93()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def LoadEverything93():
    if CHECK_MANDATORY_FIELDS:
        CPMessage('    - Loading Domains, Ranges, Critical Ranges and Mandatory Fields to Memory', 2)
    else:
        CPMessage('    - Loading Domains, Ranges, and Critical Ranges to Memory', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "LoadEverything93" 

    OriginalWorkspace = arcpy.Workspace
    
    # Get a list of all the domains in the Tools and Templates FGDB based on naming convention.
    Domains = {}
    
    
    arcpy.Workspace = Processing_Variables['Tools_and_Templates_FGDB_Path']
    Domain_List = arcpy.ListTables('Domain_*')
    
    CPMessage('        - There are ' + stra(len(Domain_List)) + ' domains to load.', 4)
    
    
    # Load each domain into a Python Dictionary
    for Domain in Domain_List:
        Domain_Name = Domain.split('Domain_')[1].upper()     
        result = arcpy.GetCount_management(Domain)
        Number_of_Features = int(result.GetOutput(0))
        Number_Processed = 0

        CPMessage('        - Getting individual values for ' + Domain_Name + ' (' + stra(Number_of_Features) + ' codes).', 4)
        rows = arcpy.SearchCursor(Domain)
        row = rows.Next()
        
        Domain_Values = []
        
        while row:
            if Number_Processed % 1000 == 0:
                CPMessage('            - Processed ' + stra(Number_Processed), 5)
            Current_Value = row.GetValue('CODE')
            CPMessage('                - ' + stra(Current_Value), 5)
            Domain_Values.append(Current_Value)
            
            Number_Processed = Number_Processed + 1
            row = rows.Next()
        CPMessage('            - Processed all ' + stra(Number_Processed), 5)
        del row
        del rows
        
        Domains[Domain_Name] = Domain_Values
        
    
    #### Update to match the Metadata tables, not the QC_* ones
    Domain_Cross_Reference = {}
    Ranges = {}
    CriticalRanges = {}
    if CHECK_MANDATORY_FIELDS:
        Mandatory_Fields = []
        Mandatory_Fields_T = []
        Mandatory_Fields_EM = []
        Mandatory_Fields_SEI = []

    result = arcpy.GetCount_management(Processing_Variables['TEIS_Metadata_Items'])
    Number_of_Features = int(result.GetOutput(0))
    CPMessage('        - Getting Domain Cross References, Ranges, Critical Ranges and Mandatory Fields for ' + stra(Number_of_Features) + ' items.', 4)
    Number_Processed = 0

    rows = arcpy.SearchCursor(Processing_Variables['TEIS_Metadata_Items'])
    row = rows.Next()
    while row:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + stra(Number_Processed), 4)

        ########################
        ##### Load Domains #####
        ########################

        Current_Field_Name  = row.GetValue('FGDB_Name')
        Current_Field_Name  = Current_Field_Name.upper()
        Current_Domain_Name = row.GetValue('Item_Domain_Name')
        if Current_Domain_Name in [None, '']:
            Current_Domain_Name = 'NO DOMAIN FOUND'
        Current_Domain_Name = Current_Domain_Name.upper()
        CPMessage('                - ' + Current_Field_Name + ' : ' + Current_Domain_Name, 5)
        
        if Current_Domain_Name not in Domains.keys() and Current_Domain_Name != 'NO DOMAIN FOUND':
            CPMessage('** Warning Domain Name Referenced without Values: ' + Current_Domain_Name, 0)
        elif Current_Domain_Name in Domains.keys():
            Domain_Cross_Reference[Current_Field_Name] = Current_Domain_Name
        
        #######################
        ##### Load Ranges #####
        #######################

        Current_Field_Name  = row.GetValue('FGDB_Name')
        Current_Field_Name  = Current_Field_Name.upper()
        Current_Range_Minimum = row.GetValue('Item_Range_Minimum')
        Current_Range_Maximum = row.GetValue('Item_Range_Maximum')
        CPMessage('            - ' + Current_Field_Name + ' : ' + stra(Current_Range_Minimum) + ' to ' + stra(Current_Range_Maximum), 5)
        
        if Current_Range_Minimum > Current_Range_Maximum:
            CPMessage('** Warning Range Values Invalid (may be swapped): ' + Current_Field_Name, 3)
        elif Current_Range_Minimum != None and Current_Range_Maximum != None:
            Ranges[Current_Field_Name] = {'Minimum':  Current_Range_Minimum, 'Maximum': Current_Range_Maximum}
        elif Current_Range_Minimum != None or Current_Range_Maximum != None:
            CPMessage('** Warning Range Values Invalid (min and max not defined): ' + Current_Field_Name, 3)

        ################################
        ##### Load Critical Ranges #####
        ################################

        Current_Field_Name    = row.GetValue('FGDB_Name')
        Current_Field_Name    = Current_Field_Name.upper()
        Current_Oracle_Type   = row.GetValue('Oracle_Type')
        Current_Oracle_Length = row.GetValue('Oracle_Length')
        Current_Short_Tbl_Num = row.GetValue('TEIS_Master_Short_Tbl')
        Current_Long_Tbl_Num  = row.GetValue('TEIS_Master_Long_Tbl')
        Current_Proj_Bnd_Num  = row.GetValue('TEIS_Master_Project_Boundaries')

        if (Current_Short_Tbl_Num != None or Current_Proj_Bnd_Num != None) and Current_Oracle_Type == 'NUMBER':
            Current_Critical_Range_Maximum = 10 ** Current_Oracle_Length - 1
            Current_Critical_Range_Minimum = 0
            CriticalRanges[Current_Field_Name] = {'Minimum':  Current_Critical_Range_Minimum, 'Maximum': Current_Critical_Range_Maximum}
            CPMessage('            - ' + Current_Field_Name + ' : ' + stra(Current_Critical_Range_Minimum) + ' to ' + stra(Current_Critical_Range_Maximum), 5)

        #################################
        ##### Load Mandatory Fields #####
        #################################

        if CHECK_MANDATORY_FIELDS:
            Current_Field_Name  = row.GetValue('FGDB_Name')
            Current_Field_Name  = Current_Field_Name.upper()
            Current_Mandatory = row.GetValue('Item_Mandatory')
            Current_Mandatory_T = row.GetValue('Item_Mandatory_T')
            Current_Mandatory_EM = row.GetValue('Item_Mandatory_EM')
            Current_Mandatory_SEI = row.GetValue('Item_Mandatory_SEI')
            CPMessage('            - ' + Current_Field_Name + ' : ', 5)

            if Current_Mandatory == 'Y':
                CPMessage('                - Mandatory for all project types', 5)
                Mandatory_Fields.append(Current_Field_Name)
            if Current_Mandatory_T == 'Y':
                CPMessage('                - Mandatory for %T% project types', 5)
                Mandatory_Fields_T.append(Current_Field_Name)
            if Current_Mandatory_EM == 'Y':
                CPMessage('                - Mandatory for %EM% project types', 5)
                Mandatory_Fields_EM.append(Current_Field_Name)
            if Current_Mandatory_SEI == 'Y':
                CPMessage('                - Mandatory for %SEI% project types', 5)
                Mandatory_Fields_SEI.append(Current_Field_Name)

        Number_Processed = Number_Processed + 1
        row = rows.Next()
    CPMessage('            - Processed all ' + stra(Number_Processed), 4)
    del row
    del rows
 
    Processing_Variables['FieldsXDomains'] = dict(Domain_Cross_Reference)
    Processing_Variables['Domains'] = Domains
    Processing_Variables['Ranges'] = Ranges
    Processing_Variables['CriticalRanges'] = CriticalRanges
    Processing_Variables['Rows'] = ['TDEC_1','TDEC_2','TDEC_3','SDEC_1','SDEC_2','SDEC_3','TTEX_1','TTEX_1A','TTEX_1B','TTEX_1C','TTEX_2','TTEX_2A','TTEX_2B','TTEX_2C','TTEX_3','TTEX_3A','TTEX_3B','TTEX_3C','TTTEX_1','TTTEX_1A','TTTEX_1B','TTTEX_1C','TTTEX_2','TTTEX_2A','TTTEX_2B','TTTEX_2C','TTTEX_3','TTTEX_3A','TTTEX_3B','TTTEX_3C','STTEX_1','STTEX_1A','STTEX_1B','STTEX_1C','STTEX_2','STTEX_2A','STTEX_2B','STTEX_2C','STTEX_3','STTEX_3A','STTEX_3B','STTEX_3C','SURF_E1','SURF_E1A','SURF_E1B','SURF_E1C','SURF_E2','SURF_E2A','SURF_E2B','SURF_E2C','SURF_E3','SURF_E3A','SURF_E3B','SURF_E3C','TSURF_E1','TSURF_E1A','TSURF_E1B','TSURF_E1C','TSURF_E2','TSURF_E2A','TSURF_E2B','TSURF_E2C','TSURF_E3','TSURF_E3A','TSURF_E3B','TSURF_E3C','SSURF_E1','SSURF_E1A','SSURF_E1B','SSURF_E1C','SSURF_E2','SSURF_E2A','SSURF_E2B','SSURF_E2C','SSURF_E3','SSURF_E3A','SSURF_E3B','SSURF_E3C','GEOP_SCM1','GEOP_INZ1','GEOP_SCM1A','GEOP_SCM1B','GEOP_SCM1C','GEOP_SCM2','GEOP_INZ2','GEOP_SCM2A','GEOP_SCM2B','GEOP_SCM2C','GEOP_SCM3','GEOP_INZ3','GEOP_SCM3A','GEOP_SCM3B','GEOP_SCM3C','SLPLL_1','SLPUL_1','SLPLL_2','SLPUL_2','SITE_S1','SITE_S2','SITE_S3']
    if CHECK_MANDATORY_FIELDS:
        Processing_Variables['Mandatory_Fields'] = Mandatory_Fields
        Processing_Variables['Mandatory_Fields_T'] = Mandatory_Fields_T
        Processing_Variables['Mandatory_Fields_EM'] = Mandatory_Fields_EM
        Processing_Variables['Mandatory_Fields_SEI'] = Mandatory_Fields_SEI
    # Domains is a dictionary of keys with a field name and values as a list.
    
    # print '=' * 79
    # print Processing_Variables['Domains']
    # print '=' * 79
    
    arcpy.Workspace = OriginalWorkspace
    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# LoadEverything101()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def LoadEverything101():
    if CHECK_MANDATORY_FIELDS:
        CPMessage('    - Loading Domains, Ranges, Critical Ranges and Mandatory Fields to Memory', 2)
    else:
        CPMessage('    - Loading Domains, Ranges, and Critical Ranges to Memory', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "LoadEverything101" 

    OriginalWorkspace = arcpy.env.workspace
    
    # Get a list of all the domains in the Tools and Templates FGDB based on naming convention.
    Domains = {}
    
    
    arcpy.env.workspace = Processing_Variables['Tools_and_Templates_FGDB_Path']
    Domain_List = arcpy.ListTables('Domain_*')
    
    CPMessage('        - There are ' + stra(len(Domain_List)) + ' domains to load.', 4)
    
    
    # Load each domain into a Python Dictionary
    for Domain in Domain_List:
        Domain_Name = Domain.split('Domain_')[1].upper()     
        Number_of_Features = int(arcpy.GetCount_management(Domain).getOutput(0))
        Number_Processed = 0

        CPMessage('        - Getting individual values for ' + Domain_Name + ' (' + stra(Number_of_Features) + ' codes).', 4)

        Domain_Values = []

        exec_string = '''with arcpy.da.SearchCursor(Domain, ['CODE']) as cursor:
    for row in cursor:
        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + stra(Number_Processed), 5)
        Current_Value = row[0]
        CPMessage('                - ' + stra(Current_Value), 5)
        Domain_Values.append(Current_Value)

        Number_Processed = Number_Processed + 1'''

        exec(exec_string)
        CPMessage('            - Processed all ' + stra(Number_Processed), 5)
        del row
        del cursor
        
        Domains[Domain_Name] = Domain_Values
        
    
    #### Update to match the Metadata tables, not the QC_* ones
    Domain_Cross_Reference = {}
    Ranges = {}
    CriticalRanges = {}
    if CHECK_MANDATORY_FIELDS:
        Mandatory_Fields = []
        Mandatory_Fields_T = []
        Mandatory_Fields_EM = []
        Mandatory_Fields_SEI = []

    Number_of_Features = int(arcpy.GetCount_management(Processing_Variables['TEIS_Metadata_Items']).getOutput(0))
    CPMessage('        - Getting Domain Cross References, Ranges, Critical Ranges and Mandatory Fields for ' + stra(Number_of_Features) + ' items.', 4)
    Number_Processed = 0

    exec_string = '''with arcpy.da.SearchCursor(Processing_Variables['TEIS_Metadata_Items'], ['FGDB_Name','Item_Domain_Name','Item_Range_Minimum','Item_Range_Maximum','Oracle_Type','Oracle_Length','TEIS_Master_Short_Tbl','TEIS_Master_Long_Tbl','TEIS_Master_Project_Boundaries','Item_Mandatory','Item_Mandatory_T','Item_Mandatory_EM','Item_Mandatory_SEI']) as cursor:
    for row in cursor:

        if Number_Processed % 1000 == 0:
            CPMessage('            - Processed ' + stra(Number_Processed), 4)

        ########################
        ##### Load Domains #####
        ########################

        Current_Field_Name  = row[0]
        Current_Field_Name  = Current_Field_Name.upper()
        Current_Domain_Name = row[1]
        if Current_Domain_Name in [None, '']:
            Current_Domain_Name = 'NO DOMAIN FOUND'
        Current_Domain_Name = Current_Domain_Name.upper()
        CPMessage('                - ' + Current_Field_Name + ' : ' + Current_Domain_Name, 5)

        if Current_Domain_Name not in Domains.keys() and Current_Domain_Name != 'NO DOMAIN FOUND':
            CPMessage('** Warning Domain Name Referenced without Values: ' + Current_Domain_Name, 0)
        elif Current_Domain_Name in Domains.keys():
            Domain_Cross_Reference[Current_Field_Name] = Current_Domain_Name

        #######################
        ##### Load Ranges #####
        #######################

        Current_Range_Minimum = row[2]
        Current_Range_Maximum = row[3]
        CPMessage('            - ' + Current_Field_Name + ' : ' + stra(Current_Range_Minimum) + ' to ' + stra(Current_Range_Maximum), 5)

        if Current_Range_Minimum > Current_Range_Maximum:
            CPMessage('** Warning Range Values Invalid (may be swapped): ' + Current_Field_Name, 3)
        elif Current_Range_Minimum != None and Current_Range_Maximum != None:
            Ranges[Current_Field_Name] = {'Minimum':  Current_Range_Minimum, 'Maximum': Current_Range_Maximum}
        elif Current_Range_Minimum != None or Current_Range_Maximum != None:
            CPMessage('** Warning Range Values Invalid (min and max not defined): ' + Current_Field_Name, 3)

        ################################
        ##### Load Critical Ranges #####
        ################################

        Current_Oracle_Type   = row[4]
        Current_Oracle_Length = row[5]
        Current_Short_Tbl_Num = row[6]
        Current_Long_Tbl_Num  = row[7]
        Current_Proj_Bnd_Num  = row[8]

        if (Current_Short_Tbl_Num != None or Current_Proj_Bnd_Num != None) and Current_Oracle_Type == 'NUMBER':
            Current_Critical_Range_Maximum = 10 ** Current_Oracle_Length - 1
            Current_Critical_Range_Minimum = 0
            CriticalRanges[Current_Field_Name] = {'Minimum':  Current_Critical_Range_Minimum, 'Maximum': Current_Critical_Range_Maximum}
            CPMessage('            - ' + Current_Field_Name + ' : ' + stra(Current_Critical_Range_Minimum) + ' to ' + stra(Current_Critical_Range_Maximum), 5)

        #################################
        ##### Load Mandatory Fields #####
        #################################

        if CHECK_MANDATORY_FIELDS:
            Current_Mandatory = row[9]
            Current_Mandatory_T = row[10]
            Current_Mandatory_EM = row[11]
            Current_Mandatory_SEI = row[12]
            CPMessage('            - ' + Current_Field_Name + ' : ', 5)

            if Current_Mandatory == 'Y':
                CPMessage('                - Mandatory for all project types', 5)
                Mandatory_Fields.append(Current_Field_Name)
            if Current_Mandatory_T == 'Y':
                CPMessage('                - Mandatory for %T% project types', 5)
                Mandatory_Fields_T.append(Current_Field_Name)
            if Current_Mandatory_EM == 'Y':
                CPMessage('                - Mandatory for %EM% project types', 5)
                Mandatory_Fields_EM.append(Current_Field_Name)
            if Current_Mandatory_SEI == 'Y':
                CPMessage('                - Mandatory for %SEI% project types', 5)
                Mandatory_Fields_SEI.append(Current_Field_Name)

        Number_Processed = Number_Processed + 1'''

    exec(exec_string)
    CPMessage('            - Processed all ' + stra(Number_Processed), 4)
    del row
    del cursor
 
    Processing_Variables['FieldsXDomains'] = dict(Domain_Cross_Reference)
    Processing_Variables['Domains'] = Domains
    Processing_Variables['Ranges'] = Ranges
    Processing_Variables['CriticalRanges'] = CriticalRanges
    Processing_Variables['Rows'] = ['TDEC_1','TDEC_2','TDEC_3','SDEC_1','SDEC_2','SDEC_3','TTEX_1','TTEX_1A','TTEX_1B','TTEX_1C','TTEX_2','TTEX_2A','TTEX_2B','TTEX_2C','TTEX_3','TTEX_3A','TTEX_3B','TTEX_3C','TTTEX_1','TTTEX_1A','TTTEX_1B','TTTEX_1C','TTTEX_2','TTTEX_2A','TTTEX_2B','TTTEX_2C','TTTEX_3','TTTEX_3A','TTTEX_3B','TTTEX_3C','STTEX_1','STTEX_1A','STTEX_1B','STTEX_1C','STTEX_2','STTEX_2A','STTEX_2B','STTEX_2C','STTEX_3','STTEX_3A','STTEX_3B','STTEX_3C','SURF_E1','SURF_E1A','SURF_E1B','SURF_E1C','SURF_E2','SURF_E2A','SURF_E2B','SURF_E2C','SURF_E3','SURF_E3A','SURF_E3B','SURF_E3C','TSURF_E1','TSURF_E1A','TSURF_E1B','TSURF_E1C','TSURF_E2','TSURF_E2A','TSURF_E2B','TSURF_E2C','TSURF_E3','TSURF_E3A','TSURF_E3B','TSURF_E3C','SSURF_E1','SSURF_E1A','SSURF_E1B','SSURF_E1C','SSURF_E2','SSURF_E2A','SSURF_E2B','SSURF_E2C','SSURF_E3','SSURF_E3A','SSURF_E3B','SSURF_E3C','GEOP_SCM1','GEOP_INZ1','GEOP_SCM1A','GEOP_SCM1B','GEOP_SCM1C','GEOP_SCM2','GEOP_INZ2','GEOP_SCM2A','GEOP_SCM2B','GEOP_SCM2C','GEOP_SCM3','GEOP_INZ3','GEOP_SCM3A','GEOP_SCM3B','GEOP_SCM3C','SLPLL_1','SLPUL_1','SLPLL_2','SLPUL_2','SITE_S1','SITE_S2','SITE_S3']

    if CHECK_MANDATORY_FIELDS:
        Processing_Variables['Mandatory_Fields'] = Mandatory_Fields
        Processing_Variables['Mandatory_Fields_T'] = Mandatory_Fields_T
        Processing_Variables['Mandatory_Fields_EM'] = Mandatory_Fields_EM
        Processing_Variables['Mandatory_Fields_SEI'] = Mandatory_Fields_SEI
    # Domains is a dictionary of keys with a field name and values as a list.
    
    # print '=' * 79
    # print Processing_Variables['Domains']
    # print '=' * 79
    
    arcpy.env.workspace = OriginalWorkspace
    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# CheckEverything93()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CheckEverything93():
    if CHECK_MANDATORY_FIELDS:
        CPMessage('    - Checking Domains, Ranges, Critical Ranges, Rows, Mandatory Fields', 2)
    else:
        CPMessage('    - Checking Domains, Ranges, Critical Ranges, Rows', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "CheckEverything93" 

    det_rows = arcpy.InsertCursor(Processing_Variables['DomainErrorTable'])
    ret_rows = arcpy.InsertCursor(Processing_Variables['RangeErrorTable'])
    cret_rows = arcpy.InsertCursor(Processing_Variables['CriticalRangeErrorTable'])
    rowet_rows = arcpy.InsertCursor(Processing_Variables['RowErrorTable'])
    if CHECK_MANDATORY_FIELDS:
        mfet_rows = arcpy.InsertCursor(Processing_Variables['MandatoryFieldErrorTable'])
    set_rows = arcpy.InsertCursor(Processing_Variables['SummaryErrorTable'])
    if arcpy.Exists(Processing_Variables['GeometryErrorTable']):
        get_rows = arcpy.InsertCursor(Processing_Variables['GeometryErrorTable'])

    # Get the Field List
    Field_List = arcpy.ListFields(INPUT_FOR_TESTING)
    Field_Name_List = []
    for Field in Field_List:
        Field_Name_List.append(Field.name)
    
    Domain_Check_Fields = []
    
    CPMessage('        - FieldsXDomains: ' + stra(Processing_Variables['FieldsXDomains']), 5)
    
    for Current_Field in Field_List:

        if Current_Field.Name in Processing_Variables['FieldsXDomains']:
            Domain_Check_Fields.append(Current_Field.Name)
            CPMessage('            - Checking Field: ' + stra(Current_Field.Name), 5)
        else:
            CPMessage('            - Not Checking Field: ' + stra(Current_Field.Name), 5)

    CPMessage('        - Checking Fields for Domains: ' + stra(Domain_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    Range_Check_Fields = []
    
    CPMessage('        - Ranges: ' + stra(Processing_Variables['Ranges'].keys()), 5)
    
    for Current_Field in Field_List:
        if Current_Field.Name.upper() in Processing_Variables['Ranges']:
            Range_Check_Fields.append(Current_Field.Name)
            CPMessage('            - Checking Field: ' + stra(Current_Field.Name), 5)
        else:
            CPMessage('            - Not Checking Field: ' + stra(Current_Field.Name), 5)

    CPMessage('        - Checking Fields for Ranges: ' + stra(Range_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    Critical_Range_Check_Fields = []
    
    CPMessage('        - Critical Ranges: ' + stra(Processing_Variables['CriticalRanges'].keys()), 5)
    
    for Current_Field in Field_List:
        if Current_Field.Name.upper() in Processing_Variables['CriticalRanges']:
            Critical_Range_Check_Fields.append(Current_Field.Name)
            CPMessage('            - Checking Field: ' + stra(Current_Field.Name), 5)
        else:
            CPMessage('            - Not Checking Field: ' + stra(Current_Field.Name), 5)

    CPMessage('        - Checking Fields for Critical Ranges: ' + stra(Critical_Range_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    if CHECK_MANDATORY_FIELDS:
        Mandatory_Check_Fields = []
        Mandatory_Check_Fields_T = []
        Mandatory_Check_Fields_EM = []
        Mandatory_Check_Fields_SEI = []

        for Current_Field in Field_List:

            if Current_Field.Name in Processing_Variables['Mandatory_Fields']:
                Mandatory_Check_Fields.append(Current_Field.Name)

            if Current_Field.Name in Processing_Variables['Mandatory_Fields_T']:
                Mandatory_Check_Fields_T.append(Current_Field.Name)

            if Current_Field.Name in Processing_Variables['Mandatory_Fields_EM']:
                Mandatory_Check_Fields_EM.append(Current_Field.Name)

            if Current_Field.Name in Processing_Variables['Mandatory_Fields_SEI']:
                Mandatory_Check_Fields_SEI.append(Current_Field.Name)

        CPMessage('        - Checking Mandatory Fields: ' + stra(Mandatory_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)
        CPMessage('        - Checking Mandatory Fields for %T% Project Types: ' + stra(Mandatory_Check_Fields_T).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)
        CPMessage('        - Checking Mandatory Fields for %EM% Project Types: ' + stra(Mandatory_Check_Fields_EM).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)
        CPMessage('        - Checking Mandatory Fields for %SEI% Project Types: ' + stra(Mandatory_Check_Fields_SEI).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    #  Key on TEIS_ID, or, for tables (like Project_Details) without TEIS_ID, key on the OID field
    Object_ID_Field = arcpy.ListFields(INPUT_FOR_TESTING, "*", "OID")[0].Name

    if "TEIS_ID" in Field_Name_List:
        Object_ID_Field = "TEIS_ID"

    CPMessage('        - Keying on the input table\'s ' + Object_ID_Field + ' field.', 4)
    
    result = arcpy.GetCount_management(INPUT_FOR_TESTING)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Checking individual values for ' + stra(Number_of_Features) + ' features.', 4)
    rows = arcpy.SearchCursor(INPUT_FOR_TESTING)
    row = rows.Next()
    Error_List = []
    error_count = 0
    dtCalcStart = time.time()
    while row:
        summary_error_list = []
        
        #########################
        ##### Check Domains #####
        #########################

        for Field_Name in Domain_Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            #### Testing numeric domains here!
            Test_Value = stra(row.GetValue(Field_Name))
            Test_Domain_Name = Processing_Variables['FieldsXDomains'][Field_Name]
            # print ' TEST DOMAIN:', Processing_Variables['Domains'][Test_Domain_Name]
            #CPMessage('            - Testing: ' + Field_Name + ' with value ' + stra(Test_Value), 5)

            
            if Test_Value in Processing_Variables['Domains'][Test_Domain_Name] or stra(Test_Value) == 'None' or stra(Test_Value) == '' or stra(Test_Value) == ' ':
                # The value is null or in the allowed list
                #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                pass
            else:
                Current_OID = row.GetValue(Object_ID_Field)
                #CPMessage('            - Invalid Value : ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                #Error_List.append((Current_OID, 'Invalid Value in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                problem_string = 'Invalid Value in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                det_row = det_rows.NewRow()
                det_row.SetValue('CLASS', INPUT_FOR_TESTING)
                det_row.SetValue('FEATURE_ID', Current_OID)
                det_row.SetValue('PROBLEM', problem_string)
                det_rows.InsertRow(det_row)
                summary_error_list.append(problem_string)
                
        ########################
        ##### Check Ranges #####
        ########################

        for Field_Name in Range_Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            Test_Value = row.GetValue(Field_Name)
            
            Test_Range_Name = Field_Name
            # print ' TEST Range:', Processing_Variables['Ranges'][Field_Name]
            #CPMessage('            - Testing: ' + Field_Name + ' with value ' + stra(Test_Value), 5)
            
            Test_Validation = 'Number'
            
            if stra(Test_Value) in ['None', '', ' ']:
                Test_Validation = 'Empty'
                #CPMessage('            - Value OK : NULL (' + stra(Test_Value) + ').', 5)
            if Test_Validation == 'Number':
                try:
                    Test_Value = int(Test_Value)
                except:
                    Current_OID = row.GetValue(Object_ID_Field)
                    Test_Validation = 'NotNumber'
                    #CPMessage('            - Invalid Value (Not an Integer): ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, 'Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = 'Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    ret_row = ret_rows.NewRow()
                    ret_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    ret_row.SetValue('FEATURE_ID', Current_OID)
                    ret_row.SetValue('PROBLEM', problem_string)
                    ret_rows.InsertRow(ret_row)
                    summary_error_list.append(problem_string)

            if Test_Validation == 'Number':
                if (Processing_Variables['Ranges'][Test_Range_Name]['Minimum'] <= Test_Value  <= Processing_Variables['Ranges'][Test_Range_Name]['Maximum']):
                    # The value is in the allowed list
                    #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                    pass
                else:
                    Current_OID = row.GetValue(Object_ID_Field)
                    #CPMessage('            - Invalid Value : ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, 'Invalid Value (out of range) in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = 'Invalid Value (out of range) in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    ret_row = ret_rows.NewRow()
                    ret_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    ret_row.SetValue('FEATURE_ID', Current_OID)
                    ret_row.SetValue('PROBLEM', problem_string)
                    ret_rows.InsertRow(ret_row)
                    summary_error_list.append(problem_string)

        #################################
        ##### Check Critical Ranges #####
        #################################

        for Field_Name in Critical_Range_Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            Test_Value = row.GetValue(Field_Name)
            
            Test_Range_Name = Field_Name
            # print ' TEST Range:', Processing_Variables['CriticalRanges'][Field_Name]
            #CPMessage('            - Testing: ' + Field_Name + ' with value ' + stra(Test_Value), 5)
            
            Test_Validation = 'Number'
            
            if stra(Test_Value) in ['None', '', ' ']:
                Test_Validation = 'Empty'
                #CPMessage('            - Value OK : NULL (' + stra(Test_Value) + ').', 5)
            if Test_Validation == 'Number':
                try:
                    Test_Value = int(Test_Value)
                except:
                    Current_OID = row.GetValue(Object_ID_Field)
                    Test_Validation = 'NotNumber'
                    #CPMessage('            ** CRITICAL ERROR ** - Invalid Value (Not an Integer): ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, '** CRITICAL ERROR ** - Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = '** CRITICAL ERROR ** - Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    cret_row = cret_rows.NewRow()
                    cret_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    cret_row.SetValue('FEATURE_ID', Current_OID)
                    cret_row.SetValue('PROBLEM', problem_string)
                    cret_rows.InsertRow(cret_row)
                    summary_error_list.append(problem_string)

            if Test_Validation == 'Number':
                if (Processing_Variables['CriticalRanges'][Test_Range_Name]['Minimum'] <= Test_Value  <= Processing_Variables['CriticalRanges'][Test_Range_Name]['Maximum']):
                    # The value is in the allowed list
                    #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                    pass
                else:
                    Current_OID = row.GetValue(Object_ID_Field)
                    #CPMessage('            ** CRITICAL ERROR ** - Invalid Value : ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, '** CRITICAL ERROR ** - invalid number of digits in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = '** CRITICAL ERROR ** - invalid number of digits in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    cret_row = cret_rows.NewRow()
                    cret_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    cret_row.SetValue('FEATURE_ID', Current_OID)
                    cret_row.SetValue('PROBLEM', problem_string)
                    cret_rows.InsertRow(cret_row)
                    summary_error_list.append(problem_string)
        
        ######################
        ##### Check Rows #####
        ######################

        Current_OID = row.GetValue(Object_ID_Field)
        
        ## CPMessage('            - Testing OID ' + stra(Current_OID) + '.', 5)
        
        if 'TDEC_1' in Field_Name_List and 'TDEC_2' in Field_Name_List and 'TDEC_3' in Field_Name_List:
            TDEC_1 = row.GetValue('TDEC_1')
            TDEC_2 = row.GetValue('TDEC_2')
            TDEC_3 = row.GetValue('TDEC_3')
            if TDEC_1 == None:
                TDEC_1 = 0
            if TDEC_2 == None:
                TDEC_2 = 0
            if TDEC_3 == None:
                TDEC_3 = 0
            TDEC_SUM = TDEC_1 + TDEC_2 + TDEC_3
            CPMessage('                    - T-Deciles add to: ' + stra(TDEC_SUM) + '.', 5)
            if TDEC_SUM != 10 and TDEC_SUM != 0:
                #CPMessage('            - T-Deciles of OID ' + stra(Current_OID) + ' do not add to 10 or zero: ' + stra(TDEC_SUM) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile sum (TDEC_1 + TDEC_2 + TDEC_3 = ' + stra(TDEC_SUM)  + r').'))
                problem_string = 'Invalid decile sum (TDEC_1 + TDEC_2 + TDEC_3 = ' + stra(TDEC_SUM)  + r').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
                
                
            # Check that the deciles are in order (1 >= 2 >= 3)
            
            if (TDEC_1 < TDEC_2):
                #CPMessage('            - TDEC_1 of OID ' + stra(Current_OID) + ' is smaller than TDEC_2: ' + stra(TDEC_1) + ' vs ' + stra(TDEC_2) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (TDEC_1 and TDEC_2).'))
                problem_string = 'Invalid decile values in (TDEC_1 and TDEC_2).'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
            
            if (TDEC_2 < TDEC_3):
                CPMessage('            - TDEC_2 of OID ' + stra(Current_OID) + ' is smaller than TDEC_3: ' + stra(TDEC_2) + ' vs ' + stra(TDEC_3) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (TDEC_2 and TDEC_3).'))
                problem_string = 'Invalid decile values in (TDEC_2 and TDEC_3).'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)

            # Check that if there is a TDEC value, then the SURFM is filled in.
            if (TDEC_1 > 0) and ('SURFM_1' in Field_Name_List):
                SURFM_1 = stra(row.GetValue('SURFM_1'))
                if SURFM_1 in ['', 'None']:
                    #CPMessage('            - SURFM_1 of OID ' + stra(Current_OID) + ' is blank, but TDEC_1 is not.', 3)
                    #Error_List.append((Current_OID, r'SURFM_1 is unexpectedly blank.'))
                    problem_string = 'SURFM_1 is unexpectedly blank.'
                    rowet_row = rowet_rows.NewRow()
                    rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    rowet_row.SetValue('FEATURE_ID', Current_OID)
                    rowet_row.SetValue('PROBLEM', problem_string)
                    rowet_rows.InsertRow(rowet_row)
                    summary_error_list.append(problem_string)

            if (TDEC_2 > 0) and ('SURFM_2' in Field_Name_List):
                SURFM_2 = stra(row.GetValue('SURFM_2'))
                if SURFM_2 in ['', 'None']:
                    #CPMessage('            - SURFM_2 of OID ' + stra(Current_OID) + ' is blank, but TDEC_2 is not.', 3)
                    #Error_List.append((Current_OID, r'SURFM_2 is unexpectedly blank.'))
                    problem_string = 'SURFM_2 is unexpectedly blank.'
                    rowet_row = rowet_rows.NewRow()
                    rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    rowet_row.SetValue('FEATURE_ID', Current_OID)
                    rowet_row.SetValue('PROBLEM', problem_string)
                    rowet_rows.InsertRow(rowet_row)
                    summary_error_list.append(problem_string)

            if (TDEC_3 > 0) and ('SURFM_3' in Field_Name_List):
                SURFM_3 = stra(row.GetValue('SURFM_3'))
                if SURFM_3 in ['', 'None']:
                    #CPMessage('            - SURFM_3 of OID ' + stra(Current_OID) + ' is blank, but TDEC_3 is not.', 3)
                    #Error_List.append((Current_OID, r'SURFM_3 is unexpectedly blank.'))
                    problem_string = 'SURFM_3 is unexpectedly blank.'
                    rowet_row = rowet_rows.NewRow()
                    rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    rowet_row.SetValue('FEATURE_ID', Current_OID)
                    rowet_row.SetValue('PROBLEM', problem_string)
                    rowet_rows.InsertRow(rowet_row)
                    summary_error_list.append(problem_string)

           
        else:
            CPMessage('                    - T-Deciles not present.', 5)
        
        if 'SDEC_1' in Field_Name_List and 'SDEC_2' in Field_Name_List and 'SDEC_3' in Field_Name_List:
            SDEC_1 = row.GetValue('SDEC_1')
            SDEC_2 = row.GetValue('SDEC_2')
            SDEC_3 = row.GetValue('SDEC_3')
            if SDEC_1 == None:
                SDEC_1 = 0
            if SDEC_2 == None:
                SDEC_2 = 0
            if SDEC_3 == None:
                SDEC_3 = 0            
            SDEC_SUM = SDEC_1 + SDEC_2 + SDEC_3
            CPMessage('                    - S-Deciles add to: ' + stra(SDEC_SUM), 5)
            if SDEC_SUM != 10 and SDEC_SUM != 0:
                #CPMessage('            - S-Deciles of OID ' + stra(Current_OID) + ' do not add to 10 or zero: ' + stra(SDEC_SUM) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile sum (SDEC_1 + SDEC_2 + SDEC_3 = ' + stra(SDEC_SUM)  + r').'))
                problem_string = 'Invalid decile sum (SDEC_1 + SDEC_2 + SDEC_3 = ' + stra(SDEC_SUM)  + r').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)

            # Check that the deciles are in order (1 >= 2 >= 3)

            if (SDEC_1 < SDEC_2):
                #CPMessage('            - SDEC_1 of OID ' + stra(Current_OID) + ' is smaller than SDEC_2: ' + stra(SDEC_1) + ' vs ' + stra(SDEC_2) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (SDEC_1 and SDEC_2).'))
                problem_string = 'Invalid decile values in (SDEC_1 and SDEC_2).'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)

            if (SDEC_2 < SDEC_3):
                #CPMessage('            - SDEC_2 of OID ' + stra(Current_OID) + ' is smaller than SDEC_3: ' + stra(SDEC_2) + ' vs ' + stra(SDEC_3) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (SDEC_2 and SDEC_3).'))
                problem_string = 'Invalid decile values in (SDEC_2 and SDEC_3).'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)

        else:
            CPMessage('                    - S-Deciles not present.', 5)                
        

        if 'TTEX_1' in Field_Name_List and 'TTEX_1A' in Field_Name_List and 'TTEX_1B' in Field_Name_List and 'TTEX_1C' in Field_Name_List:
            TTEX_1 = stra(row.GetValue('TTEX_1'))
            TTEX_1A = row.GetValue('TTEX_1A')
            TTEX_1B = row.GetValue('TTEX_1B')
            TTEX_1C = row.GetValue('TTEX_1C')
            TTEX_1_Composite = stra(TTEX_1C) + stra(TTEX_1B) + stra(TTEX_1A)
            TTEX_1_Composite = TTEX_1_Composite.replace(' ', '')
            TTEX_1_Composite = TTEX_1_Composite.replace('None', '')
            TTEX_1 = TTEX_1.replace(' ', '')
            TTEX_1 = TTEX_1.replace('None', '')
            #CPMessage('                    - TTEX_1_Composite: ' + TTEX_1_Composite + ' vs ' + stra(TTEX_1), 5)
            if TTEX_1 != TTEX_1_Composite:
                #CPMessage('            - TTEX_1 of OID ' + stra(Current_OID) + ' does not match composite (' + TTEX_1 + ' vs ' + TTEX_1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTEX_1 does not match components (' + TTEX_1 + ' vs ' + TTEX_1_Composite + ').'))
                problem_string = 'TTEX_1 does not match components (' + TTEX_1 + ' vs ' + TTEX_1_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TTEX_1 components not present.', 5)  

        if 'TTEX_2' in Field_Name_List and 'TTEX_2A' in Field_Name_List and 'TTEX_2B' in Field_Name_List and 'TTEX_2C' in Field_Name_List:
            TTEX_2 = stra(row.GetValue('TTEX_2'))
            TTEX_2A = row.GetValue('TTEX_2A')
            TTEX_2B = row.GetValue('TTEX_2B')
            TTEX_2C = row.GetValue('TTEX_2C')
            TTEX_2_Composite = stra(TTEX_2C) + stra(TTEX_2B) + stra(TTEX_2A)
            TTEX_2_Composite = TTEX_2_Composite.replace(' ', '')
            TTEX_2_Composite = TTEX_2_Composite.replace('None', '')
            TTEX_2 = TTEX_2.replace(' ', '')
            TTEX_2 = TTEX_2.replace('None', '')
            #CPMessage('                    - TTEX_2_Composite: ' + TTEX_2_Composite + ' vs ' + stra(TTEX_2), 5)
            if TTEX_2 != TTEX_2_Composite:
                #CPMessage('            - TTEX_2 of OID ' + stra(Current_OID) + ' does not match composite (' + TTEX_2 + ' vs ' + TTEX_2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTEX_2 does not match components (' + TTEX_2 + ' vs ' + TTEX_2_Composite + ').'))
                problem_string = 'TTEX_2 does not match components (' + TTEX_2 + ' vs ' + TTEX_2_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TTEX_2 components not present.', 5)  

        if 'TTEX_3' in Field_Name_List and 'TTEX_3A' in Field_Name_List and 'TTEX_3B' in Field_Name_List and 'TTEX_3C' in Field_Name_List:
            TTEX_3 = stra(row.GetValue('TTEX_3'))
            TTEX_3A = row.GetValue('TTEX_3A')
            TTEX_3B = row.GetValue('TTEX_3B')
            TTEX_3C = row.GetValue('TTEX_3C')
            TTEX_3_Composite = stra(TTEX_3C) + stra(TTEX_3B) + stra(TTEX_3A)
            TTEX_3_Composite = TTEX_3_Composite.replace(' ', '')
            TTEX_3_Composite = TTEX_3_Composite.replace('None', '')
            TTEX_3 = TTEX_3.replace(' ', '')
            TTEX_3 = TTEX_3.replace('None', '')
            #CPMessage('                    - TTEX_3_Composite: ' + TTEX_3_Composite + ' vs ' + stra(TTEX_3), 5)
            if TTEX_3 != TTEX_3_Composite:
                #CPMessage('            - TTEX_3 of OID ' + stra(Current_OID) + ' does not match composite (' + TTEX_3 + ' vs ' + TTEX_3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTEX_3 does not match components (' + TTEX_3 + ' vs ' + TTEX_3_Composite + ').'))
                problem_string = 'TTEX_3 does not match components (' + TTEX_3 + ' vs ' + TTEX_3_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TTEX_3 components not present.', 5)  


        if 'TTTEX_1' in Field_Name_List and 'TTTEX_1A' in Field_Name_List and 'TTTEX_1B' in Field_Name_List and 'TTTEX_1C' in Field_Name_List:
            TTTEX_1 = stra(row.GetValue('TTTEX_1'))
            TTTEX_1A = row.GetValue('TTTEX_1A')
            TTTEX_1B = row.GetValue('TTTEX_1B')
            TTTEX_1C = row.GetValue('TTTEX_1C')
            TTTEX_1_Composite = stra(TTTEX_1C) + stra(TTTEX_1B) + stra(TTTEX_1A)
            TTTEX_1_Composite = TTTEX_1_Composite.replace(' ', '')
            TTTEX_1_Composite = TTTEX_1_Composite.replace('None', '')
            TTTEX_1 = TTTEX_1.replace(' ', '')
            TTTEX_1 = TTTEX_1.replace('None', '')
            #CPMessage('                    - TTTEX_1_Composite: ' + TTTEX_1_Composite + ' vs ' + stra(TTTEX_1), 5)
            if TTTEX_1 != TTTEX_1_Composite:
                #CPMessage('            - TTTEX_1 of OID ' + stra(Current_OID) + ' does not match composite (' + TTTEX_1 + ' vs ' + TTTEX_1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTTEX_1 does not match components (' + TTTEX_1 + ' vs ' + TTTEX_1_Composite + ').'))
                problem_string = 'TTTEX_1 does not match components (' + TTTEX_1 + ' vs ' + TTTEX_1_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TTTEX_1 components not present.', 5)  

        if 'TTTEX_2' in Field_Name_List and 'TTTEX_2A' in Field_Name_List and 'TTTEX_2B' in Field_Name_List and 'TTTEX_2C' in Field_Name_List:
            TTTEX_2 = stra(row.GetValue('TTTEX_2'))
            TTTEX_2A = row.GetValue('TTTEX_2A')
            TTTEX_2B = row.GetValue('TTTEX_2B')
            TTTEX_2C = row.GetValue('TTTEX_2C')
            TTTEX_2_Composite = stra(TTTEX_2C) + stra(TTTEX_2B) + stra(TTTEX_2A)
            TTTEX_2_Composite = TTTEX_2_Composite.replace(' ', '')
            TTTEX_2_Composite = TTTEX_2_Composite.replace('None', '')
            TTTEX_2 = TTTEX_2.replace(' ', '')
            TTTEX_2 = TTTEX_2.replace('None', '')
            #CPMessage('                    - TTTEX_2_Composite: ' + TTTEX_2_Composite + ' vs ' + stra(TTTEX_2), 5)
            if TTTEX_2 != TTTEX_2_Composite:
                #CPMessage('            - TTTEX_2 of OID ' + stra(Current_OID) + ' does not match composite (' + TTTEX_2 + ' vs ' + TTTEX_2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTTEX_2 does not match components (' + TTTEX_2 + ' vs ' + TTTEX_2_Composite + ').'))
                problem_string = 'TTTEX_2 does not match components (' + TTTEX_2 + ' vs ' + TTTEX_2_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TTTEX_2 components not present.', 5)  

        if 'TTTEX_3' in Field_Name_List and 'TTTEX_3A' in Field_Name_List and 'TTTEX_3B' in Field_Name_List and 'TTTEX_3C' in Field_Name_List:
            TTTEX_3 = stra(row.GetValue('TTTEX_3'))
            TTTEX_3A = row.GetValue('TTTEX_3A')
            TTTEX_3B = row.GetValue('TTTEX_3B')
            TTTEX_3C = row.GetValue('TTTEX_3C')
            TTTEX_3_Composite = stra(TTTEX_3C) + stra(TTTEX_3B) + stra(TTTEX_3A)
            TTTEX_3_Composite = TTTEX_3_Composite.replace(' ', '')
            TTTEX_3_Composite = TTTEX_3_Composite.replace('None', '')
            TTTEX_3 = TTTEX_3.replace(' ', '')
            TTTEX_3 = TTTEX_3.replace('None', '')
            #CPMessage('                    - TTTEX_3_Composite: ' + TTTEX_3_Composite + ' vs ' + stra(TTTEX_3), 5)
            if TTTEX_3 != TTTEX_3_Composite:
                #CPMessage('            - TTTEX_3 of OID ' + stra(Current_OID) + ' does not match composite (' + TTTEX_3 + ' vs ' + TTTEX_3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTTEX_3 does not match components (' + TTTEX_3 + ' vs ' + TTTEX_3_Composite + ').'))
                problem_string = 'TTTEX_3 does not match components (' + TTTEX_3 + ' vs ' + TTTEX_3_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TTTEX_3 components not present.', 5)  


        if 'STTEX_1' in Field_Name_List and 'STTEX_1A' in Field_Name_List and 'STTEX_1B' in Field_Name_List and 'STTEX_1C' in Field_Name_List:
            STTEX_1 = stra(row.GetValue('STTEX_1'))
            STTEX_1A = row.GetValue('STTEX_1A')
            STTEX_1B = row.GetValue('STTEX_1B')
            STTEX_1C = row.GetValue('STTEX_1C')
            STTEX_1_Composite = stra(STTEX_1C) + stra(STTEX_1B) + stra(STTEX_1A)
            STTEX_1_Composite = STTEX_1_Composite.replace(' ', '')
            STTEX_1_Composite = STTEX_1_Composite.replace('None', '')
            STTEX_1 = STTEX_1.replace(' ', '')
            STTEX_1 = STTEX_1.replace('None', '')
            #CPMessage('                    - STTEX_1_Composite: ' + STTEX_1_Composite + ' vs ' + stra(STTEX_1), 5)
            if STTEX_1 != STTEX_1_Composite:
                #CPMessage('            - STTEX_1 of OID ' + stra(Current_OID) + ' does not match composite (' + STTEX_1 + ' vs ' + STTEX_1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'STTEX_1 does not match components (' + STTEX_1 + ' vs ' + STTEX_1_Composite + ').'))
                problem_string = 'STTEX_1 does not match components (' + STTEX_1 + ' vs ' + STTEX_1_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - STTEX_1 components not present.', 5)  

        if 'STTEX_2' in Field_Name_List and 'STTEX_2A' in Field_Name_List and 'STTEX_2B' in Field_Name_List and 'STTEX_2C' in Field_Name_List:
            STTEX_2 = stra(row.GetValue('STTEX_2'))
            STTEX_2A = row.GetValue('STTEX_2A')
            STTEX_2B = row.GetValue('STTEX_2B')
            STTEX_2C = row.GetValue('STTEX_2C')
            STTEX_2_Composite = stra(STTEX_2C) + stra(STTEX_2B) + stra(STTEX_2A)
            STTEX_2_Composite = STTEX_2_Composite.replace(' ', '')
            STTEX_2_Composite = STTEX_2_Composite.replace('None', '')
            STTEX_2 = STTEX_2.replace(' ', '')
            STTEX_2 = STTEX_2.replace('None', '')
            #CPMessage('                    - STTEX_2_Composite: ' + STTEX_2_Composite + ' vs ' + stra(STTEX_2), 5)
            if STTEX_2 != STTEX_2_Composite:
                #CPMessage('            - STTEX_2 of OID ' + stra(Current_OID) + ' does not match composite (' + STTEX_2 + ' vs ' + STTEX_2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'STTEX_2 does not match components (' + STTEX_2 + ' vs ' + STTEX_2_Composite + ').'))
                problem_string = 'STTEX_2 does not match components (' + STTEX_2 + ' vs ' + STTEX_2_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - STTEX_2 components not present.', 5)  

        if 'STTEX_3' in Field_Name_List and 'STTEX_3A' in Field_Name_List and 'STTEX_3B' in Field_Name_List and 'STTEX_3C' in Field_Name_List:
            STTEX_3 = stra(row.GetValue('STTEX_3'))
            STTEX_3A = row.GetValue('STTEX_3A')
            STTEX_3B = row.GetValue('STTEX_3B')
            STTEX_3C = row.GetValue('STTEX_3C')
            STTEX_3_Composite = stra(STTEX_3C) + stra(STTEX_3B) + stra(STTEX_3A)
            STTEX_3_Composite = STTEX_3_Composite.replace(' ', '')
            STTEX_3_Composite = STTEX_3_Composite.replace('None', '')
            STTEX_3 = STTEX_3.replace(' ', '')
            STTEX_3 = STTEX_3.replace('None', '')
            #CPMessage('                    - STTEX_3_Composite: ' + STTEX_3_Composite + ' vs ' + stra(STTEX_3), 5)
            if STTEX_3 != STTEX_3_Composite:
                #CPMessage('            - STTEX_3 of OID ' + stra(Current_OID) + ' does not match composite (' + STTEX_3 + ' vs ' + STTEX_3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'STTEX_3 does not match components (' + STTEX_3 + ' vs ' + STTEX_3_Composite + ').'))
                problem_string = 'STTEX_3 does not match components (' + STTEX_3 + ' vs ' + STTEX_3_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - STTEX_3 components not present.', 5)  


        if 'SURF_E1' in Field_Name_List and 'SURF_E1A' in Field_Name_List and 'SURF_E1B' in Field_Name_List and 'SURF_E1C' in Field_Name_List:
            SURF_E1 = stra(row.GetValue('SURF_E1'))
            SURF_E1A = row.GetValue('SURF_E1A')
            SURF_E1B = row.GetValue('SURF_E1B')
            SURF_E1C = row.GetValue('SURF_E1C')
            SURF_E1_Composite = stra(SURF_E1A) + stra(SURF_E1B) + stra(SURF_E1C)
            SURF_E1_Composite = SURF_E1_Composite.replace(' ', '')
            SURF_E1_Composite = SURF_E1_Composite.replace('None', '')
            SURF_E1 = SURF_E1.replace(' ', '')
            SURF_E1 = SURF_E1.replace('None', '')
            #CPMessage('                    - SURF_E1_Composite: ' + SURF_E1_Composite + ' vs ' + stra(SURF_E1), 5)
            if SURF_E1 != SURF_E1_Composite:
                #CPMessage('            - SURF_E1 of OID ' + stra(Current_OID) + ' does not match composite (' + SURF_E1 + ' vs ' + SURF_E1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SURF_E1 does not match components (' + SURF_E1 + ' vs ' + SURF_E1_Composite + ').'))
                problem_string = 'SURF_E1 does not match components (' + SURF_E1 + ' vs ' + SURF_E1_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SURF_E1 components not present.', 5)  

        if 'SURF_E2' in Field_Name_List and 'SURF_E2A' in Field_Name_List and 'SURF_E2B' in Field_Name_List and 'SURF_E2C' in Field_Name_List:
            SURF_E2 = stra(row.GetValue('SURF_E2'))
            SURF_E2A = row.GetValue('SURF_E2A')
            SURF_E2B = row.GetValue('SURF_E2B')
            SURF_E2C = row.GetValue('SURF_E2C')
            SURF_E2_Composite = stra(SURF_E2A) + stra(SURF_E2B) + stra(SURF_E2C)
            SURF_E2_Composite = SURF_E2_Composite.replace(' ', '')
            SURF_E2_Composite = SURF_E2_Composite.replace('None', '')
            SURF_E2 = SURF_E2.replace(' ', '')
            SURF_E2 = SURF_E2.replace('None', '')
            #CPMessage('                    - SURF_E2_Composite: ' + SURF_E2_Composite + ' vs ' + stra(SURF_E2), 5)
            if SURF_E2 != SURF_E2_Composite:
                #CPMessage('            - SURF_E2 of OID ' + stra(Current_OID) + ' does not match composite (' + SURF_E2 + ' vs ' + SURF_E2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SURF_E2 does not match components (' + SURF_E2 + ' vs ' + SURF_E2_Composite + ').'))
                problem_string = 'SURF_E2 does not match components (' + SURF_E2 + ' vs ' + SURF_E2_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SURF_E2 components not present.', 5)  

        if 'SURF_E3' in Field_Name_List and 'SURF_E3A' in Field_Name_List and 'SURF_E3B' in Field_Name_List and 'SURF_E3C' in Field_Name_List:
            SURF_E3 = stra(row.GetValue('SURF_E3'))
            SURF_E3A = row.GetValue('SURF_E3A')
            SURF_E3B = row.GetValue('SURF_E3B')
            SURF_E3C = row.GetValue('SURF_E3C')
            SURF_E3_Composite = stra(SURF_E3A) + stra(SURF_E3B) + stra(SURF_E3C)
            SURF_E3_Composite = SURF_E3_Composite.replace(' ', '')
            SURF_E3_Composite = SURF_E3_Composite.replace('None', '')
            SURF_E3 = SURF_E3.replace(' ', '')
            SURF_E3 = SURF_E3.replace('None', '')
            #CPMessage('                    - SURF_E3_Composite: ' + SURF_E3_Composite + ' vs ' + stra(SURF_E3), 5)
            if SURF_E3 != SURF_E3_Composite:
                #CPMessage('            - SURF_E3 of OID ' + stra(Current_OID) + ' does not match composite (' + SURF_E3 + ' vs ' + SURF_E3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SURF_E3 does not match components (' + SURF_E3 + ' vs ' + SURF_E3_Composite + ').'))
                problem_string = 'SURF_E3 does not match components (' + SURF_E3 + ' vs ' + SURF_E3_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SURF_E3 components not present.', 5)  



        if 'TSURF_E1' in Field_Name_List and 'TSURF_E1A' in Field_Name_List and 'TSURF_E1B' in Field_Name_List and 'TSURF_E1C' in Field_Name_List:
            TSURF_E1 = stra(row.GetValue('TSURF_E1'))
            TSURF_E1A = row.GetValue('TSURF_E1A')
            TSURF_E1B = row.GetValue('TSURF_E1B')
            TSURF_E1C = row.GetValue('TSURF_E1C')
            TSURF_E1_Composite = stra(TSURF_E1A) + stra(TSURF_E1B) + stra(TSURF_E1C)
            TSURF_E1_Composite = TSURF_E1_Composite.replace(' ', '')
            TSURF_E1_Composite = TSURF_E1_Composite.replace('None', '')
            TSURF_E1 = TSURF_E1.replace(' ', '')
            TSURF_E1 = TSURF_E1.replace('None', '')
            #CPMessage('                    - TSURF_E1_Composite: ' + TSURF_E1_Composite + ' vs ' + stra(TSURF_E1), 5)
            if TSURF_E1 != TSURF_E1_Composite:
                #CPMessage('            - TSURF_E1 of OID ' + stra(Current_OID) + ' does not match composite (' + TSURF_E1 + ' vs ' + TSURF_E1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TSURF_E1 does not match components (' + TSURF_E1 + ' vs ' + TSURF_E1_Composite + ').'))
                problem_string = 'TSURF_E1 does not match components (' + TSURF_E1 + ' vs ' + TSURF_E1_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TSURF_E1 components not present.', 5)  

        if 'TSURF_E2' in Field_Name_List and 'TSURF_E2A' in Field_Name_List and 'TSURF_E2B' in Field_Name_List and 'TSURF_E2C' in Field_Name_List:
            TSURF_E2 = stra(row.GetValue('TSURF_E2'))
            TSURF_E2A = row.GetValue('TSURF_E2A')
            TSURF_E2B = row.GetValue('TSURF_E2B')
            TSURF_E2C = row.GetValue('TSURF_E2C')
            TSURF_E2_Composite = stra(TSURF_E2A) + stra(TSURF_E2B) + stra(TSURF_E2C)
            TSURF_E2_Composite = TSURF_E2_Composite.replace(' ', '')
            TSURF_E2_Composite = TSURF_E2_Composite.replace('None', '')
            TSURF_E2 = TSURF_E2.replace(' ', '')
            TSURF_E2 = TSURF_E2.replace('None', '')
            #CPMessage('                    - TSURF_E2_Composite: ' + TSURF_E2_Composite + ' vs ' + stra(TSURF_E2), 5)
            if TSURF_E2 != TSURF_E2_Composite:
                #CPMessage('            - TSURF_E2 of OID ' + stra(Current_OID) + ' does not match composite (' + TSURF_E2 + ' vs ' + TSURF_E2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TSURF_E2 does not match components (' + TSURF_E2 + ' vs ' + TSURF_E2_Composite + ').'))
                problem_string = 'TSURF_E2 does not match components (' + TSURF_E2 + ' vs ' + TSURF_E2_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TSURF_E2 components not present.', 5)  

        if 'TSURF_E3' in Field_Name_List and 'TSURF_E3A' in Field_Name_List and 'TSURF_E3B' in Field_Name_List and 'TSURF_E3C' in Field_Name_List:
            TSURF_E3 = stra(row.GetValue('TSURF_E3'))
            TSURF_E3A = row.GetValue('TSURF_E3A')
            TSURF_E3B = row.GetValue('TSURF_E3B')
            TSURF_E3C = row.GetValue('TSURF_E3C')
            TSURF_E3_Composite = stra(TSURF_E3A) + stra(TSURF_E3B) + stra(TSURF_E3C)
            TSURF_E3_Composite = TSURF_E3_Composite.replace(' ', '')
            TSURF_E3_Composite = TSURF_E3_Composite.replace('None', '')
            TSURF_E3 = TSURF_E3.replace(' ', '')
            TSURF_E3 = TSURF_E3.replace('None', '')
            #CPMessage('                    - TSURF_E3_Composite: ' + TSURF_E3_Composite + ' vs ' + stra(TSURF_E3), 5)
            if TSURF_E3 != TSURF_E3_Composite:
                #CPMessage('            - TSURF_E3 of OID ' + stra(Current_OID) + ' does not match composite (' + TSURF_E3 + ' vs ' + TSURF_E3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TSURF_E3 does not match components (' + TSURF_E3 + ' vs ' + TSURF_E3_Composite + ').'))
                problem_string = 'TSURF_E3 does not match components (' + TSURF_E3 + ' vs ' + TSURF_E3_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - TSURF_E3 components not present.', 5)  


        if 'SSURF_E1' in Field_Name_List and 'SSURF_E1A' in Field_Name_List and 'SSURF_E1B' in Field_Name_List and 'SSURF_E1C' in Field_Name_List:
            SSURF_E1 = stra(row.GetValue('SSURF_E1'))
            SSURF_E1A = row.GetValue('SSURF_E1A')
            SSURF_E1B = row.GetValue('SSURF_E1B')
            SSURF_E1C = row.GetValue('SSURF_E1C')
            SSURF_E1_Composite = stra(SSURF_E1A) + stra(SSURF_E1B) + stra(SSURF_E1C)
            SSURF_E1_Composite = SSURF_E1_Composite.replace(' ', '')
            SSURF_E1_Composite = SSURF_E1_Composite.replace('None', '')
            SSURF_E1 = SSURF_E1.replace(' ', '')
            SSURF_E1 = SSURF_E1.replace('None', '')
            #CPMessage('                    - SSURF_E1_Composite: ' + SSURF_E1_Composite + ' vs ' + stra(SSURF_E1), 5)
            if SSURF_E1 != SSURF_E1_Composite:
                #CPMessage('            - SSURF_E1 of OID ' + stra(Current_OID) + ' does not match composite (' + SSURF_E1 + ' vs ' + SSURF_E1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SSURF_E1 does not match components (' + SSURF_E1 + ' vs ' + SSURF_E1_Composite + ').'))
                problem_string = 'SSURF_E1 does not match components (' + SSURF_E1 + ' vs ' + SSURF_E1_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SSURF_E1 components not present.', 5)  

        if 'SSURF_E2' in Field_Name_List and 'SSURF_E2A' in Field_Name_List and 'SSURF_E2B' in Field_Name_List and 'SSURF_E2C' in Field_Name_List:
            SSURF_E2 = stra(row.GetValue('SSURF_E2'))
            SSURF_E2A = row.GetValue('SSURF_E2A')
            SSURF_E2B = row.GetValue('SSURF_E2B')
            SSURF_E2C = row.GetValue('SSURF_E2C')
            SSURF_E2_Composite = stra(SSURF_E2A) + stra(SSURF_E2B) + stra(SSURF_E2C)
            SSURF_E2_Composite = SSURF_E2_Composite.replace(' ', '')
            SSURF_E2_Composite = SSURF_E2_Composite.replace('None', '')
            SSURF_E2 = SSURF_E2.replace(' ', '')
            SSURF_E2 = SSURF_E2.replace('None', '')
            #CPMessage('                    - SSURF_E2_Composite: ' + SSURF_E2_Composite + ' vs ' + stra(SSURF_E2), 5)
            if SSURF_E2 != SSURF_E2_Composite:
                #CPMessage('            - SSURF_E2 of OID ' + stra(Current_OID) + ' does not match composite (' + SSURF_E2 + ' vs ' + SSURF_E2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SSURF_E2 does not match components (' + SSURF_E2 + ' vs ' + SSURF_E2_Composite + ').'))
                problem_string = 'SSURF_E2 does not match components (' + SSURF_E2 + ' vs ' + SSURF_E2_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SSURF_E2 components not present.', 5)  

        if 'SSURF_E3' in Field_Name_List and 'SSURF_E3A' in Field_Name_List and 'SSURF_E3B' in Field_Name_List and 'SSURF_E3C' in Field_Name_List:
            SSURF_E3 = stra(row.GetValue('SSURF_E3'))
            SSURF_E3A = row.GetValue('SSURF_E3A')
            SSURF_E3B = row.GetValue('SSURF_E3B')
            SSURF_E3C = row.GetValue('SSURF_E3C')
            SSURF_E3_Composite = stra(SSURF_E3A) + stra(SSURF_E3B) + stra(SSURF_E3C)
            SSURF_E3_Composite = SSURF_E3_Composite.replace(' ', '')
            SSURF_E3_Composite = SSURF_E3_Composite.replace('None', '')
            SSURF_E3 = SSURF_E3.replace(' ', '')
            SSURF_E3 = SSURF_E3.replace('None', '')
            #CPMessage('                    - SSURF_E3_Composite: ' + SSURF_E3_Composite + ' vs ' + stra(SSURF_E3), 5)
            if SSURF_E3 != SSURF_E3_Composite:
                #CPMessage('            - SSURF_E3 of OID ' + stra(Current_OID) + ' does not match composite (' + SSURF_E3 + ' vs ' + SSURF_E3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SSURF_E3 does not match components (' + SSURF_E3 + ' vs ' + SSURF_E3_Composite + ').'))
                problem_string = 'SSURF_E3 does not match components (' + SSURF_E3 + ' vs ' + SSURF_E3_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SSURF_E3 components not present.', 5)  

#        if 'GEOP_SCM1' in Field_Name_List and 'GEOP_SCM1A' in Field_Name_List and 'GEOP_SCM1B' in Field_Name_List and 'GEOP_SCM1C' in Field_Name_List:
        if 'GEOP_SCM1' in Field_Name_List and 'GEOP_SCM1A' in Field_Name_List and 'GEOP_SCM1B' in Field_Name_List and 'GEOP_SCM1C' in Field_Name_List and 'GEOP_INZ1' in Field_Name_List:
            GEOP_SCM1 = stra(row.GetValue('GEOP_SCM1'))
            GEOP_INZ1 = row.GetValue('GEOP_INZ1')
            GEOP_SCM1A = row.GetValue('GEOP_SCM1A')
            GEOP_SCM1B = row.GetValue('GEOP_SCM1B')
            GEOP_SCM1C = row.GetValue('GEOP_SCM1C')
            if stra(GEOP_INZ1) not in ['',' ','None']:
                GEOP_SCM1_Composite = stra(GEOP_INZ1) + stra(GEOP_SCM1A) + stra(GEOP_SCM1B)
            else:
                GEOP_SCM1_Composite = stra(GEOP_SCM1A) + stra(GEOP_SCM1B) + stra(GEOP_SCM1C)
            GEOP_SCM1_Composite = GEOP_SCM1_Composite.replace(' ', '')
            GEOP_SCM1_Composite = GEOP_SCM1_Composite.replace('None', '')
            GEOP_SCM1 = GEOP_SCM1.replace(' ', '')
            GEOP_SCM1 = GEOP_SCM1.replace('None', '')
            #CPMessage('                    - GEOP_SCM1_Composite: ' + GEOP_SCM1_Composite + ' vs ' + stra(GEOP_SCM1), 5)
            if GEOP_SCM1 != GEOP_SCM1_Composite:
                #CPMessage('            - GEOP_SCM1 of OID ' + stra(Current_OID) + ' does not match composite (' + GEOP_SCM1 + ' vs ' + GEOP_SCM1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'GEOP_SCM1 does not match components (' + GEOP_SCM1 + ' vs ' + GEOP_SCM1_Composite + ').'))
                problem_string = 'GEOP_SCM1 does not match components (' + GEOP_SCM1 + ' vs ' + GEOP_SCM1_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - GEOP_SCM1 components not present.', 5)  

#        if 'GEOP_SCM2' in Field_Name_List and 'GEOP_SCM2A' in Field_Name_List and 'GEOP_SCM2B' in Field_Name_List and 'GEOP_SCM2C' in Field_Name_List:
        if 'GEOP_SCM2' in Field_Name_List and 'GEOP_SCM2A' in Field_Name_List and 'GEOP_SCM2B' in Field_Name_List and 'GEOP_SCM2C' in Field_Name_List and 'GEOP_INZ2' in Field_Name_List:
            GEOP_SCM2 = stra(row.GetValue('GEOP_SCM2'))
            GEOP_INZ2 = row.GetValue('GEOP_INZ2')
            GEOP_SCM2A = row.GetValue('GEOP_SCM2A')
            GEOP_SCM2B = row.GetValue('GEOP_SCM2B')
            GEOP_SCM2C = row.GetValue('GEOP_SCM2C')
            if stra(GEOP_INZ2) not in ['',' ','None']:
                GEOP_SCM2_Composite = stra(GEOP_INZ2) + stra(GEOP_SCM2A) + stra(GEOP_SCM2B)
            else:
                GEOP_SCM2_Composite = stra(GEOP_SCM2A) + stra(GEOP_SCM2B) + stra(GEOP_SCM2C)
            GEOP_SCM2_Composite = GEOP_SCM2_Composite.replace(' ', '')
            GEOP_SCM2_Composite = GEOP_SCM2_Composite.replace('None', '')
            GEOP_SCM2 = GEOP_SCM2.replace(' ', '')
            GEOP_SCM2 = GEOP_SCM2.replace('None', '')
            #CPMessage('                    - GEOP_SCM2_Composite: ' + GEOP_SCM2_Composite + ' vs ' + stra(GEOP_SCM2), 5)
            if GEOP_SCM2 != GEOP_SCM2_Composite:
                #CPMessage('            - GEOP_SCM2 of OID ' + stra(Current_OID) + ' does not match composite (' + GEOP_SCM2 + ' vs ' + GEOP_SCM2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'GEOP_SCM2 does not match components (' + GEOP_SCM2 + ' vs ' + GEOP_SCM2_Composite + ').'))
                problem_string = 'GEOP_SCM2 does not match components (' + GEOP_SCM2 + ' vs ' + GEOP_SCM2_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - GEOP_SCM2 components not present.', 5)  

#        if 'GEOP_SCM3' in Field_Name_List and 'GEOP_SCM3A' in Field_Name_List and 'GEOP_SCM3B' in Field_Name_List and 'GEOP_SCM3C' in Field_Name_List:
        if 'GEOP_SCM3' in Field_Name_List and 'GEOP_SCM3A' in Field_Name_List and 'GEOP_SCM3B' in Field_Name_List and 'GEOP_SCM3C' in Field_Name_List and 'GEOP_INZ3' in Field_Name_List:
            GEOP_SCM3 = stra(row.GetValue('GEOP_SCM3'))
            GEOP_INZ3 = row.GetValue('GEOP_INZ3')
            GEOP_SCM3A = row.GetValue('GEOP_SCM3A')
            GEOP_SCM3B = row.GetValue('GEOP_SCM3B')
            GEOP_SCM3C = row.GetValue('GEOP_SCM3C')
            if stra(GEOP_INZ3) not in ['',' ','None']:
                GEOP_SCM3_Composite = stra(GEOP_INZ3) + stra(GEOP_SCM3A) + stra(GEOP_SCM3B)
            else:
                GEOP_SCM3_Composite = stra(GEOP_SCM3A) + stra(GEOP_SCM3B) + stra(GEOP_SCM3C)
            GEOP_SCM3_Composite = GEOP_SCM3_Composite.replace(' ', '')
            GEOP_SCM3_Composite = GEOP_SCM3_Composite.replace('None', '')
            GEOP_SCM3 = GEOP_SCM3.replace(' ', '')
            GEOP_SCM3 = GEOP_SCM3.replace('None', '')
            #CPMessage('                    - GEOP_SCM3_Composite: ' + GEOP_SCM3_Composite + ' vs ' + stra(GEOP_SCM3), 5)
            if GEOP_SCM3 != GEOP_SCM3_Composite:
                #CPMessage('            - GEOP_SCM3 of OID ' + stra(Current_OID) + ' does not match composite (' + GEOP_SCM3 + ' vs ' + GEOP_SCM3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'GEOP_SCM3 does not match components (' + GEOP_SCM3 + ' vs ' + GEOP_SCM3_Composite + ').'))
                problem_string = 'GEOP_SCM3 does not match components (' + GEOP_SCM3 + ' vs ' + GEOP_SCM3_Composite + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - GEOP_SCM3 components not present.', 5)  


        if 'SLPLL_1' in Field_Name_List and 'SLPUL_1' in Field_Name_List:
            SLPLL_1 = row.GetValue('SLPLL_1')
            if stra(SLPLL_1) == 'None':
                SLPLL_1 = 0
            SLPUL_1 = row.GetValue('SLPUL_1')
            if stra(SLPUL_1) == '0':
                #CPMessage('            - SLPUL_1 of OID ' + stra(Current_OID) + ' is zero.', 3)
                pass
            if stra(SLPUL_1) == 'None':
                SLPUL_1 = 0            
            if SLPLL_1 >= SLPUL_1 and SLPUL_1 != 0:
                #CPMessage('            - SLPLL_1 of OID ' + stra(Current_OID) + ' is higher than or equal to SLPUL_1 (' + stra(SLPLL_1) + ' vs ' + stra(SLPUL_1) + ').', 3)
                #Error_List.append((Current_OID, r'SLPLL_1 is higher than or equal to SLPUL_1 (' + stra(SLPLL_1) + ' vs ' + stra(SLPUL_1) + ').'))
                problem_string = 'SLPLL_1 is higher than or equal to SLPUL_1 (' + stra(SLPLL_1) + ' vs ' + stra(SLPUL_1) + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SLPLL_1, SLPUL_1 not present.', 5)  


        if 'SLPLL_2' in Field_Name_List and 'SLPUL_2' in Field_Name_List:
            SLPLL_2 = row.GetValue('SLPLL_2')
            if stra(SLPLL_2) == 'None':
                SLPLL_2 = 0
            SLPUL_2 = row.GetValue('SLPUL_2')
            if stra(SLPUL_2) == '0':
                #CPMessage('            - SLPUL_2 of OID ' + stra(Current_OID) + ' is zero.', 3)            
                pass
            if stra(SLPUL_2) == 'None':
                SLPUL_2 = 0            
            if SLPLL_2 >= SLPUL_2 and SLPUL_2 != 0:
                #CPMessage('            - SLPLL_2 of OID ' + stra(Current_OID) + ' is higher than or equal to SLPUL_2 (' + stra(SLPLL_2) + ' vs ' + stra(SLPUL_2) + ').', 3)
                #Error_List.append((Current_OID, r'SLPLL_2 is higher than or equal to SLPUL_2 (' + stra(SLPLL_2) + ' vs ' + stra(SLPUL_2) + ').'))
                problem_string = 'SLPLL_2 is higher than or equal to SLPUL_2 (' + stra(SLPLL_2) + ' vs ' + stra(SLPUL_2) + ').'
                rowet_row = rowet_rows.NewRow()
                rowet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                rowet_row.SetValue('FEATURE_ID', Current_OID)
                rowet_row.SetValue('PROBLEM', problem_string)
                rowet_rows.InsertRow(rowet_row)
                summary_error_list.append(problem_string)
        else:
            CPMessage('                    - SLPLL_2, SLPUL_2 not present.', 5)  


        #if 'SITE_S1' in Field_Name_List and 'SDEC_1' in Field_Name_List:
        #    SITE_S1 = row.GetValue('SITE_S1')
        #    SDEC_1 = row.GetValue('SDEC_1')
        #    if stra(SDEC_1) != 'None' and stra(SITE_S1) in ['None','']:
        #        CPMessage('            - SITE_S1 of OID ' + stra(Current_OID) + ' is unexpectedly blank.', 3)

        #if 'SITE_S2' in Field_Name_List and 'SDEC_2' in Field_Name_List:
        #    SITE_S2 = row.GetValue('SITE_S2')
        #    SDEC_2 = row.GetValue('SDEC_2')
        #    if stra(SDEC_2) != 'None' and stra(SITE_S2) in ['None','']:
        #        CPMessage('            - SITE_S2 of OID ' + stra(Current_OID) + ' is unexpectedly blank.', 3)

        #if 'SITE_S3' in Field_Name_List and 'SDEC_3' in Field_Name_List:
        #    SITE_S3 = row.GetValue('SITE_S3')
        #    SDEC_3 = row.GetValue('SDEC_3')
        #    if stra(SDEC_3) != 'None' and stra(SITE_S3) in ['None','']:
        #        CPMessage('            - SITE_S3 of OID ' + stra(Current_OID) + ' is unexpectedly blank.', 3)

        ##################################
        ##### Check Mandatory Fields #####
        ##################################

        if CHECK_MANDATORY_FIELDS:
            for Mandatory_Field_Name in Mandatory_Check_Fields:
                Test_Value = stra(row.GetValue(Mandatory_Field_Name))
                CPMessage('            - Testing: ' + Mandatory_Field_Name + ' for mandatory value', 5)
                if stra(Test_Value) not in ['None','',' ']:
                    #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                    pass
                else:
                    Current_OID = row.GetValue(Object_ID_Field)
                    #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, 'Mandatory value missing in ' + Mandatory_Field_Name + '.'))
                    problem_string = 'Mandatory value missing in ' + Mandatory_Field_Name + '.'
                    mfet_row = mfet_rows.NewRow()
                    mfet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                    mfet_row.SetValue('FEATURE_ID', Current_OID)
                    mfet_row.SetValue('PROBLEM', problem_string)
                    mfet_rows.InsertRow(mfet_row)
                    summary_error_list.append(problem_string)

            Current_Project_Type = stra(row.GetValue('PROJ_TYPE'))

            if 'T' in Current_Project_Type:
                for Mandatory_Field_Name_T in Mandatory_Check_Fields_T:
                    Test_Value = stra(row.GetValue(Mandatory_Field_Name_T))
                    CPMessage('            - Testing: ' + Mandatory_Field_Name_T + ' for mandatory value', 5)
                    if stra(Test_Value) not in ['None','',' ']:
                        #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                        pass
                    else:
                        Current_OID = row.GetValue(Object_ID_Field)
                        #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name_T + ' at OID ' + stra(Current_OID) + '.', 4)
                        #Error_List.append((Current_OID, 'Mandatory value (for %T% project types) missing in ' + Mandatory_Field_Name_T + '.'))
                        problem_string = 'Mandatory value (for %T% project types) missing in ' + Mandatory_Field_Name_T + '.'
                        mfet_row = mfet_rows.NewRow()
                        mfet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                        mfet_row.SetValue('FEATURE_ID', Current_OID)
                        mfet_row.SetValue('PROBLEM', problem_string)
                        mfet_rows.InsertRow(mfet_row)
                        summary_error_list.append(problem_string)

            if 'EM' in Current_Project_Type:
                for Mandatory_Field_Name_EM in Mandatory_Check_Fields_EM:
                    Test_Value = stra(row.GetValue(Mandatory_Field_Name_EM))
                    CPMessage('            - Testing: ' + Mandatory_Field_Name_EM + ' for mandatory value', 5)
                    if stra(Test_Value) not in ['None','',' ']:
                        #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                        pass
                    else:
                        Current_OID = row.GetValue(Object_ID_Field)
                        #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name_EM + ' at OID ' + stra(Current_OID) + '.', 4)
                        #Error_List.append((Current_OID, 'Mandatory value (for %EM% project types) missing in ' + Mandatory_Field_Name_EM + '.'))
                        problem_string = 'Mandatory value (for %EM% project types) missing in ' + Mandatory_Field_Name_EM + '.'
                        mfet_row = mfet_rows.NewRow()
                        mfet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                        mfet_row.SetValue('FEATURE_ID', Current_OID)
                        mfet_row.SetValue('PROBLEM', problem_string)
                        mfet_rows.InsertRow(mfet_row)
                        summary_error_list.append(problem_string)

            if 'SEI' in Current_Project_Type:
                for Mandatory_Field_Name_SEI in Mandatory_Check_Fields_SEI:
                    Test_Value = stra(row.GetValue(Mandatory_Field_Name_SEI))
                    CPMessage('            - Testing: ' + Mandatory_Field_Name_SEI + ' for mandatory value', 5)
                    if stra(Test_Value) not in ['None','',' ']:
                        #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                        pass
                    else:
                        Current_OID = row.GetValue(Object_ID_Field)
                        #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name_SEI + ' at OID ' + stra(Current_OID) + '.', 4)
                        #Error_List.append((Current_OID, 'Mandatory value (for %SEI% project types) missing in ' + Mandatory_Field_Name_SEI + '.'))
                        problem_string = 'Mandatory value (for %SEI% project types) missing in ' + Mandatory_Field_Name_SEI + '.'
                        mfet_row = mfet_rows.NewRow()
                        mfet_row.SetValue('CLASS', INPUT_FOR_TESTING)
                        mfet_row.SetValue('FEATURE_ID', Current_OID)
                        mfet_row.SetValue('PROBLEM', problem_string)
                        mfet_rows.InsertRow(mfet_row)
                        summary_error_list.append(problem_string)

        ###############################
        ##### Write Error Summary #####
        ###############################

        if 0 < len(summary_error_list) < 51:
            if len(summary_error_list) == 1:
                problem_string = '1 Error: '
            else:
                problem_string = stra(len(summary_error_list)) + ' Errors: '
            for error in summary_error_list:
                problem_string += error + ' '
            problem_string = problem_string[:-1]
            problem_string = problem_string[:499]
            #CPMessage('            - Error summary for OID ' + stra(Current_OID) + ': ' + problem_string, 4)
            set_row = set_rows.NewRow()
            set_row.SetValue('FEATURE_ID', Current_OID)
            set_row.SetValue('PROBLEMS', problem_string)
            set_rows.InsertRow(set_row)
            error_count += 1
        elif len(summary_error_list) >= 51:
            problem_string = '** MANY ERRORS (' + stra(len(summary_error_list)) + ')'
            set_row = set_rows.NewRow()
            set_row.SetValue('FEATURE_ID', Current_OID)
            set_row.SetValue('PROBLEMS', problem_string)
            set_rows.InsertRow(set_row)
            error_count += 1

        Number_Processed = Number_Processed + 1
        if (0 < Number_Processed <= 10000 and Number_Processed % 1000 == 0) or (10000 < Number_Processed and Number_Processed % 10000 == 0):
            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
            CPMessage('            - Processed ' + stra(Number_Processed) + ' of ' + stra(Number_of_Features) + ' (' + stra(error_count) + ' records found with errors) (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)
        
        row = rows.Next()
    CPMessage('            - Processed all ' + stra(Number_Processed), 4)
    del row
    del rows

    # Can't delete row object if a new row was never created (ie. if there were no errors of that type to be written)
    try:
        del det_row
    except:
        pass
        
    try:
        del ret_row
    except:
        pass

    try:
        del cret_row
    except:
        pass

    try:
        del rowet_row
    except:
        pass

    try:
        del mfet_row
    except:
        pass

    try:
        del set_row
    except:
        pass

    try:
        del get_row
    except:
        pass

    try:
        del get_rows
    except:
        pass

    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# CheckEverything101()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def CheckEverything101():
    if CHECK_MANDATORY_FIELDS:
        CPMessage('    - Checking Domains, Ranges, Critical Ranges, Rows, Mandatory Fields', 2)
    else:
        CPMessage('    - Checking Domains, Ranges, Critical Ranges, Rows', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "CheckEverything101"

    # Get the Field List
    Field_List = arcpy.ListFields(INPUT_FOR_TESTING)
    Field_Name_List = []
    for Field in Field_List:
        Field_Name_List.append(Field.name)
    
    Cursor_Field_List = []
    
    Domain_Check_Fields = []
    
    CPMessage('        - FieldsXDomains: ' + stra(Processing_Variables['FieldsXDomains']), 5)
    
    for Current_Field in Field_List:

        if Current_Field.name in Processing_Variables['FieldsXDomains']:
            Domain_Check_Fields.append(Current_Field.name)
            if Current_Field.name not in Cursor_Field_List:
                Cursor_Field_List.append(Current_Field.name)
            CPMessage('            - Checking Field: ' + stra(Current_Field.name), 5)
        else:
            CPMessage('            - Not Checking Field: ' + stra(Current_Field.name), 5)

    CPMessage('        - Checking Fields for Domains: ' + stra(Domain_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    Range_Check_Fields = []
    
    CPMessage('        - Ranges: ' + stra(Processing_Variables['Ranges'].keys()), 5)
    
    for Current_Field in Field_List:
        if Current_Field.name.upper() in Processing_Variables['Ranges']:
            Range_Check_Fields.append(Current_Field.name)
            if Current_Field.name not in Cursor_Field_List:
                Cursor_Field_List.append(Current_Field.name)
            CPMessage('            - Checking Field: ' + stra(Current_Field.name), 5)
        else:
            CPMessage('            - Not Checking Field: ' + stra(Current_Field.name), 5)

    CPMessage('        - Checking Fields for Ranges: ' + stra(Range_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    Critical_Range_Check_Fields = []
    
    CPMessage('        - Critical Ranges: ' + stra(Processing_Variables['CriticalRanges'].keys()), 5)
    
    for Current_Field in Field_List:
        if Current_Field.name.upper() in Processing_Variables['CriticalRanges']:
            Critical_Range_Check_Fields.append(Current_Field.name)
            if Current_Field.name not in Cursor_Field_List:
                Cursor_Field_List.append(Current_Field.name)
            CPMessage('            - Checking Field: ' + stra(Current_Field.name), 5)
        else:
            CPMessage('            - Not Checking Field: ' + stra(Current_Field.name), 5)

    CPMessage('        - Checking Fields for Critical Ranges: ' + stra(Critical_Range_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    for Current_Field_Name in Processing_Variables['Rows']:
        if Current_Field_Name in Field_Name_List and Current_Field_Name not in Cursor_Field_List:
            Cursor_Field_List.append(Current_Field_Name)
    
    if CHECK_MANDATORY_FIELDS:
        Mandatory_Check_Fields = []
        Mandatory_Check_Fields_T = []
        Mandatory_Check_Fields_EM = []
        Mandatory_Check_Fields_SEI = []

        for Current_Field in Field_List:

            if Current_Field.name in Processing_Variables['Mandatory_Fields']:
                Mandatory_Check_Fields.append(Current_Field.name)
                if Current_Field.name not in Cursor_Field_List:
                    Cursor_Field_List.append(Current_Field.name)

            if Current_Field.name in Processing_Variables['Mandatory_Fields_T']:
                Mandatory_Check_Fields_T.append(Current_Field.name)
                if Current_Field.name not in Cursor_Field_List:
                    Cursor_Field_List.append(Current_Field.name)

            if Current_Field.name in Processing_Variables['Mandatory_Fields_EM']:
                Mandatory_Check_Fields_EM.append(Current_Field.name)
                if Current_Field.name not in Cursor_Field_List:
                    Cursor_Field_List.append(Current_Field.name)

            if Current_Field.name in Processing_Variables['Mandatory_Fields_SEI']:
                Mandatory_Check_Fields_SEI.append(Current_Field.name)
                if Current_Field.name not in Cursor_Field_List:
                    Cursor_Field_List.append(Current_Field.name)

        CPMessage('        - Checking Mandatory Fields: ' + stra(Mandatory_Check_Fields).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)
        CPMessage('        - Checking Mandatory Fields for %T% Project Types: ' + stra(Mandatory_Check_Fields_T).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)
        CPMessage('        - Checking Mandatory Fields for %EM% Project Types: ' + stra(Mandatory_Check_Fields_EM).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)
        CPMessage('        - Checking Mandatory Fields for %SEI% Project Types: ' + stra(Mandatory_Check_Fields_SEI).replace('[','').replace(']','').replace("', u'",', ').replace("u'","'").replace("'",""), 4)

    #  Key on TEIS_ID, or, for tables (like Project_Details) without TEIS_ID, key on the OID field
    Object_ID_Field = arcpy.ListFields(INPUT_FOR_TESTING, "*", "OID")[0].name
    if Object_ID_Field not in Cursor_Field_List:
        Cursor_Field_List.append(Object_ID_Field)

    if "TEIS_ID" in Field_Name_List:
        Object_ID_Field = "TEIS_ID"
        if "TEIS_ID" not in Cursor_Field_List:
            Cursor_Field_List.append("TEIS_ID")

    if "PROJ_TYPE" in Field_Name_List and "PROJ_TYPE" not in Cursor_Field_List:
        Cursor_Field_List.append("PROJ_TYPE")

    if "SHAPE@XY" not in Cursor_Field_List:
        Cursor_Field_List.append("SHAPE@XY")

    CPMessage('        - Keying on the input table\'s ' + Object_ID_Field + ' field.', 4)
    
    Number_of_Features = int(arcpy.GetCount_management(INPUT_FOR_TESTING).getOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Checking individual values for ' + stra(Number_of_Features) + ' features.', 4)

    Error_List = []

    FieldName_Dict = {}
    x = 0
    for CurrentFieldName in Cursor_Field_List:
        FieldName_Dict[CurrentFieldName] = x
        x += 1

    # start an edit session in the FGDB of the input dataset
    edit = arcpy.da.Editor(os.path.split(INPUT_FOR_TESTING)[0])
    edit.startEditing(False,False)
    
    det_cursor = arcpy.da.InsertCursor(Processing_Variables['DomainErrorTable'],['CLASS','FEATURE_ID','PROBLEM','BAPID'])
    ret_cursor = arcpy.da.InsertCursor(Processing_Variables['RangeErrorTable'],['CLASS','FEATURE_ID','PROBLEM','BAPID'])
    cret_cursor = arcpy.da.InsertCursor(Processing_Variables['CriticalRangeErrorTable'],['CLASS','FEATURE_ID','PROBLEM','BAPID'])
    rowet_cursor = arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM','BAPID'])
    if CHECK_MANDATORY_FIELDS:
        mfet_cursor = arcpy.da.InsertCursor(Processing_Variables['MandatoryFieldErrorTable'],['CLASS','FEATURE_ID','PROBLEM','BAPID'])
    set_cursor = arcpy.da.InsertCursor(Processing_Variables['SummaryErrorTable'],['FEATURE_ID','PROBLEMS','BAPID'])

    dtCalcStart = time.time()
    error_count = 0
    exec_string = '''with arcpy.da.SearchCursor(INPUT_FOR_TESTING, Cursor_Field_List) as cursor:

    for row in cursor:

        summary_error_list = []

        #########################
        ##### Check Domains #####
        #########################

        for Field_Name in Domain_Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            #### Testing numeric domains here!
            Test_Value = stra(row[FieldName_Dict[Field_Name]])       # stra(row.GetValue(Field_Name))
            Test_Domain_Name = Processing_Variables['FieldsXDomains'][Field_Name]
            # print ' TEST DOMAIN:', Processing_Variables['Domains'][Test_Domain_Name]
            #CPMessage('            - Testing: ' + Field_Name + ' with value ' + stra(Test_Value), 5)


            if Test_Value in Processing_Variables['Domains'][Test_Domain_Name] or stra(Test_Value) == 'None' or stra(Test_Value) == '' or stra(Test_Value) == ' ':
                # The value is null or in the allowed list
                #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                pass
            else:
                Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                BAPID = row[FieldName_Dict['BAPID']]
                #CPMessage('            - Invalid Value : ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                #Error_List.append((Current_OID, 'Invalid Value in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                problem_string = 'Invalid Value in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['DomainErrorTable'],['CLASS','FEATURE_ID','PROBLEM','BAPID']) as det_cursor:
                #    det_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del det_cursor
                det_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)

        ########################
        ##### Check Ranges #####
        ########################

        for Field_Name in Range_Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            Test_Value = row[FieldName_Dict[Field_Name]]      # row.GetValue(Field_Name)

            Test_Range_Name = Field_Name
            # print ' TEST Range:', Processing_Variables['Ranges'][Field_Name]
            #CPMessage('            - Testing: ' + Field_Name + ' with value ' + stra(Test_Value), 5)

            Test_Validation = 'Number'

            if stra(Test_Value) in ['None', '', ' ']:
                Test_Validation = 'Empty'
                #CPMessage('            - Value OK : NULL (' + stra(Test_Value) + ').', 5)
            if Test_Validation == 'Number':
                try:
                    Test_Value = int(Test_Value)
                except:
                    Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                    Test_Validation = 'NotNumber'
                    #CPMessage('            - Invalid Value (Not an Integer): ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, 'Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = 'Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    #with arcpy.da.InsertCursor(Processing_Variables['RangeErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as ret_cursor:
                    #    ret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del ret_cursor
                    ret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)

            if Test_Validation == 'Number':
                if (Processing_Variables['Ranges'][Test_Range_Name]['Minimum'] <= Test_Value  <= Processing_Variables['Ranges'][Test_Range_Name]['Maximum']):
                    # The value is in the allowed list
                    #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                    pass
                else:
                    Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                    #CPMessage('            - Invalid Value : ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, 'Invalid Value (out of range) in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = 'Invalid Value (out of range) in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    #with arcpy.da.InsertCursor(Processing_Variables['RangeErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as ret_cursor:
                    #    ret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del ret_cursor
                    ret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)

        #################################
        ##### Check Critical Ranges #####
        #################################

        for Field_Name in Critical_Range_Check_Fields:
            # Get all item values, check to see if they match the valid ones.
            Test_Value = row[FieldName_Dict[Field_Name]]      # row.GetValue(Field_Name)

            Test_Range_Name = Field_Name
            # print ' TEST Range:', Processing_Variables['CriticalRanges'][Field_Name]
            #CPMessage('            - Testing: ' + Field_Name + ' with value ' + stra(Test_Value), 5)

            Test_Validation = 'Number'

            if stra(Test_Value) in ['None', '', ' ']:
                Test_Validation = 'Empty'
                #CPMessage('            - Value OK : NULL (' + stra(Test_Value) + ').', 5)
            if Test_Validation == 'Number':
                try:
                    Test_Value = int(Test_Value)
                except:
                    Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                    Test_Validation = 'NotNumber'
                    #CPMessage('            ** CRITICAL ERROR ** - Invalid Value (Not an Integer): ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, '** CRITICAL ERROR ** - Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = '** CRITICAL ERROR ** - Invalid Value (not an integer) in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    #with arcpy.da.InsertCursor(Processing_Variables['CriticalRangeErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as cret_cursor:
                    #    cret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del cret_cursor
                    cret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)

            if Test_Validation == 'Number':
                if (Processing_Variables['CriticalRanges'][Test_Range_Name]['Minimum'] <= Test_Value <= Processing_Variables['CriticalRanges'][Test_Range_Name]['Maximum']):
                    # The value is in the allowed list
                    #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                    pass
                else:
                    Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                    #CPMessage('            ** CRITICAL ERROR ** - Invalid Value : ' + stra(Test_Value) + ' in ' + Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, '** CRITICAL ERROR ** - invalid number of digits in ' + Field_Name + ' (' + stra(Test_Value) + ').'))
                    problem_string = '** CRITICAL ERROR ** - invalid number of digits in ' + Field_Name + ' (' + stra(Test_Value) + ').'
                    #with arcpy.da.InsertCursor(Processing_Variables['CriticalRangeErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as cret_cursor:
                    #    cret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del cret_cursor
                    cret_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)

        ######################
        ##### Check Rows #####
        ######################

        Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)

        ## CPMessage('            - Testing OID ' + stra(Current_OID) + '.', 5)

        if 'TDEC_1' in FieldName_Dict and 'TDEC_2' in FieldName_Dict and 'TDEC_3' in FieldName_Dict:
            TDEC_1 = row[FieldName_Dict['TDEC_1']]     # row.GetValue('TDEC_1')
            TDEC_2 = row[FieldName_Dict['TDEC_2']]     # row.GetValue('TDEC_2')
            TDEC_3 = row[FieldName_Dict['TDEC_3']]     # row.GetValue('TDEC_3')
            if TDEC_1 == None:
                TDEC_1 = 0
            if TDEC_2 == None:
                TDEC_2 = 0
            if TDEC_3 == None:
                TDEC_3 = 0
            TDEC_SUM = TDEC_1 + TDEC_2 + TDEC_3
            #CPMessage('                    - T-Deciles add to: ' + stra(TDEC_SUM) + '.', 5)
            if TDEC_SUM != 10 and TDEC_SUM != 0:
                #CPMessage('            - T-Deciles of OID ' + stra(Current_OID) + ' do not add to 10 or zero: ' + stra(TDEC_SUM) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile sum (TDEC_1 + TDEC_2 + TDEC_3 = ' + stra(TDEC_SUM)  + r').'))
                problem_string = 'Invalid decile sum (TDEC_1 + TDEC_2 + TDEC_3 = ' + stra(TDEC_SUM)  + r').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)


            # Check that the deciles are in order (1 >= 2 >= 3)

            if (TDEC_1 < TDEC_2):
                #CPMessage('            - TDEC_1 of OID ' + stra(Current_OID) + ' is smaller than TDEC_2: ' + stra(TDEC_1) + ' vs ' + stra(TDEC_2) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (TDEC_1 and TDEC_2).'))
                problem_string = 'Invalid decile values in (TDEC_1 and TDEC_2).'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)

            if (TDEC_2 < TDEC_3):
                #CPMessage('            - TDEC_2 of OID ' + stra(Current_OID) + ' is smaller than TDEC_3: ' + stra(TDEC_2) + ' vs ' + stra(TDEC_3) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (TDEC_2 and TDEC_3).'))
                problem_string = 'Invalid decile values in (TDEC_2 and TDEC_3).'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)

            # Check that if there is a TDEC value, then the SURFM is filled in.
            if (TDEC_1 > 0) and ('SURFM_1' in FieldName_Dict):
                SURFM_1 = stra(row[FieldName_Dict['SURFM_1']])     # stra(row.GetValue('SURFM_1'))
                if SURFM_1 in ['', 'None']:
                    #CPMessage('            - SURFM_1 of OID ' + stra(Current_OID) + ' is blank, but TDEC_1 is not.', 3)
                    #Error_List.append((Current_OID, r'SURFM_1 is unexpectedly blank.'))
                    problem_string = 'SURFM_1 is unexpectedly blank.'
                    #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                    #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del rowet_cursor
                    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)

            if (TDEC_2 > 0) and ('SURFM_2' in FieldName_Dict):
                SURFM_2 = stra(row[FieldName_Dict['SURFM_2']])     # stra(row.GetValue('SURFM_2'))
                if SURFM_2 in ['', 'None']:
                    #CPMessage('            - SURFM_2 of OID ' + stra(Current_OID) + ' is blank, but TDEC_2 is not.', 3)
                    #Error_List.append((Current_OID, r'SURFM_2 is unexpectedly blank.'))
                    problem_string = 'SURFM_2 is unexpectedly blank.'
                    #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                    #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del rowet_cursor
                    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)

            if (TDEC_3 > 0) and ('SURFM_3' in FieldName_Dict):
                SURFM_3 = stra(row[FieldName_Dict['SURFM_3']])     # stra(row.GetValue('SURFM_3'))
                if SURFM_3 in ['', 'None']:
                    #CPMessage('            - SURFM_3 of OID ' + stra(Current_OID) + ' is blank, but TDEC_3 is not.', 3)
                    #Error_List.append((Current_OID, r'SURFM_3 is unexpectedly blank.'))
                    problem_string = 'SURFM_3 is unexpectedly blank.'
                    #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                    #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del rowet_cursor
                    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)


        #else:
        #    CPMessage('                    - T-Deciles not present.', 5)

        if 'SDEC_1' in FieldName_Dict and 'SDEC_2' in FieldName_Dict and 'SDEC_3' in FieldName_Dict:
            SDEC_1 = row[FieldName_Dict['SDEC_1']]     # row.GetValue('SDEC_1')
            SDEC_2 = row[FieldName_Dict['SDEC_2']]     # row.GetValue('SDEC_2')
            SDEC_3 = row[FieldName_Dict['SDEC_3']]     # row.GetValue('SDEC_3')
            if SDEC_1 == None:
                SDEC_1 = 0
            if SDEC_2 == None:
                SDEC_2 = 0
            if SDEC_3 == None:
                SDEC_3 = 0            
            SDEC_SUM = SDEC_1 + SDEC_2 + SDEC_3
            CPMessage('                    - S-Deciles add to: ' + stra(SDEC_SUM), 5)
            if SDEC_SUM != 10 and SDEC_SUM != 0:
                #CPMessage('            - S-Deciles of OID ' + stra(Current_OID) + ' do not add to 10 or zero: ' + stra(SDEC_SUM) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile sum (SDEC_1 + SDEC_2 + SDEC_3 = ' + stra(SDEC_SUM)  + r').'))
                problem_string = 'Invalid decile sum (SDEC_1 + SDEC_2 + SDEC_3 = ' + stra(SDEC_SUM)  + r').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)

            # Check that the deciles are in order (1 >= 2 >= 3)

            if (SDEC_1 < SDEC_2):
                #CPMessage('            - SDEC_1 of OID ' + stra(Current_OID) + ' is smaller than SDEC_2: ' + stra(SDEC_1) + ' vs ' + stra(SDEC_2) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (SDEC_1 and SDEC_2).'))
                problem_string = 'Invalid decile values in (SDEC_1 and SDEC_2).'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)

            if (SDEC_2 < SDEC_3):
                #CPMessage('            - SDEC_2 of OID ' + stra(Current_OID) + ' is smaller than SDEC_3: ' + stra(SDEC_2) + ' vs ' + stra(SDEC_3) + '.', 3)
                #Error_List.append((Current_OID, r'Invalid decile values in (SDEC_2 and SDEC_3).'))
                problem_string = 'Invalid decile values in (SDEC_2 and SDEC_3).'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)

        #else:
        #    CPMessage('                    - S-Deciles not present.', 5)                


        if 'TTEX_1' in FieldName_Dict and 'TTEX_1A' in FieldName_Dict and 'TTEX_1B' in FieldName_Dict and 'TTEX_1C' in FieldName_Dict:
            TTEX_1 = stra(row[FieldName_Dict['TTEX_1']])     # stra(row.GetValue('TTEX_1'))
            TTEX_1A = row[FieldName_Dict['TTEX_1A']]     # row.GetValue('TTEX_1A')
            TTEX_1B = row[FieldName_Dict['TTEX_1B']]     # row.GetValue('TTEX_1B')
            TTEX_1C = row[FieldName_Dict['TTEX_1C']]     # row.GetValue('TTEX_1C')
            TTEX_1_Composite = stra(TTEX_1C) + stra(TTEX_1B) + stra(TTEX_1A)
            TTEX_1_Composite = TTEX_1_Composite.replace(' ', '')
            TTEX_1_Composite = TTEX_1_Composite.replace('None', '')
            TTEX_1 = TTEX_1.replace(' ', '')
            TTEX_1 = TTEX_1.replace('None', '')
            #CPMessage('                    - TTEX_1_Composite: ' + TTEX_1_Composite + ' vs ' + stra(TTEX_1), 5)
            if TTEX_1 != TTEX_1_Composite:
                #CPMessage('            - TTEX_1 of OID ' + stra(Current_OID) + ' does not match composite (' + TTEX_1 + ' vs ' + TTEX_1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTEX_1 does not match components (' + TTEX_1 + ' vs ' + TTEX_1_Composite + ').'))
                problem_string = 'TTEX_1 does not match components (' + TTEX_1 + ' vs ' + TTEX_1_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TTEX_1 components not present.', 5)  

        if 'TTEX_2' in FieldName_Dict and 'TTEX_2A' in FieldName_Dict and 'TTEX_2B' in FieldName_Dict and 'TTEX_2C' in FieldName_Dict:
            TTEX_2 = stra(row[FieldName_Dict['TTEX_2']])     # stra(row.GetValue('TTEX_2'))
            TTEX_2A = row[FieldName_Dict['TTEX_2A']]     # row.GetValue('TTEX_2A')
            TTEX_2B = row[FieldName_Dict['TTEX_2B']]     # row.GetValue('TTEX_2B')
            TTEX_2C = row[FieldName_Dict['TTEX_2C']]     # row.GetValue('TTEX_2C')
            TTEX_2_Composite = stra(TTEX_2C) + stra(TTEX_2B) + stra(TTEX_2A)
            TTEX_2_Composite = TTEX_2_Composite.replace(' ', '')
            TTEX_2_Composite = TTEX_2_Composite.replace('None', '')
            TTEX_2 = TTEX_2.replace(' ', '')
            TTEX_2 = TTEX_2.replace('None', '')
            #CPMessage('                    - TTEX_2_Composite: ' + TTEX_2_Composite + ' vs ' + stra(TTEX_2), 5)
            if TTEX_2 != TTEX_2_Composite:
                #CPMessage('            - TTEX_2 of OID ' + stra(Current_OID) + ' does not match composite (' + TTEX_2 + ' vs ' + TTEX_2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTEX_2 does not match components (' + TTEX_2 + ' vs ' + TTEX_2_Composite + ').'))
                problem_string = 'TTEX_2 does not match components (' + TTEX_2 + ' vs ' + TTEX_2_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TTEX_2 components not present.', 5)  

        if 'TTEX_3' in FieldName_Dict and 'TTEX_3A' in FieldName_Dict and 'TTEX_3B' in FieldName_Dict and 'TTEX_3C' in FieldName_Dict:
            TTEX_3 = stra(row[FieldName_Dict['TTEX_3']])     # stra(row.GetValue('TTEX_3'))
            TTEX_3A = row[FieldName_Dict['TTEX_3A']]     # row.GetValue('TTEX_3A')
            TTEX_3B = row[FieldName_Dict['TTEX_3B']]     # row.GetValue('TTEX_3B')
            TTEX_3C = row[FieldName_Dict['TTEX_3C']]     # row.GetValue('TTEX_3C')
            TTEX_3_Composite = stra(TTEX_3C) + stra(TTEX_3B) + stra(TTEX_3A)
            TTEX_3_Composite = TTEX_3_Composite.replace(' ', '')
            TTEX_3_Composite = TTEX_3_Composite.replace('None', '')
            TTEX_3 = TTEX_3.replace(' ', '')
            TTEX_3 = TTEX_3.replace('None', '')
            #CPMessage('                    - TTEX_3_Composite: ' + TTEX_3_Composite + ' vs ' + stra(TTEX_3), 5)
            if TTEX_3 != TTEX_3_Composite:
                #CPMessage('            - TTEX_3 of OID ' + stra(Current_OID) + ' does not match composite (' + TTEX_3 + ' vs ' + TTEX_3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTEX_3 does not match components (' + TTEX_3 + ' vs ' + TTEX_3_Composite + ').'))
                problem_string = 'TTEX_3 does not match components (' + TTEX_3 + ' vs ' + TTEX_3_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TTEX_3 components not present.', 5)  


        if 'TTTEX_1' in FieldName_Dict and 'TTTEX_1A' in FieldName_Dict and 'TTTEX_1B' in FieldName_Dict and 'TTTEX_1C' in FieldName_Dict:
            TTTEX_1 = stra(row[FieldName_Dict['TTTEX_1']])     # stra(row.GetValue('TTTEX_1'))
            TTTEX_1A = row[FieldName_Dict['TTTEX_1A']]     # row.GetValue('TTTEX_1A')
            TTTEX_1B = row[FieldName_Dict['TTTEX_1B']]     # row.GetValue('TTTEX_1B')
            TTTEX_1C = row[FieldName_Dict['TTTEX_1C']]     # row.GetValue('TTTEX_1C')
            TTTEX_1_Composite = stra(TTTEX_1C) + stra(TTTEX_1B) + stra(TTTEX_1A)
            TTTEX_1_Composite = TTTEX_1_Composite.replace(' ', '')
            TTTEX_1_Composite = TTTEX_1_Composite.replace('None', '')
            TTTEX_1 = TTTEX_1.replace(' ', '')
            TTTEX_1 = TTTEX_1.replace('None', '')
            #CPMessage('                    - TTTEX_1_Composite: ' + TTTEX_1_Composite + ' vs ' + stra(TTTEX_1), 5)
            if TTTEX_1 != TTTEX_1_Composite:
                #CPMessage('            - TTTEX_1 of OID ' + stra(Current_OID) + ' does not match composite (' + TTTEX_1 + ' vs ' + TTTEX_1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTTEX_1 does not match components (' + TTTEX_1 + ' vs ' + TTTEX_1_Composite + ').'))
                problem_string = 'TTTEX_1 does not match components (' + TTTEX_1 + ' vs ' + TTTEX_1_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TTTEX_1 components not present.', 5)  

        if 'TTTEX_2' in FieldName_Dict and 'TTTEX_2A' in FieldName_Dict and 'TTTEX_2B' in FieldName_Dict and 'TTTEX_2C' in FieldName_Dict:
            TTTEX_2 = stra(row[FieldName_Dict['TTTEX_2']])     # stra(row.GetValue('TTTEX_2'))
            TTTEX_2A = row[FieldName_Dict['TTTEX_2A']]     # row.GetValue('TTTEX_2A')
            TTTEX_2B = row[FieldName_Dict['TTTEX_2B']]     # row.GetValue('TTTEX_2B')
            TTTEX_2C = row[FieldName_Dict['TTTEX_2C']]     # row.GetValue('TTTEX_2C')
            TTTEX_2_Composite = stra(TTTEX_2C) + stra(TTTEX_2B) + stra(TTTEX_2A)
            TTTEX_2_Composite = TTTEX_2_Composite.replace(' ', '')
            TTTEX_2_Composite = TTTEX_2_Composite.replace('None', '')
            TTTEX_2 = TTTEX_2.replace(' ', '')
            TTTEX_2 = TTTEX_2.replace('None', '')
            #CPMessage('                    - TTTEX_2_Composite: ' + TTTEX_2_Composite + ' vs ' + stra(TTTEX_2), 5)
            if TTTEX_2 != TTTEX_2_Composite:
                #CPMessage('            - TTTEX_2 of OID ' + stra(Current_OID) + ' does not match composite (' + TTTEX_2 + ' vs ' + TTTEX_2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTTEX_2 does not match components (' + TTTEX_2 + ' vs ' + TTTEX_2_Composite + ').'))
                problem_string = 'TTTEX_2 does not match components (' + TTTEX_2 + ' vs ' + TTTEX_2_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TTTEX_2 components not present.', 5)  

        if 'TTTEX_3' in FieldName_Dict and 'TTTEX_3A' in FieldName_Dict and 'TTTEX_3B' in FieldName_Dict and 'TTTEX_3C' in FieldName_Dict:
            TTTEX_3 = stra(row[FieldName_Dict['TTTEX_3']])     # stra(row.GetValue('TTTEX_3'))
            TTTEX_3A = row[FieldName_Dict['TTTEX_3A']]     # row.GetValue('TTTEX_3A')
            TTTEX_3B = row[FieldName_Dict['TTTEX_3B']]     # row.GetValue('TTTEX_3B')
            TTTEX_3C = row[FieldName_Dict['TTTEX_3C']]     # row.GetValue('TTTEX_3C')
            TTTEX_3_Composite = stra(TTTEX_3C) + stra(TTTEX_3B) + stra(TTTEX_3A)
            TTTEX_3_Composite = TTTEX_3_Composite.replace(' ', '')
            TTTEX_3_Composite = TTTEX_3_Composite.replace('None', '')
            TTTEX_3 = TTTEX_3.replace(' ', '')
            TTTEX_3 = TTTEX_3.replace('None', '')
            #CPMessage('                    - TTTEX_3_Composite: ' + TTTEX_3_Composite + ' vs ' + stra(TTTEX_3), 5)
            if TTTEX_3 != TTTEX_3_Composite:
                #CPMessage('            - TTTEX_3 of OID ' + stra(Current_OID) + ' does not match composite (' + TTTEX_3 + ' vs ' + TTTEX_3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TTTEX_3 does not match components (' + TTTEX_3 + ' vs ' + TTTEX_3_Composite + ').'))
                problem_string = 'TTTEX_3 does not match components (' + TTTEX_3 + ' vs ' + TTTEX_3_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TTTEX_3 components not present.', 5)  


        if 'STTEX_1' in FieldName_Dict and 'STTEX_1A' in FieldName_Dict and 'STTEX_1B' in FieldName_Dict and 'STTEX_1C' in FieldName_Dict:
            STTEX_1 = stra(row[FieldName_Dict['STTEX_1']])     # stra(row.GetValue('STTEX_1'))
            STTEX_1A = row[FieldName_Dict['STTEX_1A']]     # row.GetValue('STTEX_1A')
            STTEX_1B = row[FieldName_Dict['STTEX_1B']]     # row.GetValue('STTEX_1B')
            STTEX_1C = row[FieldName_Dict['STTEX_1C']]     # row.GetValue('STTEX_1C')
            STTEX_1_Composite = stra(STTEX_1C) + stra(STTEX_1B) + stra(STTEX_1A)
            STTEX_1_Composite = STTEX_1_Composite.replace(' ', '')
            STTEX_1_Composite = STTEX_1_Composite.replace('None', '')
            STTEX_1 = STTEX_1.replace(' ', '')
            STTEX_1 = STTEX_1.replace('None', '')
            #CPMessage('                    - STTEX_1_Composite: ' + STTEX_1_Composite + ' vs ' + stra(STTEX_1), 5)
            if STTEX_1 != STTEX_1_Composite:
                #CPMessage('            - STTEX_1 of OID ' + stra(Current_OID) + ' does not match composite (' + STTEX_1 + ' vs ' + STTEX_1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'STTEX_1 does not match components (' + STTEX_1 + ' vs ' + STTEX_1_Composite + ').'))
                problem_string = 'STTEX_1 does not match components (' + STTEX_1 + ' vs ' + STTEX_1_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - STTEX_1 components not present.', 5)  

        if 'STTEX_2' in FieldName_Dict and 'STTEX_2A' in FieldName_Dict and 'STTEX_2B' in FieldName_Dict and 'STTEX_2C' in FieldName_Dict:
            STTEX_2 = stra(row[FieldName_Dict['STTEX_2']])     # stra(row.GetValue('STTEX_2'))
            STTEX_2A = row[FieldName_Dict['STTEX_2A']]     # row.GetValue('STTEX_2A')
            STTEX_2B = row[FieldName_Dict['STTEX_2B']]     # row.GetValue('STTEX_2B')
            STTEX_2C = row[FieldName_Dict['STTEX_2C']]     # row.GetValue('STTEX_2C')
            STTEX_2_Composite = stra(STTEX_2C) + stra(STTEX_2B) + stra(STTEX_2A)
            STTEX_2_Composite = STTEX_2_Composite.replace(' ', '')
            STTEX_2_Composite = STTEX_2_Composite.replace('None', '')
            STTEX_2 = STTEX_2.replace(' ', '')
            STTEX_2 = STTEX_2.replace('None', '')
            #CPMessage('                    - STTEX_2_Composite: ' + STTEX_2_Composite + ' vs ' + stra(STTEX_2), 5)
            if STTEX_2 != STTEX_2_Composite:
                #CPMessage('            - STTEX_2 of OID ' + stra(Current_OID) + ' does not match composite (' + STTEX_2 + ' vs ' + STTEX_2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'STTEX_2 does not match components (' + STTEX_2 + ' vs ' + STTEX_2_Composite + ').'))
                problem_string = 'STTEX_2 does not match components (' + STTEX_2 + ' vs ' + STTEX_2_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - STTEX_2 components not present.', 5)  

        if 'STTEX_3' in FieldName_Dict and 'STTEX_3A' in FieldName_Dict and 'STTEX_3B' in FieldName_Dict and 'STTEX_3C' in FieldName_Dict:
            STTEX_3 = stra(row[FieldName_Dict['STTEX_3']])     # stra(row.GetValue('STTEX_3'))
            STTEX_3A = row[FieldName_Dict['STTEX_3A']]     # row.GetValue('STTEX_3A')
            STTEX_3B = row[FieldName_Dict['STTEX_3B']]     # row.GetValue('STTEX_3B')
            STTEX_3C = row[FieldName_Dict['STTEX_3C']]     # row.GetValue('STTEX_3C')
            STTEX_3_Composite = stra(STTEX_3C) + stra(STTEX_3B) + stra(STTEX_3A)
            STTEX_3_Composite = STTEX_3_Composite.replace(' ', '')
            STTEX_3_Composite = STTEX_3_Composite.replace('None', '')
            STTEX_3 = STTEX_3.replace(' ', '')
            STTEX_3 = STTEX_3.replace('None', '')
            #CPMessage('                    - STTEX_3_Composite: ' + STTEX_3_Composite + ' vs ' + stra(STTEX_3), 5)
            if STTEX_3 != STTEX_3_Composite:
                #CPMessage('            - STTEX_3 of OID ' + stra(Current_OID) + ' does not match composite (' + STTEX_3 + ' vs ' + STTEX_3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'STTEX_3 does not match components (' + STTEX_3 + ' vs ' + STTEX_3_Composite + ').'))
                problem_string = 'STTEX_3 does not match components (' + STTEX_3 + ' vs ' + STTEX_3_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - STTEX_3 components not present.', 5)  


        if 'SURF_E1' in FieldName_Dict and 'SURF_E1A' in FieldName_Dict and 'SURF_E1B' in FieldName_Dict and 'SURF_E1C' in FieldName_Dict:
            SURF_E1 = stra(row[FieldName_Dict['SURF_E1']])     # stra(row.GetValue('SURF_E1'))
            SURF_E1A = row[FieldName_Dict['SURF_E1A']]     # row.GetValue('SURF_E1A')
            SURF_E1B = row[FieldName_Dict['SURF_E1B']]     # row.GetValue('SURF_E1B')
            SURF_E1C = row[FieldName_Dict['SURF_E1C']]     # row.GetValue('SURF_E1C')
            SURF_E1_Composite = stra(SURF_E1A) + stra(SURF_E1B) + stra(SURF_E1C)
            SURF_E1_Composite = SURF_E1_Composite.replace(' ', '')
            SURF_E1_Composite = SURF_E1_Composite.replace('None', '')
            SURF_E1 = SURF_E1.replace(' ', '')
            SURF_E1 = SURF_E1.replace('None', '')
            #CPMessage('                    - SURF_E1_Composite: ' + SURF_E1_Composite + ' vs ' + stra(SURF_E1), 5)
            if SURF_E1 != SURF_E1_Composite:
                #CPMessage('            - SURF_E1 of OID ' + stra(Current_OID) + ' does not match composite (' + SURF_E1 + ' vs ' + SURF_E1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SURF_E1 does not match components (' + SURF_E1 + ' vs ' + SURF_E1_Composite + ').'))
                problem_string = 'SURF_E1 does not match components (' + SURF_E1 + ' vs ' + SURF_E1_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - SURF_E1 components not present.', 5)  

        if 'SURF_E2' in FieldName_Dict and 'SURF_E2A' in FieldName_Dict and 'SURF_E2B' in FieldName_Dict and 'SURF_E2C' in FieldName_Dict:
            SURF_E2 = stra(row[FieldName_Dict['SURF_E2']])     # stra(row.GetValue('SURF_E2'))
            SURF_E2A = row[FieldName_Dict['SURF_E2A']]     # row.GetValue('SURF_E2A')
            SURF_E2B = row[FieldName_Dict['SURF_E2B']]     # row.GetValue('SURF_E2B')
            SURF_E2C = row[FieldName_Dict['SURF_E2C']]     # row.GetValue('SURF_E2C')
            SURF_E2_Composite = stra(SURF_E2A) + stra(SURF_E2B) + stra(SURF_E2C)
            SURF_E2_Composite = SURF_E2_Composite.replace(' ', '')
            SURF_E2_Composite = SURF_E2_Composite.replace('None', '')
            SURF_E2 = SURF_E2.replace(' ', '')
            SURF_E2 = SURF_E2.replace('None', '')
            #CPMessage('                    - SURF_E2_Composite: ' + SURF_E2_Composite + ' vs ' + stra(SURF_E2), 5)
            if SURF_E2 != SURF_E2_Composite:
                #CPMessage('            - SURF_E2 of OID ' + stra(Current_OID) + ' does not match composite (' + SURF_E2 + ' vs ' + SURF_E2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SURF_E2 does not match components (' + SURF_E2 + ' vs ' + SURF_E2_Composite + ').'))
                problem_string = 'SURF_E2 does not match components (' + SURF_E2 + ' vs ' + SURF_E2_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - SURF_E2 components not present.', 5)  

        if 'SURF_E3' in FieldName_Dict and 'SURF_E3A' in FieldName_Dict and 'SURF_E3B' in FieldName_Dict and 'SURF_E3C' in FieldName_Dict:
            SURF_E3 = stra(row[FieldName_Dict['SURF_E3']])     # stra(row.GetValue('SURF_E3'))
            SURF_E3A = row[FieldName_Dict['SURF_E3A']]     # row.GetValue('SURF_E3A')
            SURF_E3B = row[FieldName_Dict['SURF_E3B']]     # row.GetValue('SURF_E3B')
            SURF_E3C = row[FieldName_Dict['SURF_E3C']]     # row.GetValue('SURF_E3C')
            SURF_E3_Composite = stra(SURF_E3A) + stra(SURF_E3B) + stra(SURF_E3C)
            SURF_E3_Composite = SURF_E3_Composite.replace(' ', '')
            SURF_E3_Composite = SURF_E3_Composite.replace('None', '')
            SURF_E3 = SURF_E3.replace(' ', '')
            SURF_E3 = SURF_E3.replace('None', '')
            #CPMessage('                    - SURF_E3_Composite: ' + SURF_E3_Composite + ' vs ' + stra(SURF_E3), 5)
            if SURF_E3 != SURF_E3_Composite:
                #CPMessage('            - SURF_E3 of OID ' + stra(Current_OID) + ' does not match composite (' + SURF_E3 + ' vs ' + SURF_E3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SURF_E3 does not match components (' + SURF_E3 + ' vs ' + SURF_E3_Composite + ').'))
                problem_string = 'SURF_E3 does not match components (' + SURF_E3 + ' vs ' + SURF_E3_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - SURF_E3 components not present.', 5)  



        if 'TSURF_E1' in FieldName_Dict and 'TSURF_E1A' in FieldName_Dict and 'TSURF_E1B' in FieldName_Dict and 'TSURF_E1C' in FieldName_Dict:
            TSURF_E1 = stra(row[FieldName_Dict['TSURF_E1']])     # stra(row.GetValue('TSURF_E1'))
            TSURF_E1A = row[FieldName_Dict['TSURF_E1A']]     # row.GetValue('TSURF_E1A')
            TSURF_E1B = row[FieldName_Dict['TSURF_E1B']]     # row.GetValue('TSURF_E1B')
            TSURF_E1C = row[FieldName_Dict['TSURF_E1C']]     # row.GetValue('TSURF_E1C')
            TSURF_E1_Composite = stra(TSURF_E1A) + stra(TSURF_E1B) + stra(TSURF_E1C)
            TSURF_E1_Composite = TSURF_E1_Composite.replace(' ', '')
            TSURF_E1_Composite = TSURF_E1_Composite.replace('None', '')
            TSURF_E1 = TSURF_E1.replace(' ', '')
            TSURF_E1 = TSURF_E1.replace('None', '')
            #CPMessage('                    - TSURF_E1_Composite: ' + TSURF_E1_Composite + ' vs ' + stra(TSURF_E1), 5)
            if TSURF_E1 != TSURF_E1_Composite:
                #CPMessage('            - TSURF_E1 of OID ' + stra(Current_OID) + ' does not match composite (' + TSURF_E1 + ' vs ' + TSURF_E1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TSURF_E1 does not match components (' + TSURF_E1 + ' vs ' + TSURF_E1_Composite + ').'))
                problem_string = 'TSURF_E1 does not match components (' + TSURF_E1 + ' vs ' + TSURF_E1_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TSURF_E1 components not present.', 5)  

        if 'TSURF_E2' in FieldName_Dict and 'TSURF_E2A' in FieldName_Dict and 'TSURF_E2B' in FieldName_Dict and 'TSURF_E2C' in FieldName_Dict:
            TSURF_E2 = stra(row[FieldName_Dict['TSURF_E2']])     # stra(row.GetValue('TSURF_E2'))
            TSURF_E2A = row[FieldName_Dict['TSURF_E2A']]     # row.GetValue('TSURF_E2A')
            TSURF_E2B = row[FieldName_Dict['TSURF_E2B']]     # row.GetValue('TSURF_E2B')
            TSURF_E2C = row[FieldName_Dict['TSURF_E2C']]     # row.GetValue('TSURF_E2C')
            TSURF_E2_Composite = stra(TSURF_E2A) + stra(TSURF_E2B) + stra(TSURF_E2C)
            TSURF_E2_Composite = TSURF_E2_Composite.replace(' ', '')
            TSURF_E2_Composite = TSURF_E2_Composite.replace('None', '')
            TSURF_E2 = TSURF_E2.replace(' ', '')
            TSURF_E2 = TSURF_E2.replace('None', '')
            #CPMessage('                    - TSURF_E2_Composite: ' + TSURF_E2_Composite + ' vs ' + stra(TSURF_E2), 5)
            if TSURF_E2 != TSURF_E2_Composite:
                #CPMessage('            - TSURF_E2 of OID ' + stra(Current_OID) + ' does not match composite (' + TSURF_E2 + ' vs ' + TSURF_E2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TSURF_E2 does not match components (' + TSURF_E2 + ' vs ' + TSURF_E2_Composite + ').'))
                problem_string = 'TSURF_E2 does not match components (' + TSURF_E2 + ' vs ' + TSURF_E2_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TSURF_E2 components not present.', 5)  

        if 'TSURF_E3' in FieldName_Dict and 'TSURF_E3A' in FieldName_Dict and 'TSURF_E3B' in FieldName_Dict and 'TSURF_E3C' in FieldName_Dict:
            TSURF_E3 = stra(row[FieldName_Dict['TSURF_E3']])     # stra(row.GetValue('TSURF_E3'))
            TSURF_E3A = row[FieldName_Dict['TSURF_E3A']]     # row.GetValue('TSURF_E3A')
            TSURF_E3B = row[FieldName_Dict['TSURF_E3B']]     # row.GetValue('TSURF_E3B')
            TSURF_E3C = row[FieldName_Dict['TSURF_E3C']]     # row.GetValue('TSURF_E3C')
            TSURF_E3_Composite = stra(TSURF_E3A) + stra(TSURF_E3B) + stra(TSURF_E3C)
            TSURF_E3_Composite = TSURF_E3_Composite.replace(' ', '')
            TSURF_E3_Composite = TSURF_E3_Composite.replace('None', '')
            TSURF_E3 = TSURF_E3.replace(' ', '')
            TSURF_E3 = TSURF_E3.replace('None', '')
            #CPMessage('                    - TSURF_E3_Composite: ' + TSURF_E3_Composite + ' vs ' + stra(TSURF_E3), 5)
            if TSURF_E3 != TSURF_E3_Composite:
                #CPMessage('            - TSURF_E3 of OID ' + stra(Current_OID) + ' does not match composite (' + TSURF_E3 + ' vs ' + TSURF_E3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'TSURF_E3 does not match components (' + TSURF_E3 + ' vs ' + TSURF_E3_Composite + ').'))
                problem_string = 'TSURF_E3 does not match components (' + TSURF_E3 + ' vs ' + TSURF_E3_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - TSURF_E3 components not present.', 5)  


        if 'SSURF_E1' in FieldName_Dict and 'SSURF_E1A' in FieldName_Dict and 'SSURF_E1B' in FieldName_Dict and 'SSURF_E1C' in FieldName_Dict:
            SSURF_E1 = stra(row[FieldName_Dict['SSURF_E1']])     # stra(row.GetValue('SSURF_E1'))
            SSURF_E1A = row[FieldName_Dict['SSURF_E1A']]     # row.GetValue('SSURF_E1A')
            SSURF_E1B = row[FieldName_Dict['SSURF_E1B']]     # row.GetValue('SSURF_E1B')
            SSURF_E1C = row[FieldName_Dict['SSURF_E1C']]     # row.GetValue('SSURF_E1C')
            SSURF_E1_Composite = stra(SSURF_E1A) + stra(SSURF_E1B) + stra(SSURF_E1C)
            SSURF_E1_Composite = SSURF_E1_Composite.replace(' ', '')
            SSURF_E1_Composite = SSURF_E1_Composite.replace('None', '')
            SSURF_E1 = SSURF_E1.replace(' ', '')
            SSURF_E1 = SSURF_E1.replace('None', '')
            #CPMessage('                    - SSURF_E1_Composite: ' + SSURF_E1_Composite + ' vs ' + stra(SSURF_E1), 5)
            if SSURF_E1 != SSURF_E1_Composite:
                #CPMessage('            - SSURF_E1 of OID ' + stra(Current_OID) + ' does not match composite (' + SSURF_E1 + ' vs ' + SSURF_E1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SSURF_E1 does not match components (' + SSURF_E1 + ' vs ' + SSURF_E1_Composite + ').'))
                problem_string = 'SSURF_E1 does not match components (' + SSURF_E1 + ' vs ' + SSURF_E1_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - SSURF_E1 components not present.', 5)  

        if 'SSURF_E2' in FieldName_Dict and 'SSURF_E2A' in FieldName_Dict and 'SSURF_E2B' in FieldName_Dict and 'SSURF_E2C' in FieldName_Dict:
            SSURF_E2 = stra(row[FieldName_Dict['SSURF_E2']])     # stra(row.GetValue('SSURF_E2'))
            SSURF_E2A = row[FieldName_Dict['SSURF_E2A']]     # row.GetValue('SSURF_E2A')
            SSURF_E2B = row[FieldName_Dict['SSURF_E2B']]     # row.GetValue('SSURF_E2B')
            SSURF_E2C = row[FieldName_Dict['SSURF_E2C']]     # row.GetValue('SSURF_E2C')
            SSURF_E2_Composite = stra(SSURF_E2A) + stra(SSURF_E2B) + stra(SSURF_E2C)
            SSURF_E2_Composite = SSURF_E2_Composite.replace(' ', '')
            SSURF_E2_Composite = SSURF_E2_Composite.replace('None', '')
            SSURF_E2 = SSURF_E2.replace(' ', '')
            SSURF_E2 = SSURF_E2.replace('None', '')
            #CPMessage('                    - SSURF_E2_Composite: ' + SSURF_E2_Composite + ' vs ' + stra(SSURF_E2), 5)
            if SSURF_E2 != SSURF_E2_Composite:
                #CPMessage('            - SSURF_E2 of OID ' + stra(Current_OID) + ' does not match composite (' + SSURF_E2 + ' vs ' + SSURF_E2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SSURF_E2 does not match components (' + SSURF_E2 + ' vs ' + SSURF_E2_Composite + ').'))
                problem_string = 'SSURF_E2 does not match components (' + SSURF_E2 + ' vs ' + SSURF_E2_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
            #CPMessage('                    - SSURF_E2 components not present.', 5)  

        if 'SSURF_E3' in FieldName_Dict and 'SSURF_E3A' in FieldName_Dict and 'SSURF_E3B' in FieldName_Dict and 'SSURF_E3C' in FieldName_Dict:
            SSURF_E3 = stra(row[FieldName_Dict['SSURF_E3']])     # stra(row.GetValue('SSURF_E3'))
            SSURF_E3A = row[FieldName_Dict['SSURF_E3A']]     # row.GetValue('SSURF_E3A')
            SSURF_E3B = row[FieldName_Dict['SSURF_E3B']]     # row.GetValue('SSURF_E3B')
            SSURF_E3C = row[FieldName_Dict['SSURF_E3C']]     # row.GetValue('SSURF_E3C')
            SSURF_E3_Composite = stra(SSURF_E3A) + stra(SSURF_E3B) + stra(SSURF_E3C)
            SSURF_E3_Composite = SSURF_E3_Composite.replace(' ', '')
            SSURF_E3_Composite = SSURF_E3_Composite.replace('None', '')
            SSURF_E3 = SSURF_E3.replace(' ', '')
            SSURF_E3 = SSURF_E3.replace('None', '')
            #CPMessage('                    - SSURF_E3_Composite: ' + SSURF_E3_Composite + ' vs ' + stra(SSURF_E3), 5)
            if SSURF_E3 != SSURF_E3_Composite:
                #CPMessage('            - SSURF_E3 of OID ' + stra(Current_OID) + ' does not match composite (' + SSURF_E3 + ' vs ' + SSURF_E3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'SSURF_E3 does not match components (' + SSURF_E3 + ' vs ' + SSURF_E3_Composite + ').'))
                problem_string = 'SSURF_E3 does not match components (' + SSURF_E3 + ' vs ' + SSURF_E3_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - SSURF_E3 components not present.', 5)  

#        if 'GEOP_SCM1' in FieldName_Dict and 'GEOP_SCM1A' in FieldName_Dict and 'GEOP_SCM1B' in FieldName_Dict and 'GEOP_SCM1C' in FieldName_Dict:
        if 'GEOP_SCM1' in FieldName_Dict and 'GEOP_SCM1A' in FieldName_Dict and 'GEOP_SCM1B' in FieldName_Dict and 'GEOP_SCM1C' in FieldName_Dict and 'GEOP_INZ1' in FieldName_Dict:
            GEOP_SCM1 = stra(row[FieldName_Dict['GEOP_SCM1']])     # stra(row.GetValue('GEOP_SCM1'))
            GEOP_INZ1 = row[FieldName_Dict['GEOP_INZ1']]     # row.GetValue('GEOP_INZ1')
            GEOP_SCM1A = row[FieldName_Dict['GEOP_SCM1A']]     # row.GetValue('GEOP_SCM1A')
            GEOP_SCM1B = row[FieldName_Dict['GEOP_SCM1B']]     # row.GetValue('GEOP_SCM1B')
            GEOP_SCM1C = row[FieldName_Dict['GEOP_SCM1C']]     # row.GetValue('GEOP_SCM1C')
            if stra(GEOP_INZ1) not in ['',' ','None']:
                GEOP_SCM1_Composite = stra(GEOP_INZ1) + stra(GEOP_SCM1A) + stra(GEOP_SCM1B)
            else:
                GEOP_SCM1_Composite = stra(GEOP_SCM1A) + stra(GEOP_SCM1B) + stra(GEOP_SCM1C)
            GEOP_SCM1_Composite = GEOP_SCM1_Composite.replace(' ', '')
            GEOP_SCM1_Composite = GEOP_SCM1_Composite.replace('None', '')
            GEOP_SCM1 = GEOP_SCM1.replace(' ', '')
            GEOP_SCM1 = GEOP_SCM1.replace('None', '')
            #CPMessage('                    - GEOP_SCM1_Composite: ' + GEOP_SCM1_Composite + ' vs ' + stra(GEOP_SCM1), 5)
            if GEOP_SCM1 != GEOP_SCM1_Composite:
                #CPMessage('            - GEOP_SCM1 of OID ' + stra(Current_OID) + ' does not match composite (' + GEOP_SCM1 + ' vs ' + GEOP_SCM1_Composite + ').', 3)
                #Error_List.append((Current_OID, r'GEOP_SCM1 does not match components (' + GEOP_SCM1 + ' vs ' + GEOP_SCM1_Composite + ').'))
                problem_string = 'GEOP_SCM1 does not match components (' + GEOP_SCM1 + ' vs ' + GEOP_SCM1_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - GEOP_SCM1 components not present.', 5)  

#        if 'GEOP_SCM2' in FieldName_Dict and 'GEOP_SCM2A' in FieldName_Dict and 'GEOP_SCM2B' in FieldName_Dict and 'GEOP_SCM2C' in FieldName_Dict:
        if 'GEOP_SCM2' in FieldName_Dict and 'GEOP_SCM2A' in FieldName_Dict and 'GEOP_SCM2B' in FieldName_Dict and 'GEOP_SCM2C' in FieldName_Dict and 'GEOP_INZ2' in FieldName_Dict:
            GEOP_SCM2 = stra(row[FieldName_Dict['GEOP_SCM2']])     # stra(row.GetValue('GEOP_SCM2'))
            GEOP_INZ2 = row[FieldName_Dict['GEOP_INZ2']]     # row.GetValue('GEOP_INZ2')
            GEOP_SCM2A = row[FieldName_Dict['GEOP_SCM2A']]     # row.GetValue('GEOP_SCM2A')
            GEOP_SCM2B = row[FieldName_Dict['GEOP_SCM2B']]     # row.GetValue('GEOP_SCM2B')
            GEOP_SCM2C = row[FieldName_Dict['GEOP_SCM2C']]     # row.GetValue('GEOP_SCM2C')
            if stra(GEOP_INZ2) not in ['',' ','None']:
                GEOP_SCM2_Composite = stra(GEOP_INZ2) + stra(GEOP_SCM2A) + stra(GEOP_SCM2B)
            else:
                GEOP_SCM2_Composite = stra(GEOP_SCM2A) + stra(GEOP_SCM2B) + stra(GEOP_SCM2C)
            GEOP_SCM2_Composite = GEOP_SCM2_Composite.replace(' ', '')
            GEOP_SCM2_Composite = GEOP_SCM2_Composite.replace('None', '')
            GEOP_SCM2 = GEOP_SCM2.replace(' ', '')
            GEOP_SCM2 = GEOP_SCM2.replace('None', '')
            #CPMessage('                    - GEOP_SCM2_Composite: ' + GEOP_SCM2_Composite + ' vs ' + stra(GEOP_SCM2), 5)
            if GEOP_SCM2 != GEOP_SCM2_Composite:
                #CPMessage('            - GEOP_SCM2 of OID ' + stra(Current_OID) + ' does not match composite (' + GEOP_SCM2 + ' vs ' + GEOP_SCM2_Composite + ').', 3)
                #Error_List.append((Current_OID, r'GEOP_SCM2 does not match components (' + GEOP_SCM2 + ' vs ' + GEOP_SCM2_Composite + ').'))
                problem_string = 'GEOP_SCM2 does not match components (' + GEOP_SCM2 + ' vs ' + GEOP_SCM2_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - GEOP_SCM2 components not present.', 5)  

#        if 'GEOP_SCM3' in FieldName_Dict and 'GEOP_SCM3A' in FieldName_Dict and 'GEOP_SCM3B' in FieldName_Dict and 'GEOP_SCM3C' in FieldName_Dict:
        if 'GEOP_SCM3' in FieldName_Dict and 'GEOP_SCM3A' in FieldName_Dict and 'GEOP_SCM3B' in FieldName_Dict and 'GEOP_SCM3C' in FieldName_Dict and 'GEOP_INZ3' in FieldName_Dict:
            GEOP_SCM3 = stra(row[FieldName_Dict['GEOP_SCM3']])     # stra(row.GetValue('GEOP_SCM3'))
            GEOP_INZ3 = row[FieldName_Dict['GEOP_INZ3']]     # row.GetValue('GEOP_INZ3')
            GEOP_SCM3A = row[FieldName_Dict['GEOP_SCM3A']]     # row.GetValue('GEOP_SCM3A')
            GEOP_SCM3B = row[FieldName_Dict['GEOP_SCM3B']]     # row.GetValue('GEOP_SCM3B')
            GEOP_SCM3C = row[FieldName_Dict['GEOP_SCM3C']]     # row.GetValue('GEOP_SCM3C')
            if stra(GEOP_INZ3) not in ['',' ','None']:
                GEOP_SCM3_Composite = stra(GEOP_INZ3) + stra(GEOP_SCM3A) + stra(GEOP_SCM3B)
            else:
                GEOP_SCM3_Composite = stra(GEOP_SCM3A) + stra(GEOP_SCM3B) + stra(GEOP_SCM3C)
            GEOP_SCM3_Composite = GEOP_SCM3_Composite.replace(' ', '')
            GEOP_SCM3_Composite = GEOP_SCM3_Composite.replace('None', '')
            GEOP_SCM3 = GEOP_SCM3.replace(' ', '')
            GEOP_SCM3 = GEOP_SCM3.replace('None', '')
            #CPMessage('                    - GEOP_SCM3_Composite: ' + GEOP_SCM3_Composite + ' vs ' + stra(GEOP_SCM3), 5)
            if GEOP_SCM3 != GEOP_SCM3_Composite:
                #CPMessage('            - GEOP_SCM3 of OID ' + stra(Current_OID) + ' does not match composite (' + GEOP_SCM3 + ' vs ' + GEOP_SCM3_Composite + ').', 3)
                #Error_List.append((Current_OID, r'GEOP_SCM3 does not match components (' + GEOP_SCM3 + ' vs ' + GEOP_SCM3_Composite + ').'))
                problem_string = 'GEOP_SCM3 does not match components (' + GEOP_SCM3 + ' vs ' + GEOP_SCM3_Composite + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - GEOP_SCM3 components not present.', 5)  


        if 'SLPLL_1' in FieldName_Dict and 'SLPUL_1' in FieldName_Dict:
            SLPLL_1 = row[FieldName_Dict['SLPLL_1']]     # row.GetValue('SLPLL_1')
            if stra(SLPLL_1) == 'None':
                SLPLL_1 = 0
            SLPUL_1 = row[FieldName_Dict['SLPUL_1']]     # row.GetValue('SLPUL_1')
            if stra(SLPUL_1) == '0':
                #CPMessage('            - SLPUL_1 of OID ' + stra(Current_OID) + ' is zero.', 3)
                pass
            if stra(SLPUL_1) == 'None':
                SLPUL_1 = 0            
            if SLPLL_1 >= SLPUL_1 and SLPUL_1 != 0:
                #CPMessage('            - SLPLL_1 of OID ' + stra(Current_OID) + ' is higher than or equal to SLPUL_1 (' + stra(SLPLL_1) + ' vs ' + stra(SLPUL_1) + ').', 3)
                #Error_List.append((Current_OID, r'SLPLL_1 is higher than or equal to SLPUL_1 (' + stra(SLPLL_1) + ' vs ' + stra(SLPUL_1) + ').'))
                problem_string = 'SLPLL_1 is higher than or equal to SLPUL_1 (' + stra(SLPLL_1) + ' vs ' + stra(SLPUL_1) + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - SLPLL_1, SLPUL_1 not present.', 5)  


        if 'SLPLL_2' in FieldName_Dict and 'SLPUL_2' in FieldName_Dict:
            SLPLL_2 = row[FieldName_Dict['SLPLL_2']]     # row.GetValue('SLPLL_2')
            if stra(SLPLL_2) == 'None':
                SLPLL_2 = 0
            SLPUL_2 = row[FieldName_Dict['SLPUL_2']]     # row.GetValue('SLPUL_2')
            if stra(SLPUL_2) == '0':
                #CPMessage('            - SLPUL_2 of OID ' + stra(Current_OID) + ' is zero.', 3)            
                pass
            if stra(SLPUL_2) == 'None':
                SLPUL_2 = 0            
            if SLPLL_2 >= SLPUL_2 and SLPUL_2 != 0:
                #CPMessage('            - SLPLL_2 of OID ' + stra(Current_OID) + ' is higher than or equal to SLPUL_2 (' + stra(SLPLL_2) + ' vs ' + stra(SLPUL_2) + ').', 3)
                #Error_List.append((Current_OID, r'SLPLL_2 is higher than or equal to SLPUL_2 (' + stra(SLPLL_2) + ' vs ' + stra(SLPUL_2) + ').'))
                problem_string = 'SLPLL_2 is higher than or equal to SLPUL_2 (' + stra(SLPLL_2) + ' vs ' + stra(SLPUL_2) + ').'
                #with arcpy.da.InsertCursor(Processing_Variables['RowErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as rowet_cursor:
                #    rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                #del rowet_cursor
                rowet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                summary_error_list.append(problem_string)
        #else:
        #    CPMessage('                    - SLPLL_2, SLPUL_2 not present.', 5)  


        #if 'SITE_S1' in FieldName_Dict and 'SDEC_1' in FieldName_Dict:
        #    SITE_S1 = row[FieldName_Dict['SITE_S1']]     # row.GetValue('SITE_S1')
        #    SDEC_1 = row[FieldName_Dict['SDEC_1']]     # row.GetValue('SDEC_1')
        #    if stra(SDEC_1) != 'None' and stra(SITE_S1) in ['None','']:
        #        CPMessage('            - SITE_S1 of OID ' + stra(Current_OID) + ' is unexpectedly blank.', 3)

        #if 'SITE_S2' in FieldName_Dict and 'SDEC_2' in FieldName_Dict:
        #    SITE_S2 = row[FieldName_Dict['SITE_S2']]     # row.GetValue('SITE_S2')
        #    SDEC_2 = row[FieldName_Dict['SDEC_2']]     # row.GetValue('SDEC_2')
        #    if stra(SDEC_2) != 'None' and stra(SITE_S2) in ['None','']:
        #        CPMessage('            - SITE_S2 of OID ' + stra(Current_OID) + ' is unexpectedly blank.', 3)

        #if 'SITE_S3' in FieldName_Dict and 'SDEC_3' in FieldName_Dict:
        #    SITE_S3 = row[FieldName_Dict['SITE_S3']]     # row.GetValue('SITE_S3')
        #    SDEC_3 = row[FieldName_Dict['SDEC_3']]     # row.GetValue('SDEC_3')
        #    if stra(SDEC_3) != 'None' and stra(SITE_S3) in ['None','']:
        #        CPMessage('            - SITE_S3 of OID ' + stra(Current_OID) + ' is unexpectedly blank.', 3)

        ##################################
        ##### Check Mandatory Fields #####
        ##################################

        if CHECK_MANDATORY_FIELDS:
            for Mandatory_Field_Name in Mandatory_Check_Fields:
                Test_Value = stra(row[FieldName_Dict[Mandatory_Field_Name]])     # stra(row.GetValue(Mandatory_Field_Name))
                #CPMessage('            - Testing: ' + Mandatory_Field_Name + ' for mandatory value', 5)
                if stra(Test_Value) not in ['None','',' ']:
                    #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                    pass
                else:
                    Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                    #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name + ' at OID ' + stra(Current_OID) + '.', 4)
                    #Error_List.append((Current_OID, 'Mandatory value missing in ' + Mandatory_Field_Name + '.'))
                    problem_string = 'Mandatory value missing in ' + Mandatory_Field_Name + '.'
                    #with arcpy.da.InsertCursor(Processing_Variables['MandatoryFieldErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as mfet_cursor:
                    #    mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    #del mfet_cursor
                    mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                    summary_error_list.append(problem_string)

            Current_Project_Type = stra(row[FieldName_Dict['PROJ_TYPE']])     # stra(row.GetValue('PROJ_TYPE'))

            if 'T' in Current_Project_Type:
                for Mandatory_Field_Name_T in Mandatory_Check_Fields_T:
                    Test_Value = stra(row[FieldName_Dict[Mandatory_Field_Name_T]])     # stra(row.GetValue(Mandatory_Field_Name_T))
                    CPMessage('            - Testing: ' + Mandatory_Field_Name_T + ' for mandatory value', 5)
                    if stra(Test_Value) not in ['None','',' ']:
                        #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                        pass
                    else:
                        Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                        #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name_T + ' at OID ' + stra(Current_OID) + '.', 4)
                        #Error_List.append((Current_OID, 'Mandatory value (for %T% project types) missing in ' + Mandatory_Field_Name_T + '.'))
                        problem_string = 'Mandatory value (for %T% project types) missing in ' + Mandatory_Field_Name_T + '.'
                        #with arcpy.da.InsertCursor(Processing_Variables['MandatoryFieldErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as mfet_cursor:
                        #    mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                        #del mfet_cursor
                        mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                        summary_error_list.append(problem_string)

            if 'EM' in Current_Project_Type:
                for Mandatory_Field_Name_EM in Mandatory_Check_Fields_EM:
                    Test_Value = stra(row[FieldName_Dict[Mandatory_Field_Name_EM]])     # stra(row.GetValue(Mandatory_Field_Name_EM))
                    CPMessage('            - Testing: ' + Mandatory_Field_Name_EM + ' for mandatory value', 5)
                    if stra(Test_Value) not in ['None','',' ']:
                        #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                        pass
                    else:
                        Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                        #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name_EM + ' at OID ' + stra(Current_OID) + '.', 4)
                        #Error_List.append((Current_OID, 'Mandatory value (for %EM% project types) missing in ' + Mandatory_Field_Name_EM + '.'))
                        problem_string = 'Mandatory value (for %EM% project types) missing in ' + Mandatory_Field_Name_EM + '.'
                        #with arcpy.da.InsertCursor(Processing_Variables['MandatoryFieldErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as mfet_cursor:
                        #    mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                        #del mfet_cursor
                        mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                        summary_error_list.append(problem_string)

            if 'SEI' in Current_Project_Type:
                for Mandatory_Field_Name_SEI in Mandatory_Check_Fields_SEI:
                    Test_Value = stra(row[FieldName_Dict[Mandatory_Field_Name_SEI]])     # stra(row.GetValue(Mandatory_Field_Name_SEI))
                    CPMessage('            - Testing: ' + Mandatory_Field_Name_SEI + ' for mandatory value', 5)
                    if stra(Test_Value) not in ['None','',' ']:
                        #CPMessage('            - Value OK : ' + stra(Test_Value), 5)
                        pass
                    else:
                        Current_OID = row[FieldName_Dict[Object_ID_Field]]     # row.GetValue(Object_ID_Field)
                        #CPMessage('            - Mandatory value missing in ' + Mandatory_Field_Name_SEI + ' at OID ' + stra(Current_OID) + '.', 4)
                        #Error_List.append((Current_OID, 'Mandatory value (for %SEI% project types) missing in ' + Mandatory_Field_Name_SEI + '.'))
                        problem_string = 'Mandatory value (for %SEI% project types) missing in ' + Mandatory_Field_Name_SEI + '.'
                        #with arcpy.da.InsertCursor(Processing_Variables['MandatoryFieldErrorTable'],['CLASS','FEATURE_ID','PROBLEM']) as mfet_cursor:
                        #    mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                        #del mfet_cursor
                        mfet_cursor.insertRow([INPUT_FOR_TESTING,Current_OID,problem_string,BAPID])
                        summary_error_list.append(problem_string)

        ###############################
        ##### Write Error Summary #####
        ###############################

        if 0 < len(summary_error_list) < 51:
            if len(summary_error_list) == 1:
                problem_string = '1 Error: '
            else:
                problem_string = stra(len(summary_error_list)) + ' Errors: '
            for error in summary_error_list:
                problem_string += error + ' '
            problem_string = problem_string[:-1]
            problem_string = problem_string[:499]
            #CPMessage('            - Error summary for OID ' + stra(Current_OID) + ': ' + problem_string, 4)
            set_cursor.insertRow([Current_OID,problem_string,BAPID])
            error_count += 1
        elif len(summary_error_list) >= 51:
            problem_string = '** MANY ERRORS (' + stra(len(summary_error_list)) + ')'
            #with arcpy.da.InsertCursor(Processing_Variables['SummaryErrorTable'],['FEATURE_ID','PROBLEMS']) as set_cursor:
            #    set_cursor.insertRow([Current_OID,problem_string,BAPID])
            #del set_cursor
            set_cursor.insertRow([Current_OID,problem_string,BAPID])
            error_count += 1

        Number_Processed = Number_Processed + 1
        if (0 < Number_Processed <= 10000 and Number_Processed % 1000 == 0) or (10000 < Number_Processed <= 100000 and Number_Processed % 10000 == 0) or (100000 < Number_Processed and Number_Processed % 100000 == 0):
            dtCalcNow = time.time()
            dtCalcElapsed = dtCalcNow - dtCalcStart
            dtCalcRemaining = (dtCalcElapsed / Number_Processed) * (Number_of_Features - Number_Processed)
            CPMessage('            - Processed ' + stra(Number_Processed) + ' of ' + stra(Number_of_Features) + ' (' + stra(error_count) + ' records found with errors) (' + SanitizeElapsedTime(dtCalcElapsed)[:-1] + ' elapsed, ' + SanitizeElapsedTime(dtCalcRemaining)[:-1] + ' remaining)', 4)
            edit.stopEditing(True)
            edit.startEditing(False,False)

edit.stopEditing(True)'''

    exec(exec_string)
    CPMessage('            - Processed all ' + stra(Number_Processed), 4)

    del det_cursor
    del ret_cursor
    del cret_cursor
    del rowet_cursor
    if CHECK_MANDATORY_FIELDS:
        del mfet_cursor
    del set_cursor

    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# TestGeometry93()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TestGeometry93():
    CPMessage('    - Testing Geometry', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TestGeometry93" 

    result = arcpy.GetCount_management(INPUT_FOR_TESTING)
    Number_of_Features = int(result.GetOutput(0))
    Number_Processed = 0
    
    CPMessage('        - Scanning Geometry for ' + stra(Number_of_Features) + ' features.', 4)
    arcpy.CheckGeometry_management(INPUT_FOR_TESTING, Processing_Variables['GeometryErrorTable'])
    result = arcpy.GetCount_management(Processing_Variables['GeometryErrorTable'])
    Number_of_Errors = int(result.GetOutput(0))
    
    if Number_of_Errors > 0:
        CPMessage('            - Found ' + stra(Number_of_Errors) + ' geometry issues.', 3)
    else:
        CPMessage('            - Found no geometry issues.', 4)


    return(Processing_Variables['PROCESS_STATUS'])


#---------------------------------------------------------------------------------------------------------
# TestTolerance93()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TestTolerance93():
    CPMessage('    - Testing Tolerance', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TestTolerance93" 

    desc = arcpy.Describe(INPUT_FOR_TESTING)
    if desc.SpatialReference.XYTolerance <> 0.001:
        CPMessage('** XYTolerance is ' + stra(desc.SpatialReference.XYTolerance) + ', not the correct value of 0.001m.', 3)
        sys.exit()
    else:
        CPMessage('        - XYTolerance set to correct value of .001m.', 4)


    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# TestProjection93()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TestProjection93():
    CPMessage('    - Testing Projection', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TestProjection93" 

    # Check the Albers Projection Parameters (not just name)
    
    desc = arcpy.Describe(INPUT_FOR_TESTING)   
    CorrectAlbersSR = arcpy.CreateObject("spatialreference")
    
    CorrectAlbersSR.CreateFromFile(Processing_Variables['Albers_PRJ'])
    
    TestAlbersSR = desc.SpatialReference
    
    Tests = []
    
    ProjectionOK = True
    
    for Parameter in ['Type', 'Name', 'IsHighPrecision', 'SemiMajorAxis', 'SemiMinorAxis', 'SpheroidName', 'DatumName', 'AngularUnitName', 'FalseEasting', 'FalseNorthing', 'CentralMeridian', 'LongitudeOfOrigin', 'StandardParallel1', 'StandardParallel2', 'LinearUnitName']:
        Tests.append(('CorrectAlbersSR.' + Parameter, 'TestAlbersSR.' + Parameter))

    for Test in Tests:
        exec('Test_Result = ' + stra(Test[0]) + ' == ' + stra(Test[1]))
        if Test_Result:
            CPMessage('        - Match: ' + stra(Test[0]) + ' with ' + stra(Test[1]), 5)
        else:
            CPMessage('** Projection Issue: ' + stra(Test[0]) + ' does not match ' + stra(Test[1]), 3)
            sys.exit()
    CPMessage('        - Defined projection matches NAD_1983_BC_Environment_Albers.prj.', 4)

    ## (rough) feature envelope for the province of BC in Albers projection coordinates
    BC_Minimum_X = 263000
    BC_Maximum_X = 1881000
    BC_Minimum_Y = 349000
    BC_Maximum_Y = 1746000

    if desc.Extent.XMin < BC_Minimum_X or desc.Extent.XMax > BC_Maximum_X or desc.Extent.YMin < BC_Minimum_Y or desc.Extent.YMax > BC_Maximum_Y:
        CPMessage('** Projection Issue: Feature class extent lies outside of BC boundaries. Check that the dataset is in the proper projection.', 3)
        sys.exit()
    else:
        CPMessage('        - Feature class extent OK (lies completely within BC boundaries).', 4)

    return(Processing_Variables['PROCESS_STATUS'])



#---------------------------------------------------------------------------------------------------------
# TestGeometry101()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TestGeometry101():
    CPMessage('    - Testing Geometry', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TestGeometry101" 

    Number_of_Features = int(arcpy.GetCount_management(INPUT_FOR_TESTING).getOutput(0))
    Number_Processed = 0
    
    dtCalcStart = time.time()
    CPMessage('        - Scanning Geometry for ' + stra(Number_of_Features) + ' features (may take several hours for large datasets).', 4)
    arcpy.CheckGeometry_management(INPUT_FOR_TESTING, Processing_Variables['GeometryErrorTable'])
    dtCalcNow = time.time()
    dtCalcElapsed = dtCalcNow - dtCalcStart
    CPMessage('            - Scan completed in ' + SanitizeElapsedTime(dtCalcElapsed), 4)

    Number_of_Errors = int(arcpy.GetCount_management(Processing_Variables['GeometryErrorTable']).getOutput(0))
    
    if Number_of_Errors > 0:
        CPMessage('            - Found ' + stra(Number_of_Errors) + ' geometry issues.', 3)
    else:
        CPMessage('            - Found no geometry issues.', 4)

    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# TestTolerance101()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TestTolerance101():
    CPMessage('    - Testing Tolerance', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TestTolerance101" 

    desc = arcpy.Describe(INPUT_FOR_TESTING)
    if desc.SpatialReference.XYTolerance <> 0.001:
        CPMessage('** XYTolerance is ' + stra(desc.SpatialReference.XYTolerance) + ', not the correct value of 0.001m.', 3)
        sys.exit()
    else:
        CPMessage('        - XYTolerance set to correct value of .001m.', 4)


    return(Processing_Variables['PROCESS_STATUS'])

#---------------------------------------------------------------------------------------------------------
# TestProjection101()
#
#       Test if we need to do stuff, and do it if necessary.
#          Return 0 - nothing to do
#                 1 - successful
#                -1 - unknown failure
#
#---------------------------------------------------------------------------------------------------------
def TestProjection101():
    CPMessage('    - Testing Projection', 2)
    Processing_Variables['PROCESS_STATUS'] = 0
    Processing_Variables['PROCESS_INFO']   = "TestProjection101" 

    # Check the Albers Projection Parameters (not just name)
    
    desc = arcpy.Describe(INPUT_FOR_TESTING)   
    TestAlbersSR = desc.SpatialReference

    CorrectAlbersSR = arcpy.CreateObject("spatialreference")
    CorrectAlbersSR.createFromFile(Processing_Variables['Albers_PRJ'])
    
    Tests = []
    
    ProjectionOK = True
    
    # removed semiMinorAxis, semiMajorAxis and longitudeOfOrigin from the following comparison for Arc 10.1 because the results for these are 
    # both very near zero but slightly different between the data and the prj file, resulting in a false mismatch

    for Parameter in ['type', 'name', 'isHighPrecision', 'spheroidName', 'datumName', 'angularUnitName', 'falseEasting', 'falseNorthing', 'centralMeridian', 'standardParallel1', 'standardParallel2', 'linearUnitName']:
        Tests.append(('CorrectAlbersSR.' + Parameter, 'TestAlbersSR.' + Parameter))

    for Test in Tests:
        exec('Test_Result = ' + stra(Test[0]) + ' == ' + stra(Test[1]))
        if Test_Result:
            CPMessage('        - Match: ' + stra(Test[0]) + ' with ' + stra(Test[1]), 5)
        else:
            CPMessage('** Projection Issue: ' + stra(Test[0]) + ' does not match ' + stra(Test[1]), 3)
            sys.exit()
    CPMessage('        - Projection Matches Template.', 4)

    ## (rough) feature envelope for the province of BC in Albers projection coordinates
    BC_Minimum_X = 263000
    BC_Maximum_X = 1881000
    BC_Minimum_Y = 349000
    BC_Maximum_Y = 1746000

    if desc.Extent.XMin < BC_Minimum_X or desc.Extent.XMax > BC_Maximum_X or desc.Extent.YMin < BC_Minimum_Y or desc.Extent.YMax > BC_Maximum_Y:
        CPMessage('** Projection Issue: Feature class extent lies outside of BC boundaries. Check that the dataset is in the proper projection.', 3)
        sys.exit()
    else:
        CPMessage('        - Feature class extent OK (lies completely within BC boundaries).', 4)


    return(Processing_Variables['PROCESS_STATUS'])


#*********************************************************************************************************
#  Main Program
#*********************************************************************************************************

#---------------------------------------------------------------------------------------------------------
# Run Main for ArcGIS 9.3 and 10.0
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain93():

    Initialize93()
    SetupWorkspace93()
    
    input_description = arcpy.Describe(INPUT_FOR_TESTING)
    input_datatype = input_description.DatasetType
    if input_datatype == 'FeatureClass':
        if CHECK_GEOMETRY:
            TestGeometry93()
        TestTolerance93()
        TestProjection93()
    else:
        CPMessage(' ** Input dataset is not a feature class with geometry. No geometry testing needed. **', 3)

    LoadEverything93()

    CheckEverything93()

    # CreateAttributeErrorSummary()

    return Processing_Variables['PROCESS_STATUS']


#---------------------------------------------------------------------------------------------------------
# Run Main for ArcGIS 10.1
#
#   Main Method manages the overall program
#---------------------------------------------------------------------------------------------------------
def runMain101():

    Initialize101()
    SetupWorkspace101()
    
    input_description = arcpy.Describe(INPUT_FOR_TESTING)
    input_datatype = input_description.DatasetType
    if input_datatype == 'FeatureClass':
        if CHECK_GEOMETRY:
            TestGeometry101()
        TestTolerance101()
        TestProjection101()
    else:
        CPMessage(' ** Input dataset is not a feature class with geometry. No geometry testing needed. **', 3)

    LoadEverything101()

    CheckEverything101()

    # CreateAttributeErrorSummary()

    return Processing_Variables['PROCESS_STATUS']

#*********************************************************************************************************
#  Execution
#*********************************************************************************************************
d = arcpy.GetInstallInfo('desktop')
if len(d['Version'].split('.')) > 2:
    dv = d['Version'].rsplit('.',len(d['Version'].split('.'))-2)[0]
    del arcpy
    import arcpy
    runMain101()
else:
    dv = d['Version']
    if float(dv) <10.1:
        runMain93()
    else:
        print '** Unsupported ArcGIS version (' + str(d['Version']) + ') running on this machine.'


#---------------------------------------------------------------------------------------------------------
#  Clean-up
#---------------------------------------------------------------------------------------------------------


END_TIME       = time.ctime(time.time())
END_TIME_SEC   = time.time()
END_TIME_SQL   = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
EXECUTION_TIME = END_TIME_SEC - START_TIME_SEC
if EXECUTION_TIME < 120:
    EXECUTION_TIME_STRING = stra(int(EXECUTION_TIME)) + ' seconds.'
elif EXECUTION_TIME < 3600:
    EXECUTION_TIME_STRING = stra(round(EXECUTION_TIME / 60, 1)) + ' minutes.'
else:
    EXECUTION_TIME_STRING = stra(round(EXECUTION_TIME / 3600, 2)) + ' hours.'


CPMessage('    End Time: ' + END_TIME, 2)
CPMessage('    Execution Time: ' + EXECUTION_TIME_STRING, 4)
CPMessage('Complete: ' + PYTHON_SCRIPT, 1)
CPMessage('Next Script: CreateProjectBoundaries' + '\n' +'Preconditions for next step: Check the _Long_Tbl_CriticalRanges table, if there are errors fix and rerun ValidateData' ,1)

# arcpy.CheckInExtension("spatial")
#del gp

# Report Program Execution Status back to the calling program


if 'SendRunLogToEmail' in Processing_Variables:
    for Email in Processing_Variables['SendRunLogToEmail']:
        SendRunLogToEmail(Email)

#*********************************************************************************************************
#  End of Program
#*********************************************************************************************************

